/*
 * This file is part of ArmarX.
 *
 * ArmarX is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * ArmarX is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 *
 * @package    RobotAPI::ArmarXObjects::ObjectPoseObserver
 * @author     Rainer Kartmann ( rainer dot kartmann at kit dot edu )
 * @date       2020
 * @copyright  http://www.gnu.org/licenses/gpl-2.0.txt
 *             GNU General Public License
 */

#include "ObjectPoseObserver.h"

#include <RobotAPI/libraries/core/Pose.h>
#include <RobotAPI/libraries/core/FramedPose.h>
#include <RobotAPI/libraries/ArmarXObjects/ice_conversions.h>

#include <ArmarXCore/core/time/CycleUtil.h>
#include <ArmarXCore/observers/variant/Variant.h>

#include <VirtualRobot/Robot.h>
#include <VirtualRobot/RobotConfig.h>

#include <SimoxUtility/algorithm/get_map_keys_values.h>
#include <SimoxUtility/meta/EnumNames.hpp>


namespace armarx
{

    ObjectPoseObserverPropertyDefinitions::ObjectPoseObserverPropertyDefinitions(std::string prefix) :
        armarx::ObserverPropertyDefinitions(prefix)
    {
    }

    armarx::PropertyDefinitionsPtr ObjectPoseObserver::createPropertyDefinitions()
    {
        armarx::PropertyDefinitionsPtr defs(new ObjectPoseObserverPropertyDefinitions(getConfigIdentifier()));

        defs->defineOptionalProperty<std::string>("ObjectPoseTopicName", "ObjectPoseTopic", "Name of the Object Pose Topic.");
        defs->defineOptionalProperty<std::string>("KinematicUnitObserverName", "KinematicUnitObserver", "Name of the kinematic unit observer.");
        defs->topic(debugObserver);

        calibration.defineProperties(defs, "calibration.");
        data.defineProperties(defs);
        robotHead.defineProperties(defs, "head.");
        visu.defineProperties(defs, "visu.");

        return defs;
    }

    std::string ObjectPoseObserver::getDefaultName() const
    {
        return "ObjectPoseObserver";
    }

    void ObjectPoseObserver::onInitObserver()
    {
        data.setTag(getName());
        data.decay.setTag(getName());
        robotHead.setTag(getName());
        visu.setTag(getName());

        usingTopicFromProperty("ObjectPoseTopicName");
    }

    void ObjectPoseObserver::onConnectObserver()
    {
        // onConnect can be called multiple times, but addRobot will fail if called more than once with the same ID
        // So we need to always make sure to guard a call to addRobot
        if (!RobotState::hasRobot("robot"))
        {
            data.robot = RobotState::addRobot("robot", VirtualRobot::RobotIO::RobotDescription::eStructure);
        }
        data.robotStateComponent = getRobotStateComponent();

        getProxyFromProperty(robotHead.kinematicUnitObserver, "KinematicUnitObserverName", false, "", false);
        robotHead.debugObserver = debugObserver;
        robotHead.fetchDatafields();

        visu.arviz = arviz;
        if (!visu.updateTask)
        {
            visu.updateTask = new SimpleRunningTask<>([this]()
            {
                this->visualizeRun();
            });
            visu.updateTask->start();
        }

        createRemoteGuiTab();
        RemoteGui_startRunningTask();
    }

    void ObjectPoseObserver::onDisconnectComponent()
    {
    }

    void ObjectPoseObserver::onExitObserver()
    {
    }


    void ObjectPoseObserver::reportProviderAvailable(const std::string& providerName, const objpose::ProviderInfo& info, const Ice::Current&)
    {
        updateProviderInfo(providerName, info);
    }


    void ObjectPoseObserver::reportObjectPoses(
        const std::string& providerName, const objpose::data::ProvidedObjectPoseSeq& providedPoses, const Ice::Current&)
    {
        ARMARX_VERBOSE << "Received object " << providedPoses.size() << " poses from provider '" << providerName << "'.";
        updateObjectPoses(providerName, providedPoses);
    }


    void ObjectPoseObserver::updateProviderInfo(const std::string& providerName, const objpose::ProviderInfo& info)
    {
        if (!info.proxy)
        {
            ARMARX_WARNING << "Received availability signal by provider '" << providerName << "' "
                           << "with invalid provider proxy.\nIgnoring provider '" << providerName << "'.";
            return;
        }
        {
            std::scoped_lock lock(dataMutex);
            std::stringstream ss;
            for (const auto& id : info.supportedObjects)
            {
                ss << "- " << id << "\n";
            }
            ARMARX_VERBOSE << "Provider '" << providerName << "' available.\n"
                           << "Supported objects: \n" << ss.str();
            data.providers[providerName] = info;
        }

        if (!existsChannel(providerName))
        {
            offerChannel(providerName, "Channel of provider '" + providerName + "'.");
        }
        offerOrUpdateDataField(providerName, "objectType", objpose::ObjectTypeEnumNames.to_name(info.objectType),
                               "The object type (known or unknown)");
        offerOrUpdateDataField(providerName, "numSupportedObjects", int(info.supportedObjects.size()),
                               "Number of requestable objects.");
    }


    void ObjectPoseObserver::updateObjectPoses(const std::string& providerName, const objpose::data::ProvidedObjectPoseSeq& providedPoses)
    {
        TIMING_START(ReportObjectPoses);

        std::optional<IceUtil::Time> discardUpdatesUntil;
        bool discardAll = false;
        {
            std::scoped_lock lock(robotHeadMutex);
            if (robotHead.checkHeadVelocity)
            {
                if (robotHead.isMoving())
                {
                    robotHead.movementStarts(robotHead.discardIntervalAfterMoveMS);
                    // ARMARX_IMPORTANT << "Ignoring pose update because robot head is moving! until " << robotHead.discardUpdatesUntil;
                    discardAll = true;
                }
                else if (TimeUtil::GetTime() < robotHead.discardUpdatesUntil)
                {
                    discardAll = true;
                    // ARMARX_IMPORTANT << "Ignoring pose update because robot head has moved until: " << robotHead.discardUpdatesUntil;
                }
                else
                {
                    discardUpdatesUntil = robotHead.discardUpdatesUntil;
                }
            }
        }
        if (debugObserver)
        {
            StringVariantBaseMap map;
            map["Discarding All Updates"] = new Variant(discardAll ? 1.f : 0.f);
            if (discardAll)
            {
                map["Proportion Updated Poses"] = new Variant(0.f);
            }
            debugObserver->setDebugChannel(getName(), map);
        }

        if (discardAll)
        {
            return;
        }

        {
            std::scoped_lock lock(dataMutex);
            RobotState::synchronizeLocalClone(data.robot);

            if (data.robot->hasRobotNode(calibration.robotNode))
            {
                VirtualRobot::RobotNodePtr robotNode = data.robot->getRobotNode(calibration.robotNode);
                float value = robotNode->getJointValue();
                robotNode->setJointValue(value + calibration.offset);
            }

            // This stays empty on the first report.
            objpose::ObjectPoseSeq previousPoses;
            if (auto it = data.objectPoses.find(providerName); it != data.objectPoses.end())
            {
                previousPoses = it->second;
            }

            // Build new poses.
            objpose::ObjectPoseSeq newObjectPoses;
            int numUpdated = 0;
            for (const objpose::data::ProvidedObjectPose& provided : providedPoses)
            {
                IceUtil::Time timestamp = IceUtil::Time::microSeconds(provided.timestampMicroSeconds);

                // Check whether we have an old pose for this object.
                std::optional<objpose::ObjectPose> previousPose;
                for (const objpose::ObjectPose& prev : previousPoses)
                {
                    if (prev.objectID == fromIce(provided.objectID))
                    {
                        previousPose = prev;
                    }
                }

                if (discardUpdatesUntil && timestamp < *discardUpdatesUntil)
                {
                    if (previousPose)
                    {
                        // Keep the old one
                        newObjectPoses.push_back(*previousPose);
                    }
                    else
                    {
                        // Discard the new pose.
                        // ARMARX_IMPORTANT << "Ignoring update of object " << provided.objectID << " because robot head is moved.\n"
                        //                  << "timestamp " << timestamp << " < " << robotHead.discardUpdatesUntil;
                    }
                }
                else if (previousPose && timestamp == previousPose->timestamp)
                {
                    // Keep the old one.
                    newObjectPoses.push_back(*previousPose);
                }
                else
                {
                    numUpdated++;
                    objpose::ObjectPose& newPose = newObjectPoses.emplace_back();
                    newPose.fromProvidedPose(provided, data.robot);
                    if (newPose.objectID.dataset().empty())
                    {
                        // Try to find the data set. (It might be good to cache this.)
                        if (std::optional<ObjectInfo> objectInfo = data.objectFinder.findObject(newPose.objectID))
                        {
                            newPose.objectID = { objectInfo->dataset(), newPose.objectID.className(), newPose.objectID.instanceName() };
                        }
                    }
                    if (!provided.localOOBB)
                    {
                        // Try to load oobb from disk.
                        newPose.localOOBB = data.getObjectOOBB(newPose.objectID);
                    }
                }
            }

            if (debugObserver)
            {
                debugObserver->setDebugChannel(getName(),
                {
                    { "Discarding All Updates", new Variant(discardAll ? 1 : 0) },
                    { "Proportion Updated Poses", new Variant(static_cast<float>(numUpdated) / providedPoses.size()) }
                });
            }

            data.objectPoses[providerName] = newObjectPoses;
            handleProviderUpdate(providerName);

            TIMING_END_STREAM(ReportObjectPoses, ARMARX_VERBOSE);
            if (debugObserver)
            {
                debugObserver->setDebugChannel(getName(),
                {
                    { "ReportObjectPoses [ms]", new Variant(ReportObjectPoses.toMilliSecondsDouble()) },
                });
            }
        }
    }


    void ObjectPoseObserver::handleProviderUpdate(const std::string& providerName)
    {
        // Initialized to 0 on first access.
        if (data.providers.count(providerName) == 0)
        {
            data.providers[providerName] = objpose::ProviderInfo();
        }

        if (!existsChannel(providerName))
        {
            offerChannel(providerName, "Channel of provider '" + providerName + "'.");
        }
        offerOrUpdateDataField(providerName, "objectCount", Variant(int(data.objectPoses.at(providerName).size())), "Number of provided object poses.");
    }


    objpose::data::ObjectPoseSeq ObjectPoseObserver::getObjectPoses(const Ice::Current&)
    {
        TIMING_START(GetObjectPoses);

        TIMING_START(GetObjectPosesLock);
        std::scoped_lock lock(dataMutex);
        TIMING_END_STREAM(GetObjectPosesLock, ARMARX_VERBOSE);

        const IceUtil::Time now = TimeUtil::GetTime();
        const objpose::data::ObjectPoseSeq result = objpose::toIce(data.getObjectPoses(now));

        TIMING_END_STREAM(GetObjectPoses, ARMARX_VERBOSE);

        if (debugObserver)
        {
            debugObserver->setDebugChannel(getName(),
            {
                { "getObjectPoses() [ms]", new Variant(GetObjectPoses.toMilliSecondsDouble()) },
                { "getObjectPoses() lock [ms]", new Variant(GetObjectPosesLock.toMilliSecondsDouble()) }
            });
        }

        return result;
    }

    objpose::data::ObjectPoseSeq ObjectPoseObserver::getObjectPosesByProvider(const std::string& providerName, const Ice::Current&)
    {
        TIMING_START(GetObjectPoses);

        TIMING_START(GetObjectPosesLock);
        std::scoped_lock lock(dataMutex);
        TIMING_END_STREAM(GetObjectPosesLock, ARMARX_VERBOSE);

        const IceUtil::Time now = TimeUtil::GetTime();
        const objpose::data::ObjectPoseSeq result = objpose::toIce(data.getObjectPosesByProvider(providerName, now));

        TIMING_END_STREAM(GetObjectPoses, ARMARX_VERBOSE);

        if (debugObserver)
        {
            debugObserver->setDebugChannel(getName(),
            {
                { "getObjectPosesByProvider() [ms]", new Variant(GetObjectPoses.toMilliSecondsDouble()) },
                { "getObjectPosesByProvider() lock [ms]", new Variant(GetObjectPosesLock.toMilliSecondsDouble()) }
            });
        }

        return result;
    }


    objpose::observer::RequestObjectsOutput ObjectPoseObserver::requestObjects(
        const objpose::observer::RequestObjectsInput& input, const Ice::Current&)
    {
        std::map<std::string, objpose::provider::RequestObjectsInput> providerRequests;
        std::map<std::string, objpose::ObjectPoseProviderPrx> proxies;

        objpose::observer::RequestObjectsOutput output;

        auto updateProxy = [&](const std::string & providerName)
        {
            if (proxies.count(providerName) == 0)
            {
                if (auto it = data.providers.find(providerName); it != data.providers.end())
                {
                    proxies[providerName] = it->second.proxy;
                }
                else
                {
                    ARMARX_ERROR << "No proxy for provider ' " << providerName << "'.";
                    proxies[providerName] = nullptr;
                }
            }
        };

        if (input.provider.size() > 0)
        {
            providerRequests[input.provider] = input.request;
            updateProxy(input.provider);
        }
        else
        {
            std::scoped_lock lock(dataMutex);
            for (const auto& objectID : input.request.objectIDs)
            {
                bool found = true;
                for (const auto& [providerName, info] : data.providers)
                {
                    // ToDo: optimize look up.
                    if (std::find(info.supportedObjects.begin(), info.supportedObjects.end(), objectID) != info.supportedObjects.end())
                    {
                        providerRequests[providerName].objectIDs.push_back(objectID);
                        updateProxy(providerName);
                        break;
                    }
                }
                if (!found)
                {
                    ARMARX_ERROR << "Did not find a provider for " << objectID << ".";
                    output.results[objectID].providerName = "";
                }
            }
        }

        for (const auto& [providerName, request] : providerRequests)
        {
            if (objpose::ObjectPoseProviderPrx proxy = proxies.at(providerName); proxy)
            {
                ARMARX_INFO << "Requesting " << request.objectIDs.size() << " objects by provider '"
                            << providerName << "' for " << request.relativeTimeoutMS << " ms.";
                objpose::provider::RequestObjectsOutput providerOutput = proxy->requestObjects(request);

                int successful = 0;
                for (const auto& [objectID, result] : providerOutput.results)
                {
                    objpose::observer::ObjectRequestResult& res = output.results[objectID];
                    res.providerName = providerName;
                    res.result = result;
                    successful += int(result.success);
                }
                ARMARX_INFO << successful << " of " << request.objectIDs.size() << " object requests successful.";
            }
        }
        return output;
    }

    objpose::ProviderInfoMap ObjectPoseObserver::getAvailableProvidersInfo(const Ice::Current&)
    {
        std::scoped_lock lock(dataMutex);
        return data.providers;
    }

    Ice::StringSeq ObjectPoseObserver::getAvailableProviderNames(const Ice::Current&)
    {
        std::scoped_lock lock(dataMutex);
        return simox::alg::get_keys(data.providers);
    }

    objpose::ProviderInfo ObjectPoseObserver::getProviderInfo(const std::string& providerName, const Ice::Current&)
    {
        std::scoped_lock lock(dataMutex);
        return data.getProviderInfo(providerName);
    }

    bool ObjectPoseObserver::hasProvider(const std::string& providerName, const Ice::Current&)
    {
        std::scoped_lock lock(dataMutex);
        return data.providers.count(providerName) > 0;
    }


    objpose::AttachObjectToRobotNodeOutput ObjectPoseObserver::attachObjectToRobotNode(
        const objpose::AttachObjectToRobotNodeInput& input, const Ice::Current&)
    {
        std::scoped_lock lock(dataMutex);
        return data.attachObjectToRobotNode(input);
    }

    objpose::DetachObjectFromRobotNodeOutput ObjectPoseObserver::detachObjectFromRobotNode(
        const objpose::DetachObjectFromRobotNodeInput& input, const Ice::Current&)
    {
        std::scoped_lock lock(dataMutex);
        return data.detachObjectFromRobotNode(input);
    }

    objpose::DetachAllObjectsFromRobotNodesOutput ObjectPoseObserver::detachAllObjectsFromRobotNodes(const Ice::Current&)
    {
        std::scoped_lock lock(dataMutex);
        return data.detachAllObjectsFromRobotNodes();
    }


    objpose::AgentFramesSeq ObjectPoseObserver::getAttachableFrames(const Ice::Current&)
    {
        std::scoped_lock lock(dataMutex);

        objpose::AgentFramesSeq output;
        std::vector<VirtualRobot::RobotPtr> agents = { data.robot };
        for (VirtualRobot::RobotPtr agent : agents)
        {
            objpose::AgentFrames& frames = output.emplace_back();
            frames.agent = agent->getName();
            frames.frames = agent->getRobotNodeNames();
        }
        return output;
    }

    objpose::SignalHeadMovementOutput
    ObjectPoseObserver::signalHeadMovement(const objpose::SignalHeadMovementInput& input, const Ice::Current&)
    {
        std::scoped_lock lock(robotHeadMutex);
        return robotHead.signalHeadMovement(input);
    }


    void ObjectPoseObserver::visualizeRun()
    {
        CycleUtil cycle(static_cast<int>(1000 / visu.frequencyHz));
        while (visu.updateTask && !visu.updateTask->isStopped())
        {
            {
                std::scoped_lock lock(visuMutex);

                if (visu.enabled)
                {
                    TIMING_START(Visu);

                    std::map<std::string, objpose::ObjectPoseSeq> objectPoses;
                    ObjectFinder objectFinder;
                    float minConfidence = -1;
                    {
                        std::scoped_lock lock(dataMutex);

                        const IceUtil::Time now = TimeUtil::GetTime();
                        data.updateObjectPoses(now);
                        objectPoses = data.objectPoses;
                        objectFinder = data.objectFinder;
                        if (data.decay.enabled)
                        {
                            minConfidence = data.decay.removeObjectsBelowConfidence;
                        }
                    }
                    const std::vector<viz::Layer> layers = visu.visualizeCommit(objectPoses, minConfidence, objectFinder);
                    arviz.commit(layers);

                    TIMING_END_STREAM(Visu, ARMARX_VERBOSE);

                    if (debugObserver)
                    {
                        debugObserver->setDebugChannel(getName(),
                        {
                            { "Visualize [ms]", new Variant(Visu.toMilliSecondsDouble()) },
                        });
                    }
                }
            }
            cycle.waitForCycleDuration();
        }
    }


    objpose::ObjectPoseProviderPrx ObjectPoseObserver::getProviderProxy(const std::string& providerName)
    {
        return getProxy<objpose::ObjectPoseProviderPrx>(providerName, false, "", false);
    }

    void ObjectPoseObserver::createRemoteGuiTab()
    {
        using namespace armarx::RemoteGui::Client;

        tab.visu.setup(this->visu);
        tab.decay.setup(this->data.decay);
        tab.robotHead.setup(this->robotHead);

        VBoxLayout root = {tab.visu.group, tab.decay.group, tab.robotHead.group, VSpacer()};
        RemoteGui_createTab(getName(), root, &tab);
    }

    void ObjectPoseObserver::RemoteGui_update()
    {
        // Non-atomic variables need to be guarded by a mutex if accessed by multiple threads
        {
            std::scoped_lock lock(visuMutex);
            tab.visu.update(this->visu);
        }
        {
            std::scoped_lock lock(robotHeadMutex);
            tab.robotHead.update(this->robotHead);
        }
        {
            std::scoped_lock lock(dataMutex);
            tab.decay.update(this->data.decay);
        }
    }

    void ObjectPoseObserver::Calibration::defineProperties(PropertyDefinitionsPtr defs, const std::string& prefix)
    {
        defs->optional(robotNode, prefix + "robotNode", "Robot node which can be calibrated.");
        defs->optional(offset, prefix + "offset", "Offset for the node to be calibrated.");
    }

}

