/*
 * This file is part of ArmarX.
 *
 * ArmarX is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * ArmarX is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 *
 * @package    RobotAPI::ArmarXObjects::ObjectPoseObserver
 * @author     Rainer Kartmann ( rainer dot kartmann at kit dot edu )
 * @date       2020
 * @copyright  http://www.gnu.org/licenses/gpl-2.0.txt
 *             GNU General Public License
 */

#pragma once

#include <mutex>

#include <ArmarXCore/observers/Observer.h>

#include <ArmarXGui/libraries/ArmarXGuiComponentPlugins/LightweightRemoteGuiComponentPlugin.h>

#include <RobotAPI/interface/objectpose/ObjectPoseObserver.h>
#include <RobotAPI/libraries/RobotAPIComponentPlugins/ArVizComponentPlugin.h>
#include <RobotAPI/libraries/RobotAPIComponentPlugins/RobotStateComponentPlugin.h>

#include <RobotAPI/components/ObjectPoseObserver/detail/Data.h>
#include <RobotAPI/components/ObjectPoseObserver/detail/Decay.h>
#include <RobotAPI/components/ObjectPoseObserver/detail/Visu.h>
#include <RobotAPI/components/ObjectPoseObserver/detail/RobotHeadMovement.h>

#define ICE_CURRENT_ARG const Ice::Current& = Ice::emptyCurrent


namespace armarx
{
    /**
     * @class ObjectPoseObserverPropertyDefinitions
     * @brief Property definitions of `ObjectPoseObserver`.
     */
    class ObjectPoseObserverPropertyDefinitions :
        public ObserverPropertyDefinitions
    {
    public:
        ObjectPoseObserverPropertyDefinitions(std::string prefix);
    };



    /**
     * @defgroup Component-ObjectPoseObserver ObjectPoseObserver
     * @ingroup RobotAPI-Components
     * A description of the component ObjectPoseObserver.
     *
     * @class ObjectPoseObserver
     * @ingroup Component-ObjectPoseObserver
     * @brief Brief description of class ObjectPoseObserver.
     *
     * Detailed description of class ObjectPoseObserver.
     */
    class ObjectPoseObserver :
        virtual public Observer
        , virtual public objpose::ObjectPoseObserverInterface
        , virtual public armarx::RobotStateComponentPluginUser
        , virtual public armarx::LightweightRemoteGuiComponentPluginUser
        , virtual public armarx::ArVizComponentPluginUser
    {
        class Data;

    public:
        using RobotState = armarx::RobotStateComponentPluginUser;

        /// @see armarx::ManagedIceObject::getDefaultName()
        std::string getDefaultName() const override;


        // ObjectPoseTopic interface
    public:
        void reportProviderAvailable(const std::string& providerName, const objpose::ProviderInfo& info, ICE_CURRENT_ARG) override;
        void reportObjectPoses(const std::string& providerName, const objpose::data::ProvidedObjectPoseSeq& objectPoses, ICE_CURRENT_ARG) override;

        // ObjectPoseObserverInterface interface
    public:

        // OBJECT POSES

        objpose::data::ObjectPoseSeq getObjectPoses(ICE_CURRENT_ARG) override;
        objpose::data::ObjectPoseSeq getObjectPosesByProvider(const std::string& providerName, ICE_CURRENT_ARG) override;

        // PROVIDER INFORMATION

        bool hasProvider(const std::string& providerName, ICE_CURRENT_ARG) override;
        objpose::ProviderInfo getProviderInfo(const std::string& providerName, ICE_CURRENT_ARG) override;
        Ice::StringSeq getAvailableProviderNames(ICE_CURRENT_ARG) override;
        objpose::ProviderInfoMap getAvailableProvidersInfo(ICE_CURRENT_ARG) override;


        // REQUESTING

        objpose::observer::RequestObjectsOutput requestObjects(const objpose::observer::RequestObjectsInput& input, ICE_CURRENT_ARG) override;

        // ATTACHING

        objpose::AttachObjectToRobotNodeOutput attachObjectToRobotNode(const objpose::AttachObjectToRobotNodeInput& input, ICE_CURRENT_ARG) override;
        objpose::DetachObjectFromRobotNodeOutput detachObjectFromRobotNode(const objpose::DetachObjectFromRobotNodeInput& input, ICE_CURRENT_ARG) override;
        objpose::DetachAllObjectsFromRobotNodesOutput detachAllObjectsFromRobotNodes(ICE_CURRENT_ARG) override;

        objpose::AgentFramesSeq getAttachableFrames(ICE_CURRENT_ARG) override;

        // HEAD MOVEMENT SIGNALS

        objpose::SignalHeadMovementOutput signalHeadMovement(const objpose::SignalHeadMovementInput& input, ICE_CURRENT_ARG) override;



        // Remote GUI
        void createRemoteGuiTab();
        void RemoteGui_update() override;


    protected:

        armarx::PropertyDefinitionsPtr createPropertyDefinitions() override;

        void onInitObserver() override;
        void onConnectObserver() override;

        void onDisconnectComponent() override;
        void onExitObserver() override;


    private:

        void updateProviderInfo(const std::string& providerName, const objpose::ProviderInfo& info);

        void updateObjectPoses(const std::string& providerName, const objpose::data::ProvidedObjectPoseSeq& providedPoses);
        void handleProviderUpdate(const std::string& providerName);

        objpose::ObjectPoseProviderPrx getProviderProxy(const std::string& providerName);



        // Visualization

        void visualizeRun();


    private:

        DebugObserverInterfacePrx debugObserver;

        objpose::observer::Data data;
        std::mutex dataMutex;

        objpose::observer::RobotHeadMovement robotHead;
        std::mutex robotHeadMutex;

        objpose::observer::Visu visu;
        std::mutex visuMutex;


        struct Calibration
        {
            std::string robotNode = "Neck_2_Pitch";
            float offset = 0.0f;

            void defineProperties(armarx::PropertyDefinitionsPtr defs, const std::string& prefix = "calibration.");
        };
        Calibration calibration;


        struct RemoteGuiTab : RemoteGui::Client::Tab
        {
            objpose::observer::Visu::RemoteGui visu;
            objpose::observer::Decay::RemoteGui decay;
            objpose::observer::RobotHeadMovement::RemoteGui robotHead;
        };
        RemoteGuiTab tab;

    };

}

#undef ICE_CURRENT_ARG
