#include "Data.h"

#include <RobotAPI/libraries/core/Pose.h>
#include <RobotAPI/libraries/core/FramedPose.h>
#include <RobotAPI/libraries/core/remoterobot/RemoteRobot.h>
#include <RobotAPI/libraries/ArmarXObjects/ObjectFinder.h>
#include <RobotAPI/libraries/ArmarXObjects/ice_conversions.h>

#include <ArmarXCore/core/time/TimeUtil.h>

#include <sstream>


namespace armarx::objpose::observer
{

    void Data::defineProperties(armarx::PropertyDefinitionsPtr defs, const std::string& prefix)
    {
        decay.defineProperties(defs, prefix + "decay.");
    }

    ObjectPoseSeq Data::getObjectPoses(IceUtil::Time now)
    {
        bool synchronized = false;
        ObjectPoseSeq result;

        for (auto& [providerName, objectPoses] : objectPoses)
        {
            // Update data.
            updateObjectPoses(objectPoses, now, robot, synchronized);

            // Collect results.
            for (const ObjectPose& objectPose : objectPoses)
            {
                if (!(decay.enabled && objectPose.confidence < decay.removeObjectsBelowConfidence))
                {
                    result.push_back(objectPose);
                }
            }
        }
        return result;
    }


    ObjectPoseSeq Data::getObjectPosesByProvider(
        const std::string& providerName,
        IceUtil::Time now)
    {
        bool synchronized = false;

        // Update data.
        ObjectPoseSeq& objectPoses = this->objectPoses.at(providerName);
        updateObjectPoses(objectPoses, now, robot, synchronized);

        // Collect results.
        ObjectPoseSeq result;
        for (const ObjectPose& objectPose : objectPoses)
        {
            if (!(decay.enabled && objectPose.confidence < decay.removeObjectsBelowConfidence))
            {
                result.push_back(objectPose);
            }
        }
        return result;
    }


    void Data::updateObjectPoses(IceUtil::Time now)
    {
        bool synchronized = false;
        for (auto& [providerName, objectPoses] : objectPoses)
        {
            updateObjectPoses(objectPoses, now, robot, synchronized);
        }
    }


    void Data::updateObjectPoses(
        ObjectPoseSeq& objectPoses,
        IceUtil::Time now,
        VirtualRobot::RobotPtr agent,
        bool& agentSynchronized) const
    {
        for (ObjectPose& pose : objectPoses)
        {
            updateObjectPose(pose, now, agent, agentSynchronized);
        }
    }


    void Data::updateObjectPose(
        ObjectPose& objectPose,
        IceUtil::Time now,
        VirtualRobot::RobotPtr agent,
        bool& agentSynchronized) const
    {
        updateAttachement(objectPose, agent, agentSynchronized);

        if (decay.enabled)
        {
            decay.updateConfidence(objectPose, now);
        }
    }


    void Data::updateAttachement(
        ObjectPose& objectPose, VirtualRobot::RobotPtr agent, bool& synchronized) const
    {
        if (!objectPose.attachment)
        {
            // Fetch attachment info from internal data structure.
            auto it = attachments.find(std::make_pair(objectPose.providerName, objectPose.objectID));
            if (it != attachments.end())
            {
                // Store it in the objectPose.
                objectPose.attachment = it->second;
            }
            else
            {
                // No attachment, nothing to do.
                return;
            }
        }
        ARMARX_CHECK(objectPose.attachment);

        if (!synchronized)  // Synchronize only once.
        {
            RemoteRobot::synchronizeLocalClone(agent, robotStateComponent);
            synchronized = true;
        }
        objectPose.updateAttached(agent);
    }


    ObjectPose* Data::findObjectPose(const ObjectID& objectID, const std::string& providerName)
    {
        ObjectPose* pose = nullptr;
        if (!providerName.empty())
        {
            pose = findObjectPoseByID(objectPoses.at(providerName), objectID);
        }
        else
        {
            for (auto& [_, poses] : objectPoses)
            {
                pose = findObjectPoseByID(poses, objectID);
                if (pose)
                {
                    break;
                }
            }
        }
        return pose;
    }


    std::optional<simox::OrientedBoxf> Data::getObjectOOBB(const ObjectID& id)
    {
        return oobbCache.get(id, [this](const ObjectID & id) -> std::optional<simox::OrientedBoxf>
        {
            // Try to get OOBB from repository.
            if (std::optional<ObjectInfo> objectInfo = objectFinder.findObject(id))
            {
                try
                {
                    return objectInfo->loadOOBB();
                }
                catch (const std::ios_base::failure& e)
                {
                    // Give up - no OOBB information.
                    ARMARX_WARNING << "Could not get OOBB of object " << id << ".\n- " << e.what();
                    return std::nullopt;
                }
            }
            else
            {
                return std::nullopt;
            }
        });
    }

    ProviderInfo Data::getProviderInfo(const std::string& providerName)
    {
        try
        {
            return providers.at(providerName);
        }
        catch (const std::out_of_range&)
        {
            std::stringstream ss;
            ss << "No provider with name '" << providerName << "' available.\n";
            ss << "Available are:\n";
            for (const auto& [name, _] : providers)
            {
                ss << "- '" << name << "'\n";
            }
            throw std::out_of_range(ss.str());
        }
    }



    AttachObjectToRobotNodeOutput
    Data::attachObjectToRobotNode(const AttachObjectToRobotNodeInput& input)
    {
        AttachObjectToRobotNodeOutput output;
        output.success = false;  // We are not successful until proven otherwise.

        ObjectID objectID = armarx::fromIce(input.objectID);

        if (input.agentName != "" && input.agentName != this->robot->getName())
        {
            ARMARX_WARNING << "Tried to attach object " << objectID << " to unknown agent '" << input.agentName << "'."
                           << "\n(You can leave the agent name empty if there is only one agent.)\n"
                           << "\nKnown agents: " << std::vector<std::string> {this->robot->getName()};
            return output;
        }
        VirtualRobot::RobotPtr agent = this->robot;

        if (!agent->hasRobotNode(input.frameName))
        {
            ARMARX_WARNING << "Tried to attach object " << objectID << " to unknown node '" << input.frameName
                           << "' of agent '" << agent->getName() << "'.";
            return output;
        }
        std::string frameName = input.frameName;


        // Find object pose provider name can be empty.
        ObjectPose* currentObjectPose = this->findObjectPose(objectID, input.providerName);
        if (!currentObjectPose)
        {
            ARMARX_WARNING << "Tried to attach object " << objectID << " to node '" << frameName
                           << "' of agent '" << agent->getName() << "', but object is currently not provided.";
            return output;
        }

        ObjectAttachmentInfo info;
        info.agentName = agent->getName();
        info.frameName = frameName;

        if (input.poseInFrame)
        {
            info.poseInFrame = PosePtr::dynamicCast(input.poseInFrame)->toEigen();
        }
        else
        {
            RemoteRobot::synchronizeLocalClone(agent, robotStateComponent);

            armarx::FramedPose framed(currentObjectPose->objectPoseGlobal, armarx::GlobalFrame, agent->getName());
            if (frameName == armarx::GlobalFrame)
            {
                info.poseInFrame = framed.toGlobalEigen(this->robot);
            }
            else
            {
                framed.changeFrame(this->robot, info.frameName);
                info.poseInFrame = framed.toEigen();
            }
        }
        this->attachments[std::make_pair(currentObjectPose->providerName, objectID)] = info;

        ARMARX_INFO << "Attached object " << objectID << " by provider '" << currentObjectPose->providerName << "' "
                    << "to node '" << info.frameName << "' of agent '" << info.agentName << "'.\n"
                    << "Object pose in frame: \n" << info.poseInFrame;

        output.success = true;
        output.attachment = new data::ObjectAttachmentInfo();
        output.attachment->frameName = info.frameName;
        output.attachment->agentName = info.agentName;
        output.attachment->poseInFrame = new Pose(info.poseInFrame);

        return output;
    }

    DetachObjectFromRobotNodeOutput Data::detachObjectFromRobotNode(const DetachObjectFromRobotNodeInput& input)
    {
        ObjectID objectID = armarx::fromIce(input.objectID);
        std::string providerName = input.providerName;

        std::optional<ObjectAttachmentInfo> attachment;
        {
            // Remove from latest pose (if it was cached).
            ObjectPose* objectPose = this->findObjectPose(objectID, input.providerName);
            if (objectPose)
            {
                objectPose->attachment = std::nullopt;
            }

            if (providerName.empty() && objectPose)
            {
                providerName = objectPose->providerName;
            }
            // Remove from attachment map.
            if (input.providerName.size() > 0)
            {
                auto it = this->attachments.find(std::make_pair(input.providerName, objectID));
                if (it != this->attachments.end())
                {
                    attachment = it->second;
                    this->attachments.erase(it);
                }
            }
            else
            {
                // Search for entry with matching object ID.
                for (auto it = this->attachments.begin(); it != this->attachments.end(); ++it)
                {
                    const ObjectID& id = it->first.second;
                    if (id == objectID)
                    {
                        attachment = it->second;
                        this->attachments.erase(it);
                        break;
                    }
                }
            }
        }

        DetachObjectFromRobotNodeOutput output;
        output.wasAttached = bool(attachment);
        if (attachment)
        {
            ARMARX_INFO << "Detached object " << objectID << " by provider '" << providerName << "' from robot node '"
                        << attachment->frameName << "' of agent '" << attachment->agentName << "'.";
        }
        else
        {
            ARMARX_INFO << "Tried to detach object " << objectID << " by provider '" << providerName << "' "
                        << "from robot node, but it was not attached.";
        }

        return output;
    }

    DetachAllObjectsFromRobotNodesOutput Data::detachAllObjectsFromRobotNodes()
    {
        DetachAllObjectsFromRobotNodesOutput output;
        output.numDetached = int(this->attachments.size());

        // Clear attachment map.
        this->attachments.clear();

        // Remove from poses (if it was cached).
        for (auto& [prov, poses] : this->objectPoses)
        {
            for (auto& pose : poses)
            {
                pose.attachment = std::nullopt;
            }
        }

        ARMARX_INFO << "Detached all objects (" << output.numDetached << ") from robot nodes.";

        return output;
    }


}
