#include "Visu.h"

#include <ArmarXCore/core/time/TimeUtil.h>

#include <RobotAPI/libraries/ArmarXObjects/ObjectFinder.h>


namespace armarx::objpose::observer
{

    void Visu::defineProperties(armarx::PropertyDefinitionsPtr defs, const std::string& prefix)
    {
        defs->optional(enabled, prefix + "enabled",
                       "Enable or disable visualization of objects.");
        defs->optional(frequencyHz, prefix + "frequenzyHz",
                       "Frequency of visualization.");
        defs->optional(inGlobalFrame, prefix + "inGlobalFrame",
                       "If true, show global poses. If false, show poses in robot frame.");
        defs->optional(alpha, prefix + "alpha",
                       "Alpha of objects (1 = solid, 0 = transparent).");
        defs->optional(alphaByConfidence, prefix + "alphaByConfidence",
                       "If true, use the pose confidence as alpha (if < 1.0).");
        defs->optional(oobbs, prefix + "oobbs",
                       "Enable showing oriented bounding boxes.");
        defs->optional(objectFrames, prefix + "objectFrames",
                       "Enable showing object frames.");
        defs->optional(objectFramesScale, prefix + "objectFramesScale",
                       "Scaling of object frames.");
    }


    std::vector<viz::Layer> Visu::visualizeCommit(
        const std::map<std::string, ObjectPoseSeq>& objectPoses,
        float minConfidence,
        const ObjectFinder& objectFinder) const
    {
        std::vector<viz::Layer> layers;
        for (const auto& [name, poses] : objectPoses)
        {
            layers.push_back(visualizeProvider(name, poses, minConfidence, objectFinder));
        }
        return layers;
    }

    viz::Layer Visu::visualizeProvider(
        const std::string& providerName,
        const ObjectPoseSeq& objectPoses,
        float minConfidence,
        const ObjectFinder& objectFinder) const
    {
        viz::Layer layer = arviz.layer(providerName);
        for (const ObjectPose& objectPose : objectPoses)
        {
            const bool show = not(objectPose.confidence < minConfidence);
            if (show)
            {
                visualizeObjectPose(layer, objectPose, objectFinder);
            }
        }
        return layer;
    }

    void Visu::visualizeObjectPose(
        viz::Layer& layer,
        const ObjectPose& objectPose,
        const ObjectFinder& objectFinder) const
    {
        const armarx::ObjectID id = objectPose.objectID;
        const std::string key = id.str();

        Eigen::Matrix4f pose = inGlobalFrame ? objectPose.objectPoseGlobal : objectPose.objectPoseRobot;
        {
            viz::Object object = viz::Object(key).pose(pose);
            if (std::optional<ObjectInfo> objectInfo = objectFinder.findObject(id))
            {
                object.file(objectInfo->package(), objectInfo->simoxXML().relativePath);
            }
            else
            {
                object.fileByObjectFinder(id);
            }
            if (alphaByConfidence && objectPose.confidence < 1.0f)
            {
                object.overrideColor(simox::Color::white().with_alpha(objectPose.confidence));
            }
            else if (alpha < 1)
            {
                object.overrideColor(simox::Color::white().with_alpha(alpha));
            }
            layer.add(object);
        }

        if (oobbs && objectPose.localOOBB)
        {
            const simox::OrientedBoxf& oobb = *objectPose.localOOBB;
            layer.add(viz::Box(key + " OOBB").set(oobb.transformed(pose))
                      .color(simox::Color::lime(255, 64)));
        }
        if (objectFrames)
        {
            layer.add(viz::Pose(key + " Pose").pose(pose).scale(objectFramesScale));
        }
    }


    void Visu::RemoteGui::setup(const Visu& visu)
    {
        using namespace armarx::RemoteGui::Client;

        enabled.setValue(visu.enabled);
        inGlobalFrame.setValue(visu.inGlobalFrame);
        alpha.setRange(0, 1.0);
        alpha.setValue(visu.alpha);
        alphaByConfidence.setValue(visu.alphaByConfidence);
        oobbs.setValue(visu.oobbs);
        objectFrames.setValue(visu.objectFrames);
        {
            float max = 10000;
            objectFramesScale.setRange(0, max);
            objectFramesScale.setDecimals(2);
            objectFramesScale.setSteps(int(10 * max));
            objectFramesScale.setValue(visu.objectFramesScale);
        }

        GridLayout grid;
        int row = 0;
        grid.add(Label("Enabled"), {row, 0}).add(enabled, {row, 1});
        row++;
        grid.add(Label("Global Frame"), {row, 0}).add(inGlobalFrame, {row, 1});
        row++;
        grid.add(Label("Alpha"), {row, 0}).add(alpha, {row, 1}, {1, 3});
        row++;
        grid.add(Label("Alpha By Confidence"), {row, 0}).add(alphaByConfidence, {row, 1});
        row++;
        grid.add(Label("OOBB"), {row, 0}).add(oobbs, {row, 1});
        row++;
        grid.add(Label("Object Frames"), {row, 0}).add(objectFrames, {row, 1});
        grid.add(Label("Scale:"), {row, 2}).add(objectFramesScale, {row, 3});
        row++;

        group.setLabel("Visualization");
        group.addChild(grid);
    }

    void Visu::RemoteGui::update(Visu& visu)
    {
        visu.enabled = enabled.getValue();
        visu.inGlobalFrame = inGlobalFrame.getValue();
        visu.alpha = alpha.getValue();
        visu.alphaByConfidence = alphaByConfidence.getValue();
        visu.oobbs = oobbs.getValue();
        visu.objectFrames = objectFrames.getValue();
        visu.objectFramesScale = objectFramesScale.getValue();
    }

}
