/*
* This file is part of ArmarX.
*
* ArmarX is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License version 2 as
* published by the Free Software Foundation.
*
* ArmarX is distributed in the hope that it will be useful, but
* WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program. If not, see <http://www.gnu.org/licenses/>.
*
* @package    ArmarX::Core
* @author     Kai Welke (welke _at_ kit _dot_ edu)
* @date       2011 Kai Welke
* @copyright  http://www.gnu.org/licenses/gpl-2.0.txt
*             GNU General Public License
*/

#include "PlatformUnitObserver.h"

#include "PlatformUnit.h"

#include <ArmarXCore/observers/checks/ConditionCheckUpdated.h>
#include <ArmarXCore/observers/checks/ConditionCheckEquals.h>
#include <ArmarXCore/observers/checks/ConditionCheckInRange.h>
#include <ArmarXCore/observers/checks/ConditionCheckLarger.h>
#include <ArmarXCore/observers/checks/ConditionCheckSmaller.h>
#include <ArmarXCore/observers/checks/ConditionCheckValid.h>
#include <ArmarXCore/observers/checks/ConditionCheckEqualsWithTolerance.h>

#include <RobotAPI/libraries/core/FramedPose.h>
#include <SimoxUtility/math/convert/mat4f_to_rpy.h>


namespace armarx
{
    // ********************************************************************
    // observer framework hooks
    // ********************************************************************
    void PlatformUnitObserver::onInitObserver()
    {
        // TODO: check if platformNodeName exists?
        platformNodeName = getProperty<std::string>("PlatformName").getValue();
        robotStateComponentName = getProperty<std::string>("RobotStateComponentName").getValue();


        // register all checks
        offerConditionCheck("valid", new ConditionCheckValid());
        offerConditionCheck("updated", new ConditionCheckUpdated());
        offerConditionCheck("equals", new ConditionCheckEquals());
        offerConditionCheck("inrange", new ConditionCheckInRange());
        offerConditionCheck("approx", new ConditionCheckEqualsWithTolerance());
        offerConditionCheck("larger", new ConditionCheckLarger());
        offerConditionCheck("smaller", new ConditionCheckSmaller());

        usingTopic(platformNodeName + "State");

        usingProxyFromProperty("RobotStateComponentName");
    }

    void PlatformUnitObserver::onConnectObserver()
    {
        // register all channels
        offerChannel("platformPose", "Current Position of " + platformNodeName);
        offerChannel("targetPose", "Target Position of " + platformNodeName);
        offerChannel("platformVelocity", "Current velocity of " + platformNodeName);
        offerChannel("platformOdometryPose", "Current Odometry Position of " + platformNodeName);

        // register all data fields
        offerDataField("platformPose", "positionX", VariantType::Float, "Current X position of " + platformNodeName + " in mm");
        offerDataField("platformPose", "positionY", VariantType::Float, "Current Y position of " + platformNodeName + " in mm");
        offerDataField("platformPose", "rotation", VariantType::Float, "Current Rotation of " + platformNodeName + " in radian");

        offerDataField("targetPose", "positionX", VariantType::Float, "Target X position of " + platformNodeName + " in mm");
        offerDataField("targetPose", "positionY", VariantType::Float, "Target Y position of " + platformNodeName + " in mm");
        offerDataField("targetPose", "rotation", VariantType::Float, "Target Rotation of " + platformNodeName + " in radian");

        offerDataField("platformVelocity", "velocityX", VariantType::Float, "Current X velocity of " + platformNodeName + " in mm/s");
        offerDataField("platformVelocity", "velocityY", VariantType::Float, "Current Y velocity of " + platformNodeName + " in mm/s");
        offerDataField("platformVelocity", "velocityRotation", VariantType::Float, "Current Rotation velocity of " + platformNodeName + " in radian/s");

        offerDataField("platformOdometryPose", "positionX", VariantType::Float, "Current Odometry X position of " + platformNodeName + " in mm");
        offerDataField("platformOdometryPose", "positionY", VariantType::Float, "Current Odometry Y position of " + platformNodeName + " in mm");
        offerDataField("platformOdometryPose", "rotation", VariantType::Float, "Current Odometry Rotation of " + platformNodeName + " in radian");

        robotStateComponent = getProxyFromProperty<RobotStateComponentInterfacePrx>("RobotStateComponentName");
        const std::string robotRootFrame = robotStateComponent->getSynchronizedRobot()->getRootNode()->getName();
        const std::string agentName = robotStateComponent->getRobotName();

        // odometry pose is always zero in the beginning - set it  so that it can be queried
        PoseStamped initialOdomPose;
        initialOdomPose.header.parentFrame = OdometryFrame;
        initialOdomPose.header.frame = platformNodeName; // TODO(fabian.reister): likely incorrect
        initialOdomPose.header.agent = agentName;
        initialOdomPose.header.timestampInMicroSeconds = armarx::TimeUtil::GetTime().toMicroSeconds();

        initialOdomPose.pose = Eigen::Matrix4f::Identity();

        reportPlatformOdometryPose(initialOdomPose, Ice::emptyCurrent);
    }


    void PlatformUnitObserver::reportPlatformPose(const PoseStamped& platformPose, const Ice::Current& c)
    {
        const float yaw = simox::math::mat4f_to_rpy(platformPose.pose).z();
        const Eigen::Affine3f pose(platformPose.pose);

        nameValueMapToDataFields("platformPose", pose.translation().x(), pose.translation().y(), yaw);
        updateChannel("platformPose");
    }

    void PlatformUnitObserver::reportNewTargetPose(const PoseStamped& targetPose, const Ice::Current& c)
    {
        const float yaw = simox::math::mat4f_to_rpy(targetPose.pose).z();
        const Eigen::Affine3f pose(targetPose.pose);

        nameValueMapToDataFields("targetPose", pose.translation().x(), pose.translation().y(), yaw);
        updateChannel("targetPose");
    }

    void PlatformUnitObserver::reportPlatformVelocity(const VelocityStamped& platfromVelocity, const Ice::Current& c)
    {
        setDataField("platformVelocity", "velocityX", platfromVelocity.linearVelocity.x());
        setDataField("platformVelocity", "velocityY", platfromVelocity.linearVelocity.y());
        setDataField("platformVelocity", "velocityRotation", platfromVelocity.angularVelocity.z());
        updateChannel("platformVelocity");
    }

    void armarx::PlatformUnitObserver::reportPlatformOdometryPose(const PoseStamped& odomPose, const Ice::Current&)
    {
        const float yaw = simox::math::mat4f_to_rpy(odomPose.pose).z();
        const Eigen::Affine3f pose(odomPose.pose);

        nameValueMapToDataFields("platformOdometryPose", pose.translation().x(), pose.translation().y(), yaw);
        updateChannel("platformOdometryPose");
    }

    // ********************************************************************
    // private methods
    // ********************************************************************
    void PlatformUnitObserver::nameValueMapToDataFields(std::string channelName, ::Ice::Float platformPositionX, ::Ice::Float platformPositionY, ::Ice::Float platformRotation)
    {
        setDataField(channelName, "positionX", platformPositionX);
        setDataField(channelName, "positionY", platformPositionY);
        setDataField(channelName, "rotation", platformRotation);
    }

    PropertyDefinitionsPtr PlatformUnitObserver::createPropertyDefinitions()
    {
        return PropertyDefinitionsPtr(new PlatformUnitObserverPropertyDefinitions(
                                          getConfigIdentifier()));
    }


}
