/*
 * This file is part of ArmarX.
 *
 * ArmarX is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * ArmarX is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 *
 * @package    RobotAPI::ArmarXObjects::PlatformSubUnit
 * @author     Raphael ( raphael dot grimm at kit dot edu )
 * @date       2017
 * @copyright  http://www.gnu.org/licenses/gpl-2.0.txt
 *             GNU General Public License
 */
#include "PlatformSubUnit.h"

#include <Eigen/Core>
#include <Eigen/Geometry>

#include <RobotAPI/interface/core/PoseBase.h>
#include <RobotAPI/libraries/core/FramedPose.h>

namespace armarx
{

    void armarx::PlatformSubUnit::update(const armarx::SensorAndControl& sc, const JointAndNJointControllers&)
    {
        if (!getProxy())
        {
            //this unit is not initialized yet
            ARMARX_IMPORTANT << deactivateSpam(1) << "not initialized yet - skipping this update";
            return;
        }
        if (!listenerPrx)
        {
            ARMARX_INFO << deactivateSpam(1) << "listener is not set - skipping this update";
            return;
        }
        const SensorValueBase* sensorValue = sc.sensors.at(platformSensorIndex).get();
        ARMARX_CHECK_EXPRESSION(sensorValue);
        std::lock_guard<std::mutex> guard {dataMutex};

        const auto timestamp = sc.sensorValuesTimestamp.toMicroSeconds();;

        // odom velocity is in local robot frame
        FrameHeader odomVelocityHeader;
        odomVelocityHeader.parentFrame = robotRootFrame;
        odomVelocityHeader.frame = robotRootFrame;
        odomVelocityHeader.agent = agentName;
        odomVelocityHeader.timestampInMicroSeconds = timestamp;

        // odom pose is in odom frame
        FrameHeader odomPoseHeader;
        odomPoseHeader.parentFrame = OdometryFrame;
        odomPoseHeader.frame = robotRootFrame;
        odomPoseHeader.agent = agentName;
        odomPoseHeader.timestampInMicroSeconds = timestamp;

        // odom velocity is in local robot frame
        FrameHeader globalPoseHeader;
        globalPoseHeader.parentFrame = GlobalFrame;
        globalPoseHeader.frame = robotRootFrame;
        globalPoseHeader.agent = agentName;
        globalPoseHeader.timestampInMicroSeconds = timestamp;

        //pos
        {
            ARMARX_CHECK_EXPRESSION(sensorValue->isA<SensorValueHolonomicPlatformRelativePosition>());
            const SensorValueHolonomicPlatformRelativePosition* s = sensorValue->asA<SensorValueHolonomicPlatformRelativePosition>();
            relX = s->relativePositionX;
            relY = s->relativePositionY;
            relR = s->relativePositionRotation;


            // @@@ CHECK BELOW:
            if (s->isA<SensorValueHolonomicPlatformAbsolutePosition>())
            {
                const SensorValueHolonomicPlatformAbsolutePosition* sabs = s->asA<SensorValueHolonomicPlatformAbsolutePosition>();
                abs = positionCorrection * VirtualRobot::MathTools::posrpy2eigen4f(sabs->absolutePositionX, sabs->absolutePositionY, 0, 0, 0, sabs->absolutePositionRotation);

                PoseStamped currentPose;
                currentPose.header = globalPoseHeader;
                currentPose.pose = abs;

                listenerPrx->reportPlatformPose(currentPose);
                globalPosCtrl->setGlobalPos(currentPose);
            }
            else
            {
                abs = positionCorrection * VirtualRobot::MathTools::posrpy2eigen4f(relX, relY, 0, 0, 0, relR);
            }

            PoseStamped odomPose;
            odomPose.header = odomPoseHeader;
            odomPose.pose = VirtualRobot::MathTools::posrpy2eigen4f(relX, relY, 0, 0, 0, relR);
            listenerPrx->reportPlatformOdometryPose(odomPose);
        }
        //vel
        {
            ARMARX_CHECK_EXPRESSION(sensorValue->isA<SensorValueHolonomicPlatformVelocity>());
            const SensorValueHolonomicPlatformVelocity* s = sensorValue->asA<SensorValueHolonomicPlatformVelocity>();

            VelocityStamped odomVelocity;
            odomVelocity.header = odomVelocityHeader;
            odomVelocity.linearVelocity << s->velocityX, s->velocityY, 0;
            odomVelocity.angularVelocity << 0, 0, s->velocityRotation;

            listenerPrx->reportPlatformVelocity(odomVelocity);
        }
    }

    void armarx::PlatformSubUnit::move(Ice::Float vx, Ice::Float vy, Ice::Float vr, const Ice::Current&)
    {
        //holding the mutex here could deadlock
        if (!pt->isControllerActive())
        {
            pt->activateController();
        }
        std::lock_guard<std::mutex> guard {dataMutex};
        pt->setVelocites(
            std::clamp(vx, -maxVLin, maxVLin),
            std::clamp(vy, -maxVLin, maxVLin),
            std::clamp(vr, -maxVAng, maxVAng)
        );
    }

    void armarx::PlatformSubUnit::moveTo(Ice::Float rx, Ice::Float ry, Ice::Float rr, Ice::Float lac, Ice::Float rac, const Ice::Current&)
    {
        globalPosCtrl->setTarget(rx, ry, rr, lac, rac);
        if (!globalPosCtrl->isControllerActive())
        {
            globalPosCtrl->activateController();
        };
    }

    void armarx::PlatformSubUnit::moveRelative(Ice::Float rx, Ice::Float ry, Ice::Float rr, Ice::Float lac, Ice::Float rac, const Ice::Current&)
    {
        relativePosCtrl->setTarget(rx, ry, rr, lac, rac);
        if (!relativePosCtrl->isControllerActive())
        {
            relativePosCtrl->activateController();
        }
        //holding the mutex here could deadlock
        //    std::lock_guard<std::mutex> guard {dataMutex};
        ARMARX_INFO << "target orientation: " << rr;

    }

    void armarx::PlatformSubUnit::setMaxVelocities(Ice::Float mxVLin, Ice::Float mxVAng, const Ice::Current&)
    {
        std::lock_guard<std::mutex> guard {dataMutex};
        maxVLin = std::abs(mxVLin);
        maxVAng = std::abs(mxVAng);
    }

    void armarx::PlatformSubUnit::setGlobalPose(armarx::PoseBasePtr globalPose, const Ice::Current&)
    {
        std::lock_guard<std::mutex> guard {dataMutex};
        PosePtr p = PosePtr::dynamicCast(globalPose);
        positionCorrection = p->toEigen() * abs.inverse();
    }

    armarx::PoseBasePtr armarx::PlatformSubUnit::getGlobalPose(const Ice::Current&)
    {
        std::lock_guard<std::mutex> guard {dataMutex};
        return new Pose {abs};
    }

    void armarx::PlatformSubUnit::stopPlatform(const Ice::Current& c)
    {
        move(0, 0, 0);
    }

    Eigen::Matrix4f armarx::PlatformSubUnit::getRelativePose() const
    {
        Eigen::Matrix4f rp = VirtualRobot::MathTools::rpy2eigen4f(0, 0, relR);
        rp(0, 3) = relX;
        rp(1, 3) = relY;
        return rp;
    }


    void armarx::PlatformSubUnit::reportPlatformPose(const PoseStamped& platformPose, const Ice::Current& c)
    {
        globalPosCtrl->setGlobalPos(platformPose);
    }

    void armarx::PlatformSubUnit::reportNewTargetPose(const PoseStamped& /*targetPlatformPose*/, const Ice::Current& c)
    {
    }

    void armarx::PlatformSubUnit::reportPlatformVelocity(const VelocityStamped&, const Ice::Current& c)
    {
    }

    void armarx::PlatformSubUnit::reportPlatformOdometryPose(const PoseStamped&, const Ice::Current&)
    {
    }

}