/*
 * This file is part of ArmarX.
 *
 * ArmarX is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * ArmarX is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 *
 * \package    RobotAPI::gui-plugins::ArMemMemoryViewerWidgetController
 * \author     Rainer Kartmann ( rainer dot kartmann at kit dot edu )
 * \date       2020
 * \copyright  http://www.gnu.org/licenses/gpl-2.0.txt
 *             GNU General Public License
 */

#include "ArMemMemoryViewerWidgetController.h"

#include <string>

#include <QSplitter>
#include <QTimer>

#include <ArmarXCore/observers/variant/Variant.h>
#include <ArmarXCore/core/time/TimeUtil.h>

#include <RobotAPI/libraries/armem/core/ice_conversions.h>
#include <RobotAPI/libraries/armem_gui/gui_utils.h>


namespace armarx
{

    template <class WidgetT>
    void replaceWidget(WidgetT*& old, QWidget* neu, QLayout* parentLayout)
    {
        QLayoutItem* oldItem = parentLayout->replaceWidget(old, neu);
        if (oldItem)
        {
            delete oldItem;
            delete old;
            old = nullptr;
        }
    }

    static QSplitter* useSplitter(QLayout* layout)
    {
        ARMARX_CHECK(layout);

        QSplitter* splitter;
        if (dynamic_cast<QHBoxLayout*>(layout))
        {
            splitter = new QSplitter(Qt::Orientation::Horizontal);
        }
        else if (dynamic_cast<QVBoxLayout*>(layout))
        {
            splitter = new QSplitter(Qt::Orientation::Vertical);
        }
        else
        {
            splitter = new QSplitter();
        }

        while (layout->count() > 0)
        {
            int index = 0;
            if (layout->itemAt(index)->widget())
            {
                QLayoutItem* item = layout->takeAt(index);
                splitter->addWidget(item->widget());
            }
        }
        layout->addWidget(splitter);

        return splitter;
    }


    QString ArMemMemoryViewerWidgetController::GetWidgetName()
    {
        return "MemoryX.ArMemMemoryViewer";
    }


    ArMemMemoryViewerWidgetController::ArMemMemoryViewerWidgetController()
    {
        widget.setupUi(getWidget());

        // Setup widgets
        widget.statusLabel->clear();


        // Update timer
        updateTimer = new QTimer(this);
        updateTimer->setInterval(static_cast<int>(std::round(1000 / widget.updateFrequencySpinBox->value())));
        widget.updateFrequencySpinBox->setEnabled(widget.updateAutoCheckBox->isChecked());


        memoryTree = new armem::gui::MemoryTreeWidget();
        replaceWidget(widget._memoryTreeWidget, memoryTree, widget.memoryGroupBox->layout());
        ARMARX_CHECK_NULL(widget._memoryTreeWidget);

        queryTabs = new armem::gui::QueryWidget();
        widget._queryTabWidget = nullptr;

        armarx::gui::clearLayout(widget.queryGroupBox->layout());
        widget.queryGroupBox->layout()->addWidget(queryTabs);


        instanceView = new armem::gui::InstanceView();
        instanceView->setStatusLabel(widget.statusLabel);
        replaceWidget(widget._instanceTree, instanceView, widget.instanceGroupBox->layout());
        ARMARX_CHECK_NULL(widget._instanceTree);


        useSplitter(widget.treesLayout);
        useSplitter(widget.memoryGroupBox->layout());


        // Connect signals
        connect(this, &This::connected, this, &This::updateMemory);
        connect(widget.updateButton, &QPushButton::pressed, this, &This::updateMemory);
        connect(updateTimer, &QTimer::timeout, this, &This::updateMemory);

        connect(widget.updateAutoCheckBox, &QCheckBox::toggled, widget.updateFrequencySpinBox, &QDoubleSpinBox::setEnabled);
        connect(widget.updateAutoCheckBox, &QCheckBox::toggled, this, &This::toggleTimer);
        connect(widget.updateFrequencySpinBox, &QDoubleSpinBox::editingFinished, this, &This::updateTimerFrequency);

        connect(this, &This::memoryDataChanged, this, &This::updateMemoryTree);
        connect(memoryTree, &armem::gui::MemoryTreeWidget::selectedItemChanged, this, &This::updateInstanceTree);

        connect(widget.instanceUseTypeInfoCheckBox, &QCheckBox::toggled, instanceView, &armem::gui::InstanceView::setUseTypeInfo);

        connect(memoryTree, &armem::gui::MemoryTreeWidget::updated, this, &This::memoryTreeUpdated);
        connect(instanceView, &armem::gui::InstanceView::updated, this, &This::instanceTreeUpdated);
    }

    ArMemMemoryViewerWidgetController::~ArMemMemoryViewerWidgetController()
    {
    }


    void ArMemMemoryViewerWidgetController::onInitComponent()
    {
        if (memoryName.size() > 0)
        {
            usingProxy(memoryName);
        }
        if (debugObserverName.size() > 0)
        {
            usingProxy(debugObserverName);
        }

        emit initialized();
    }

    void ArMemMemoryViewerWidgetController::onConnectComponent()
    {
        if (!memoryName.empty())
        {
            getProxy(memory, memoryName);
            memoryReader = armem::client::Reader(memory);
        }
        // DebugObserver is optional (check for null on every call)
        if (!debugObserverName.empty())
        {
            getProxy(debugObserver, debugObserverName, false, "", false);
        }

        toggleTimer(widget.updateAutoCheckBox->isChecked());

        emit connected();
    }


    void ArMemMemoryViewerWidgetController::onDisconnectComponent()
    {
        if (updateTimer)
        {
            updateTimer->stop();
        }
    }

    void ArMemMemoryViewerWidgetController::updateTimerFrequency()
    {
        updateTimer->setInterval(static_cast<int>(std::round(1000 / widget.updateFrequencySpinBox->value())));
    }

    void ArMemMemoryViewerWidgetController::toggleTimer(bool start)
    {
        if (start)
        {
            updateTimer->start();
        }
        else
        {
            updateTimer->stop();
        }
    }


    void ArMemMemoryViewerWidgetController::updateMemoryTree()
    {
        if (!memoryData)
        {
            return;
        }
        TIMING_START(GuiUpdate);
        memoryTree->update(*memoryData);
        TIMING_END_STREAM(GuiUpdate, ARMARX_VERBOSE);

        if (debugObserver)
        {
            debugObserver->setDebugDatafield(debugObserverChannelName, "GUI Update [ms]", new Variant(GuiUpdate.toMilliSecondsDouble()));
        }
    }


    void ArMemMemoryViewerWidgetController::updateMemory()
    {
        if (!memoryReader)
        {
            return;
        }

        TIMING_START(MemoryQuery);
        {
            armem::client::QueryInput input = queryTabs->queryInput();
            armem::client::QueryResult result = memoryReader.query(input);
            if (result)
            {
                this->memoryData = std::move(result.memory);
            }
            else
            {
                widget.statusLabel->setText(QString::fromStdString(result.errorMessage));
            }
        }
        TIMING_END_STREAM(MemoryQuery, ARMARX_VERBOSE);

        if (debugObserver)
        {
            debugObserver->setDebugDatafield(debugObserverChannelName, "Memory Query [ms]", new Variant(MemoryQuery.toMilliSecondsDouble()));
        }

        if (this->memoryData)
        {
            emit memoryDataChanged();
        }
        else
        {
            widget.statusLabel->setText("No query result.");
        }
    }

    void ArMemMemoryViewerWidgetController::updateInstanceTree(const armem::MemoryID& selectedID)
    {
        if (memoryData)
        {
            if (!selectedID.hasEntityName())
            {
                return;
            }
            armem::MemoryID id = selectedID;
            const armem::EntitySnapshot* snapshot = nullptr;
            if (!id.hasTimestamp())
            {
                const armem::Entity& entity = memoryData->getEntity(id);
                if (entity.empty())
                {
                    return;
                }
                snapshot = &entity.getLatestSnapshot();
                id.timestamp = snapshot->time();
            }
            if (!id.hasInstanceIndex())
            {
                if (!snapshot)
                {
                    try
                    {
                        snapshot = &memoryData->getEntitySnapshot(id);
                    }
                    catch (const armem::error::ArMemError& e)
                    {
                        widget.statusLabel->setText(e.what());
                    }
                }
                if (snapshot && snapshot->size() > 0)
                {
                    id.instanceIndex = 0;
                }
            }
            if (id.hasInstanceIndex())
            {
                instanceView->update(id, *memoryData);
            }
        }
    }


    const static std::string CONFIG_KEY_MEMORY = "MemoryName";
    const static std::string CONFIG_KEY_DEBUG_OBSERVER = "DebugObserverName";

    void ArMemMemoryViewerWidgetController::loadSettings(QSettings* settings)
    {
        memoryName = settings->value(QString::fromStdString(CONFIG_KEY_MEMORY), "Memory").toString().toStdString();
        debugObserverName = settings->value(QString::fromStdString(CONFIG_KEY_DEBUG_OBSERVER), "DebugObserver").toString().toStdString();
    }
    void ArMemMemoryViewerWidgetController::saveSettings(QSettings* settings)
    {
        settings->setValue(QString::fromStdString(CONFIG_KEY_MEMORY), QString::fromStdString(memoryName));
        settings->setValue(QString::fromStdString(CONFIG_KEY_DEBUG_OBSERVER), QString::fromStdString(debugObserverName));
    }

    QPointer<QDialog> ArMemMemoryViewerWidgetController::getConfigDialog(QWidget* parent)
    {
        if (!configDialog)
        {
            configDialog = new SimpleConfigDialog(parent);
            configDialog->addProxyFinder<armem::server::MemoryInterfacePrx>({CONFIG_KEY_MEMORY, "Memory", "*Memory"});
            configDialog->addProxyFinder<armarx::DebugObserverInterfacePrx>({CONFIG_KEY_DEBUG_OBSERVER, "Debug Observer", "DebugObserver"});
        }
        return qobject_cast<QDialog*>(configDialog);
    }

    void ArMemMemoryViewerWidgetController::configured()
    {
        if (configDialog)
        {
            memoryName = configDialog->getProxyName(CONFIG_KEY_MEMORY);
            debugObserverName = configDialog->getProxyName(CONFIG_KEY_DEBUG_OBSERVER);
        }
    }
}
