#include "CoreSegment.h"

#include <ArmarXCore/core/exceptions/local/ExpressionException.h>

#include "error.h"


namespace armarx::armem
{

    CoreSegment::CoreSegment()
    {
    }

    CoreSegment::CoreSegment(const std::string& name, aron::typenavigator::AronObjectTypeNavigatorPtr aronType) :
        CoreSegment(name, MemoryID(), aronType)
    {
    }

    CoreSegment::CoreSegment(const std::string& name, const MemoryID& parentID, aron::typenavigator::AronObjectTypeNavigatorPtr aronType) :
        CoreSegment(parentID.getMemoryID().withCoreSegmentName(name), aronType)
    {
    }

    CoreSegment::CoreSegment(const MemoryID& id, aron::typenavigator::AronObjectTypeNavigatorPtr aronType) :
        Base(id, aronType)
    {
    }

    CoreSegment::CoreSegment(const CoreSegment& other) : Base(other)
    {
        *this = other;
    }

    CoreSegment& CoreSegment::operator=(const CoreSegment& other)
    {
        other._copySelf(*this);
        other._copyElements(*this, true);
        return *this;
    }

    bool CoreSegment::equalsDeep(const CoreSegment& other) const
    {
        if (size() != other.size())
        {
            return false;
        }
        for (const auto& [key, provider] : providerSegments)
        {
            if (not other.hasProviderSegment(key))
            {
                return false;
            }
            if (not provider.equalsDeep(other.getProviderSegment(key)))
            {
                return false;
            }
        }
        return true;
    }

    CoreSegment::~CoreSegment()
    {
    }

    const std::string& CoreSegment::name() const
    {
        return id().coreSegmentName;
    }

    void CoreSegment::_copyElements(CoreSegment& other, bool withData) const
    {
        other.providerSegments.clear();
        for (const auto& [name, providerSegment] : this->providerSegments)
        {
            other.addProviderSegment(providerSegment.copy(withData));
        }
    }


    bool CoreSegment::hasProviderSegment(const std::string& name) const
    {
        return providerSegments.count(name) > 0;
    }

    ProviderSegment& CoreSegment::getProviderSegment(const std::string& name)
    {
        return const_cast<ProviderSegment&>(const_cast<const CoreSegment*>(this)->getProviderSegment(name));
    }

    const ProviderSegment& CoreSegment::getProviderSegment(const std::string& name) const
    {
        auto it = this->providerSegments.find(name);
        if (it != providerSegments.end())
        {
            return it->second;
        }
        else
        {
            throw armem::error::MissingEntry("provider segment", name, getLevelName(), this->name());
        }
    }

    const Entity& CoreSegment::getEntity(const MemoryID& id) const
    {
        _checkContainerName(id.coreSegmentName, this->name());
        return getProviderSegment(id.providerSegmentName).getEntity(id);
    }


    ProviderSegment& CoreSegment::addProviderSegment(const std::string& name, aron::typenavigator::AronObjectTypeNavigatorPtr providerSegmentType)
    {
        aron::typenavigator::AronObjectTypeNavigatorPtr type = providerSegmentType ? providerSegmentType : this->aronType();
        return addProviderSegment(ProviderSegment(name, type));
    }

    ProviderSegment& CoreSegment::addProviderSegment(const ProviderSegment& providerSegment)
    {
        return addProviderSegment(ProviderSegment(providerSegment));
    }

    ProviderSegment& CoreSegment::addProviderSegment(ProviderSegment&& providerSegment)
    {
        if (hasProviderSegment(providerSegment.name()))
        {
            throw armem::error::ContainerEntryAlreadyExists(
                providerSegment.getLevelName(), providerSegment.name(), getLevelName(), this->name());
        }

        auto it = providerSegments.emplace(providerSegment.name(), std::move(providerSegment)).first;
        it->second.id().setCoreSegmentID(id());
        it->second.setMaxHistorySize(maxHistorySize);
        return it->second;
    }

    void CoreSegment::setMaxHistorySize(long maxSize)
    {
        MaxHistorySize::setMaxHistorySize(maxSize);
        for (auto& [name, seg] : providerSegments)
        {
            seg.setMaxHistorySize(maxSize);
        }
    }

    std::string CoreSegment::getLevelName() const
    {
        return "core segment";
    }

    CoreSegment::ContainerT& CoreSegment::_underlyingContainer()
    {
        return providerSegments;
    }


    MemoryID CoreSegment::update(const EntityUpdate& update)
    {
        _checkContainerName(update.entityID.coreSegmentName, this->name());

        auto it = this->providerSegments.find(update.entityID.providerSegmentName);
        if (it != providerSegments.end())
        {
            return it->second.update(update);
        }
        else
        {
            throw armem::error::MissingEntry("provider segment", update.entityID.providerSegmentName,
                                             getLevelName(), name());
        }
    }

}
