#include "Memory.h"

#include <ArmarXCore/core/logging/Logging.h>
#include <ArmarXCore/core/exceptions/local/ExpressionException.h>

#include "error.h"


namespace armarx::armem
{

    Memory::Memory()
    {
    }

    Memory::Memory(const std::string& name) : Memory(MemoryID().withMemoryName(name))
    {
    }

    Memory::Memory(const MemoryID& id) : EntityContainer(id)
    {
    }

    Memory::Memory(const Memory& other) : EntityContainer(other)
    {
        *this = other;
    }

    Memory& Memory::operator=(const Memory& other)
    {
        other._copySelf(*this);
        other._copyElements(*this, true);
        return *this;
    }

    Memory::~Memory()
    {
    }


    const std::string& Memory::name() const
    {
        return id().memoryName;
    }

    void Memory::_copyElements(Memory& other, bool withData) const
    {
        other.coreSegments.clear();
        for (const auto& [name, coreSegment] : this->coreSegments)
        {
            other.coreSegments.emplace(name, coreSegment.copy(withData));
        }
    }


    bool Memory::hasCoreSegment(const std::string& name) const
    {
        return coreSegments.count(name) > 0;
    }

    CoreSegment& Memory::getCoreSegment(const std::string& name)
    {
        return const_cast<CoreSegment&>(const_cast<const Memory*>(this)->getCoreSegment(name));
    }
    const CoreSegment& Memory::getCoreSegment(const std::string& name) const
    {
        auto it = this->coreSegments.find(name);
        if (it != coreSegments.end())
        {
            return it->second;
        }
        else
        {
            throw armem::error::MissingEntry(CoreSegment().getLevelName(), name, getLevelName(), this->name());
        }
    }


    const Entity& Memory::getEntity(const MemoryID& id) const
    {
        _checkContainerName(id.memoryName, this->name());
        return getCoreSegment(id.coreSegmentName).getEntity(id);
    }


    CoreSegment& Memory::addCoreSegment(const std::string& name, aron::typenavigator::AronObjectTypeNavigatorPtr coreSegmentType)
    {
        return addCoreSegment(CoreSegment(name, coreSegmentType));
    }

    CoreSegment& Memory::addCoreSegment(const CoreSegment& coreSegment)
    {
        return addCoreSegment(CoreSegment(coreSegment));
    }

    CoreSegment& Memory::addCoreSegment(CoreSegment&& coreSegment)
    {
        if (coreSegments.count(coreSegment.name()) > 0)
        {
            throw armem::error::ContainerEntryAlreadyExists(coreSegment.getLevelName(), coreSegment.name(),
                    this->getLevelName(), this->name());
        }
        auto it = coreSegments.emplace(coreSegment.name(), std::move(coreSegment)).first;
        it->second.id().setMemoryID(id());
        return it->second;
    }

    std::vector<CoreSegment*> Memory::addCoreSegments(const std::vector<std::string>& names)
    {
        std::vector<CoreSegment*> segments;
        for (const auto& name : names)
        {
            try
            {
                segments.push_back(&addCoreSegment(name));
            }
            catch (const armem::error::ContainerEntryAlreadyExists& e)
            {
                ARMARX_INFO << e.what() << "\nIgnoring multiple addition.";
            }
        }
        return segments;
    }


    MemoryID Memory::update(const EntityUpdate& update)
    {
        _checkContainerName(update.entityID.memoryName, this->name());

        auto it = this->coreSegments.find(update.entityID.coreSegmentName);
        if (it != coreSegments.end())
        {
            return it->second.update(update);
        }
        else
        {
            throw armem::error::MissingEntry(CoreSegment().getLevelName(), update.entityID.coreSegmentName,
                                             getLevelName(), this->name());
        }
    }

    std::string Memory::getLevelName() const
    {
        return "memory";
    }

    Memory::ContainerT& Memory::_underlyingContainer()
    {
        return coreSegments;
    }


    bool Memory::equalsDeep(const Memory& other) const
    {
        if (size() != other.size())
        {
            return false;
        }
        for (const auto& [key, core] : coreSegments)
        {
            if (not other.hasCoreSegment(key))
            {
                return false;
            }
            if (not core.equalsDeep(other.getCoreSegment(key)))
            {
                return false;
            }
        }
        return true;
    }

}
