#include "ProviderSegment.h"

#include <ArmarXCore/core/exceptions/local/ExpressionException.h>

#include "error.h"


namespace armarx::armem
{

    ProviderSegment::ProviderSegment()
    {
    }

    ProviderSegment::ProviderSegment(const std::string& name, aron::typenavigator::AronObjectTypeNavigatorPtr aronType) :
        ProviderSegment(name, MemoryID(), aronType)
    {
    }

    ProviderSegment::ProviderSegment(const std::string& name, const MemoryID parentID, aron::typenavigator::AronObjectTypeNavigatorPtr aronType) :
        ProviderSegment(parentID.withProviderSegmentName(name), aronType)
    {
    }

    ProviderSegment::ProviderSegment(const MemoryID id, aron::typenavigator::AronObjectTypeNavigatorPtr aronType) :
        Base(id, aronType)
    {
    }

    ProviderSegment::ProviderSegment(const ProviderSegment& other) : Base(other)
    {
        *this = other;
    }

    ProviderSegment& ProviderSegment::operator=(const ProviderSegment& other)
    {
        other._copySelf(*this);
        other._copyElements(*this, true);
        return *this;
    }

    bool ProviderSegment::equalsDeep(const ProviderSegment& other) const
    {
        if (size() != other.size())
        {
            return false;
        }
        for (const auto& [key, value] : entities)
        {
            if (not other.hasEntity(key))
            {
                return false;
            }

            if (not value.equalsDeep(other.getEntity(key)))
            {
                return false;
            }
        }
        return true;
    }

    ProviderSegment::~ProviderSegment()
    {
    }

    const std::string& ProviderSegment::name() const
    {
        return id().providerSegmentName;
    }


    void ProviderSegment::_copySelf(ProviderSegment& other) const
    {
        Base::_copySelf(other);
        other.maxHistorySize = this->maxHistorySize;
    }

    void ProviderSegment::_copyElements(ProviderSegment& other, bool withData) const
    {
        other.entities.clear();
        for (const auto& [name, entity] : this->entities)
        {
            other.addEntity(entity.copy(withData));
        }
    }


    bool ProviderSegment::hasEntity(const std::string& name) const
    {
        return entities.count(name) > 0;
    }

    Entity& ProviderSegment::getEntity(const std::string& name)
    {
        return const_cast<Entity&>(const_cast<const ProviderSegment*>(this)->getEntity(name));
    }

    const Entity& ProviderSegment::getEntity(const std::string& name) const
    {
        auto it = this->entities.find(name);
        if (it != entities.end())
        {
            return it->second;
        }
        else
        {
            throw error::MissingEntry("entity", name, getLevelName(), this->name());
        }
    }

    const Entity& ProviderSegment::getEntity(const MemoryID& id) const
    {
        _checkContainerName(id.providerSegmentName, this->name());
        return getEntity(id.entityName);
    }

    MemoryID ProviderSegment::update(const EntityUpdate& update)
    {
        _checkContainerName(update.entityID.providerSegmentName, this->name());

        Entity* entity;
        auto it = this->entities.find(update.entityID.providerSegmentName);
        if (it == entities.end())
        {
            // Add entity entry.
            entity = &addEntity(update.entityID.entityName);
            entity->setMaxHistorySize(maxHistorySize);
        }
        else
        {
            entity = &it->second;
        }
        // Update entity.
        return entity->update(update);
    }

    Entity& ProviderSegment::addEntity(const std::string& name)
    {
        return addEntity(Entity(name));
    }

    Entity& ProviderSegment::addEntity(const Entity& entity)
    {
        return addEntity(Entity(entity));
    }

    Entity& ProviderSegment::addEntity(Entity&& entity)
    {
        auto it = entities.emplace(entity.name(), std::move(entity)).first;
        it->second.id().setProviderSegmentID(id());
        return it->second;
    }

    void ProviderSegment::setMaxHistorySize(long maxSize)
    {
        MaxHistorySize::setMaxHistorySize(maxSize);
        for (auto& [name, entity] : entities)
        {
            entity.setMaxHistorySize(maxSize);
        }
    }

    std::string ProviderSegment::getLevelName() const
    {
        return "provider segment";
    }

    ProviderSegment::ContainerT& ProviderSegment::_underlyingContainer()
    {
        return entities;
    }

}
