/*
* This file is part of ArmarX.
*
* ArmarX is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License version 2 as
* published by the Free Software Foundation.
*
* ArmarX is distributed in the hope that it will be useful, but
* WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program. If not, see <http://www.gnu.org/licenses/>.
*
* @author     Fabian Peller (fabian dot peller at kit dot edu)
* @copyright  http://www.gnu.org/licenses/gpl-2.0.txt
*             GNU General Public License
*/

#include "DiskReader.h"

#include <sstream>
#include <fstream>
#include <iostream>

// ArmarX
#include <RobotAPI/interface/aron.h>
#include <RobotAPI/libraries/aron/aroncore/navigators/datanavigator/AronAllDataNavigators.h>
#include <RobotAPI/libraries/aron/aroncore/navigators/typenavigator/AronAllTypeNavigators.h>

#include <RobotAPI/libraries/aron/aroncore/io/AronDataIO/AronDataConverter/AronDataConverter.h>
#include <RobotAPI/libraries/aron/aroncore/io/AronDataIO/classWriters/NavigatorWriter/AronDataNavigatorWriter.h>

#include <RobotAPI/libraries/aron/aroncore/io/AronTypeIO/AronTypeConverter/AronTypeConverter.h>
#include <RobotAPI/libraries/aron/aroncore/io/AronTypeIO/classWriters/NavigatorWriter/AronTypeNavigatorWriter.h>

#include <RobotAPI/libraries/armem/core/error.h>


namespace armarx::armem::io
{

    FileSystemLookupMemory DiskReader::readMemoryStructureFromDisk()
    {
        FileSystemLookupMemory ret;

        for (const auto& memoryFolder : std::filesystem::directory_iterator(rootPath))
        {
            if (!std::filesystem::is_directory(memoryFolder))
            {
                continue;
            }

            std::string folderName = memoryFolder.path().filename();
            FileSystemLookupMemory tmp = readMemoryStructureFromDisk(folderName);
            ret.merge(tmp);
        }
        return ret;
    }

    FileSystemLookupMemory DiskReader::readMemoryStructureFromDisk(const std::string& memoryFolderName)
    {
        FileSystemLookupMemory ret;

        std::filesystem::path p = rootPath / memoryFolderName;
        if (!std::filesystem::is_directory(p))
        {
            return ret;
        }

        for (const auto& coreSegmentFolder : std::filesystem::directory_iterator(p))
        {
            if (!std::filesystem::is_directory(coreSegmentFolder))
            {
                continue;
            }
            std::string coreSegmentName = coreSegmentFolder.path().filename();
            for (const auto& providerSegmentFolder : std::filesystem::directory_iterator(coreSegmentFolder.path()))
            {
                if (!std::filesystem::is_directory(providerSegmentFolder))
                {
                    continue;
                }
                std::string providerSegmentName = providerSegmentFolder.path().filename();
                for (const auto& entityFolder : std::filesystem::directory_iterator(providerSegmentFolder.path()))
                {
                    if (!std::filesystem::is_directory(entityFolder))
                    {
                        continue;
                    }
                    std::string entityName = entityFolder.path().filename();
                    for (const auto& entityHistoryFolder : std::filesystem::directory_iterator(entityFolder.path()))
                    {
                        if (!std::filesystem::is_directory(entityHistoryFolder))
                        {
                            continue;
                        }
                        Time entityTimestamp = armem::timeFromStringMicroSeconds(entityHistoryFolder.path().filename());

                        // TODO: make this relative to amount of files in folder
                        for (unsigned int i = 0; i < 10000; ++i)
                        {
                            std::filesystem::path entityInstance = entityHistoryFolder / std::filesystem::path(std::to_string(i) + getEntityInstanceSuffix());
                            if (!std::filesystem::is_regular_file(entityInstance))
                            {
                                break;
                            }
                            DiskMemory& mem = ret.memoryLookupTable[memoryFolderName];
                            DiskCoreSegment& mem_core = mem[coreSegmentName];
                            DiskProviderSegment& mem_prov = mem_core[providerSegmentName];
                            DiskEntity& mem_entity = mem_prov[entityName];
                            DiskEntitySnapshot& mem_entity_history = mem_entity[entityTimestamp];

                            mem_entity_history.push_back(entityInstance);
                        }
                    }
                }
            }
        }
        return ret;
    }

    EntityInstance DiskReader::readSingleInstanceFromDisk(const MemoryID& id, const aron::typenavigator::AronTypeNavigatorPtr& expectedStructure) const
    {
        std::filesystem::path p = rootPath / (id.str() + getEntityInstanceSuffix());
        return readSingleInstanceFromDisk(p, expectedStructure);
    }

    EntityInstance DiskReader::readSingleInstanceFromDisk(const std::filesystem::path& p, const aron::typenavigator::AronTypeNavigatorPtr& expectedStructure) const
    {
        if (!std::filesystem::is_regular_file(p))
        {
            throw error::PathNotARegularFile(p.string(), "During readSingleInstanceFromDisk tried to read an entityInstance from a non-file path.");
        }

        std::ifstream ifs(p);
        std::string file_content((std::istreambuf_iterator<char>(ifs)), (std::istreambuf_iterator<char>()));

        aron::datanavigator::AronDictDataNavigatorPtr dictdata = getStringAsDataNavigator(file_content, expectedStructure);
        return unwrapData(dictdata);
    }

    aron::typenavigator::AronObjectTypeNavigatorPtr DiskReader::readSingleTypeInformationFromDisk(const MemoryID& id) const
    {
        std::filesystem::path p = rootPath / id.str() / (id.getLeafItem() + getTypeSuffix());
        return readSingleTypeInformationFromDisk(p);
    }

    aron::typenavigator::AronObjectTypeNavigatorPtr DiskReader::readSingleTypeInformationFromDisk(const std::filesystem::path& p) const
    {
        if (!std::filesystem::is_regular_file(p))
        {
            return nullptr;
        }

        std::ifstream ifs(p);
        std::string file_content((std::istreambuf_iterator<char>(ifs)), (std::istreambuf_iterator<char>()));

        aron::typenavigator::AronObjectTypeNavigatorPtr objecttype = getStringAsTypeNavigator(file_content);
        return objecttype;
    }

    EntityInstance DiskReader::unwrapData(const aron::datanavigator::AronDictDataNavigatorPtr& dataWrapped) const
    {
        EntityInstance e;
        EntityInstanceMetadata& metadata = e.metadata();

        aron::datanavigator::AronDictDataNavigatorPtr data = aron::datanavigator::AronDictDataNavigator::DynamicCastAndCheck(dataWrapped->getElement(DISK_READER_WRITER_DATA_FIELD));
        e.setData(data);

        // not used right now
        //aron::datanavigator::AronLongDataNavigatorPtr timeWrapped = aron::datanavigator::AronLongDataNavigator::DynamicCastAndCheck(dataWrapped->getElement(DISK_READER_WRITER_TIME_WRAPPED_FIELD));

        aron::datanavigator::AronLongDataNavigatorPtr timeCreated = aron::datanavigator::AronLongDataNavigator::DynamicCastAndCheck(dataWrapped->getElement(DISK_READER_WRITER_TIME_CREATED_FIELD));
        metadata.timeCreated = Time::microSeconds(timeCreated->toAronLongPtr()->value);

        aron::datanavigator::AronLongDataNavigatorPtr timeSent = aron::datanavigator::AronLongDataNavigator::DynamicCastAndCheck(dataWrapped->getElement(DISK_READER_WRITER_TIME_SENT_FIELD));
        metadata.timeSent = Time::microSeconds(timeSent->toAronLongPtr()->value);

        aron::datanavigator::AronLongDataNavigatorPtr timeArrived = aron::datanavigator::AronLongDataNavigator::DynamicCastAndCheck(dataWrapped->getElement(DISK_READER_WRITER_TIME_ARRIVED_FIELD));
        metadata.timeArrived = Time::microSeconds(timeArrived->toAronLongPtr()->value);

        aron::datanavigator::AronDoubleDataNavigatorPtr confidence = aron::datanavigator::AronDoubleDataNavigator::DynamicCastAndCheck(dataWrapped->getElement(DISK_READER_WRITER_CONFIDENCE_FIELD));
        metadata.confidence = static_cast<float>(confidence->toAronDoublePtr()->value);

        return e;
    }

    aron::typenavigator::AronObjectTypeNavigatorPtr DiskReader::unwrapType(const aron::typenavigator::AronObjectTypeNavigatorPtr& t) const
    {
        return aron::typenavigator::AronObjectTypeNavigator::DynamicCastAndCheck(t->getAcceptedType(DISK_READER_WRITER_DATA_FIELD));
    }
}
