/*
* This file is part of ArmarX.
*
* ArmarX is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License version 2 as
* published by the Free Software Foundation.
*
* ArmarX is distributed in the hope that it will be useful, but
* WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program. If not, see <http://www.gnu.org/licenses/>.
*
* @author     Fabian Peller (fabian dot peller at kit dot edu)
* @copyright  http://www.gnu.org/licenses/gpl-2.0.txt
*             GNU General Public License
*/

// STD/STL
#include "DiskWriter.h"

#include <fstream>
#include <iostream>

// ArmarX
#include <RobotAPI/interface/aron.h>
#include <RobotAPI/libraries/aron/aroncore/navigators/datanavigator/AronAllDataNavigators.h>
#include <RobotAPI/libraries/aron/aroncore/navigators/typenavigator/AronAllTypeNavigators.h>

#include <RobotAPI/libraries/aron/aroncore/io/AronDataIO/AronDataWriterVisitor/AronDataWriterVisitor.h>
#include <RobotAPI/libraries/aron/aroncore/io/AronDataIO/AronDataWriter.h>
#include <RobotAPI/libraries/aron/aroncore/io/AronTypeIO/AronTypeWriterVisitor/AronTypeWriterVisitor.h>
#include <RobotAPI/libraries/aron/aroncore/io/AronTypeIO/AronTypeWriter.h>


// BaseClass
#include "../DiskReaderWriter.h"
#include "../FileSystemLookupMemory.h"

#include <RobotAPI/libraries/armem/core/Memory.h>
#include <RobotAPI/libraries/armem/core/CoreSegment.h>
#include <RobotAPI/libraries/armem/core/Entity.h>


namespace armarx::armem::io
{

    DiskWriterReturnInformation DiskWriter::writeOnDisk(const Memory& m)
    {
        MemoryID id = m.id().getMemoryID();
        DiskWriterReturnInformation ret(id.memoryName);

        std::filesystem::path p = rootPath / id.str();
        if (!ensureDirectoryPathExists(p))
        {
            ret.hasError = true;
            return ret;
        }

        for (const auto& [coreKey, coreSegment] : m.coreSegments)
        {
            MemoryID storeCore(id);
            storeCore.coreSegmentName = coreKey;
            DiskWriterReturnInformation other = writeOnDisk(storeCore, coreSegment);
            ret.append(other);
        }
        return ret;
    }

    DiskWriterReturnInformation DiskWriter::writeOnDisk(const MemoryID& i, const CoreSegment& s)
    {
        MemoryID id = i.getCoreSegmentID();
        DiskWriterReturnInformation ret(id.memoryName);

        std::filesystem::path p = rootPath / id.str();
        if (!ensureDirectoryPathExists(p))
        {
            ret.coreSegmentsError.push_back(id.str());
            ret.hasError = true;
            return ret;
        }

        if (s.hasAronType())
        {
            std::string val = getTypeAsString(wrapType(s.aronType()));

            std::filesystem::path coreSegmentProviderPath = rootPath / id.str() / (id.coreSegmentName + getTypeSuffix());
            std::ofstream ofs;
            ofs.open(coreSegmentProviderPath);
            ofs << val;
            ofs.close();
        }

        for (const auto& [providerKey, providerSegment] : s.providerSegments)
        {
            MemoryID storeProvider(id);
            storeProvider.providerSegmentName = providerKey;
            DiskWriterReturnInformation other = writeOnDisk(storeProvider, providerSegment);
            ret.append(other);
        }
        return ret;
    }

    DiskWriterReturnInformation DiskWriter::writeOnDisk(const MemoryID& i, const ProviderSegment& s)
    {
        MemoryID id = i.getProviderSegmentID();
        DiskWriterReturnInformation ret(id.memoryName);

        std::filesystem::path p = rootPath / id.str();
        if (!ensureDirectoryPathExists(p))
        {
            ret.providerSegmentsError.push_back(id.str());
            ret.hasError = true;
            return ret;
        }

        if (s.hasAronType())
        {
            std::string val = getTypeAsString(wrapType(s.aronType()));

            std::filesystem::path providerSegmentProviderPath = rootPath / id.str() / (id.providerSegmentName + getTypeSuffix());
            std::ofstream ofs;
            ofs.open(providerSegmentProviderPath);
            ofs << val;
            ofs.close();
        }

        for (const auto& [entityKey, entity] : s.entities)
        {
            MemoryID storeEntity(id);
            storeEntity.entityName = entityKey;
            DiskWriterReturnInformation other = writeOnDisk(storeEntity, entity);
            ret.append(other);
        }
        return ret;
    }

    DiskWriterReturnInformation DiskWriter::writeOnDisk(const MemoryID& i, const Entity& s)
    {
        MemoryID id = i.getEntityID();
        DiskWriterReturnInformation ret(id.memoryName);

        std::filesystem::path p = rootPath / id.str();
        if (!ensureDirectoryPathExists(p))
        {
            ret.entitiesError.push_back(id.str());
            ret.hasError = true;
            return ret;
        }

        for (const auto& [ts, snapshot] : s.history)
        {
            MemoryID storeSnapshot(id);
            storeSnapshot.timestamp = ts;
            DiskWriterReturnInformation other = writeOnDisk(storeSnapshot, snapshot);
            ret.append(other);
        }
        return ret;
    }

    DiskWriterReturnInformation DiskWriter::writeOnDisk(const MemoryID& _id, const EntitySnapshot& s)
    {
        MemoryID id = _id.getEntitySnapshotID();
        DiskWriterReturnInformation ret(id.memoryName);

        std::filesystem::path p = rootPath / id.str();
        if (!ensureDirectoryPathExists(p))
        {
            ret.historyTimestampsError.push_back(id.str());
            ret.hasError = true;
            return ret;
        }

        int i = 0;
        for (const EntityInstance& instance : s.instances)
        {
            MemoryID storeInstance(id);
            storeInstance.instanceIndex = i;

            std::filesystem::path entityElementPath = rootPath / (storeInstance.str() + getEntityInstanceSuffix());
            std::string val = getDataAsString(wrapData(instance));

            if (filePathExists(entityElementPath))
            {
                if (std::filesystem::is_directory(entityElementPath))
                {
                    ret.entityInstancesError.push_back(id.str());
                    ret.hasError = true;
                    return ret;
                }
                //std::ifstream ifs(entityElementPath);
                //std::string file_content((std::istreambuf_iterator<char>(ifs)), (std::istreambuf_iterator<char>()));
                //if (file_content == val)
                //{
                //    // already written. skip
                //    continue;
                //}
            }

            std::ofstream ofs;
            ofs.open(entityElementPath);
            ofs << val;
            ofs.close();
            ret.storedElements.memoryLookupTable[id.memoryName]
            [id.coreSegmentName]
            [id.providerSegmentName]
            [id.entityName]
            [id.timestamp].push_back(entityElementPath);

            i++;
        }
        return ret;
    }

    aron::datanavigator::AronDictDataNavigatorPtr DiskWriter::wrapData(const EntityInstance& e) const
    {
        aron::datanavigator::AronDictDataNavigatorPtr dataWrapped(new aron::datanavigator::AronDictDataNavigator());
        dataWrapped->addElement(DISK_READER_WRITER_DATA_FIELD, e.data());

        aron::datanavigator::AronLongDataNavigatorPtr timeWrapped(new aron::datanavigator::AronLongDataNavigator());
        timeWrapped->setValue(Time::now().toMicroSeconds());
        dataWrapped->addElement(DISK_READER_WRITER_TIME_STORED_FIELD, timeWrapped);


        const EntityInstanceMetadata& metadata = e.metadata();
        aron::datanavigator::AronLongDataNavigatorPtr timeCreated(new aron::datanavigator::AronLongDataNavigator());
        timeCreated->setValue(metadata.timeCreated.toMicroSeconds());
        dataWrapped->addElement(DISK_READER_WRITER_TIME_CREATED_FIELD, timeCreated);

        aron::datanavigator::AronLongDataNavigatorPtr timeSent(new aron::datanavigator::AronLongDataNavigator());
        timeSent->setValue(metadata.timeSent.toMicroSeconds());
        dataWrapped->addElement(DISK_READER_WRITER_TIME_SENT_FIELD, timeSent);

        aron::datanavigator::AronLongDataNavigatorPtr timeArrived(new aron::datanavigator::AronLongDataNavigator());
        timeArrived->setValue(metadata.timeArrived.toMicroSeconds());
        dataWrapped->addElement(DISK_READER_WRITER_TIME_ARRIVED_FIELD, timeArrived);

        aron::datanavigator::AronDoubleDataNavigatorPtr confidence(new aron::datanavigator::AronDoubleDataNavigator());
        confidence->setValue(metadata.confidence);
        dataWrapped->addElement(DISK_READER_WRITER_CONFIDENCE_FIELD, confidence);

        return dataWrapped;
    }

    aron::typenavigator::AronObjectTypeNavigatorPtr DiskWriter::wrapType(const aron::typenavigator::AronObjectTypeNavigatorPtr& t) const
    {
        aron::typenavigator::AronObjectTypeNavigatorPtr typeWrapped(new aron::typenavigator::AronObjectTypeNavigator());
        typeWrapped->setObjectName(t->getObjectName() + "__ltm_export");
        typeWrapped->addAcceptedType(DISK_READER_WRITER_DATA_FIELD, t);

        typeWrapped->addAcceptedType(DISK_READER_WRITER_TIME_STORED_FIELD, aron::typenavigator::AronLongTypeNavigatorPtr(new aron::typenavigator::AronLongTypeNavigator()));
        typeWrapped->addAcceptedType(DISK_READER_WRITER_TIME_CREATED_FIELD, aron::typenavigator::AronLongTypeNavigatorPtr(new aron::typenavigator::AronLongTypeNavigator()));
        typeWrapped->addAcceptedType(DISK_READER_WRITER_TIME_SENT_FIELD, aron::typenavigator::AronLongTypeNavigatorPtr(new aron::typenavigator::AronLongTypeNavigator()));
        typeWrapped->addAcceptedType(DISK_READER_WRITER_TIME_ARRIVED_FIELD, aron::typenavigator::AronLongTypeNavigatorPtr(new aron::typenavigator::AronLongTypeNavigator()));
        typeWrapped->addAcceptedType(DISK_READER_WRITER_CONFIDENCE_FIELD, aron::typenavigator::AronDoubleTypeNavigatorPtr(new aron::typenavigator::AronDoubleTypeNavigator()));

        return typeWrapped;
    }

    bool DiskWriter::directoryPathExists(const std::filesystem::path& p) const
    {
        return std::filesystem::exists(p) && std::filesystem::is_directory(p);
    }

    bool DiskWriter::ensureDirectoryPathExists(const std::filesystem::path& p) const
    {
        if (!std::filesystem::exists(p))
        {
            return std::filesystem::create_directory(p);
        }
        return directoryPathExists(p);
    }

    bool DiskWriter::filePathExists(const std::filesystem::path& p) const
    {
        return std::filesystem::exists(p) && std::filesystem::is_regular_file(p);
    }


    io::DiskWriterReturnInformation::DiskWriterReturnInformation(const std::string& m) :
        memoryName(m)
    {
    }

    void io::DiskWriterReturnInformation::append(const io::DiskWriterReturnInformation& o)
    {
        if (memoryName != o.memoryName)
        {
            // The memory name must be equal
            return;
        }

        storedElements.merge(o.storedElements);

        coreSegmentsError.insert(coreSegmentsError.end(), o.coreSegmentsError.begin(), o.coreSegmentsError.end());
        providerSegmentsError.insert(providerSegmentsError.end(), o.providerSegmentsError.begin(), o.providerSegmentsError.end());
        entitiesError.insert(entitiesError.end(), o.entitiesError.begin(), o.entitiesError.end());
        historyTimestampsError.insert(historyTimestampsError.end(), o.historyTimestampsError.begin(), o.historyTimestampsError.end());
        entityInstancesError.insert(entityInstancesError.end(), o.entityInstancesError.begin(), o.entityInstancesError.end());
    }
}
