/*
* This file is part of ArmarX.
*
* ArmarX is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License version 2 as
* published by the Free Software Foundation.
*
* ArmarX is distributed in the hope that it will be useful, but
* WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program. If not, see <http://www.gnu.org/licenses/>.
*
* @author     Fabian Peller (fabian dot peller at kit dot edu)
* @copyright  http://www.gnu.org/licenses/gpl-2.0.txt
*             GNU General Public License
*/

#include "FileSystemLookupMemory.h"


namespace armarx::armem::io
{

    DiskMemory FileSystemLookupMemory::getDiskMemory(const std::string& n) const
    {
        if (!hasMemory(n))
        {
            throw error::ArMemError("Could not find a memory with name '" + n + "' in lookup memory.");
        }
        return memoryLookupTable.at(n);
    }

    DiskMemory FileSystemLookupMemory::getDiskMemory(const MemoryID& id) const
    {
        if (!hasMemory(id))
        {
            throw error::InvalidMemoryID(id, "MemoryName does not exist in lookup memory.");
        }
        return memoryLookupTable.at(id.memoryName);
    }

    DiskCoreSegment FileSystemLookupMemory::getDiskCoreSegment(const MemoryID& id) const
    {
        if (!hasCoreSegment(id))
        {
            throw error::InvalidMemoryID(id, "CoreSegmentName does not exist in lookup memory.");
        }
        return getDiskMemory(id).at(id.coreSegmentName);
    }

    DiskProviderSegment FileSystemLookupMemory::getDiskProviderSegment(const MemoryID& id) const
    {
        if (!hasProviderSegment(id))
        {
            throw error::InvalidMemoryID(id, "ProviderSegmentName does not exist in lookup memory.");
        }
        return getDiskCoreSegment(id).at(id.providerSegmentName);
    }

    DiskEntity FileSystemLookupMemory::getDiskEntity(const MemoryID& id) const
    {
        if (!hasEntity(id))
        {
            throw error::InvalidMemoryID(id, "EntityName does not exist in lookup memory.");
        }
        return getDiskProviderSegment(id).at(id.entityName);
    }

    DiskEntitySnapshot FileSystemLookupMemory::getDiskEntitySnapshot(const MemoryID& id) const
    {
        if (!hasEntitySnapshot(id))
        {
            throw error::InvalidMemoryID(id, "EntitySnapshotTimestamp does not exist in lookup memory.");
        }
        return getDiskEntity(id).at(id.timestamp);
    }

    DiskEntityInstance FileSystemLookupMemory::getDiskEntityInstance(const MemoryID& id) const
    {
        if (!hasEntityInstance(id))
        {
            throw error::InvalidMemoryID(id, "EntityInstanceIndex does not exist in lookup memory.");
        }
        return getDiskEntitySnapshot(id).at(id.instanceIndex);
    }


    bool FileSystemLookupMemory::hasMemory(const std::string& n) const
    {
        MemoryID id;
        id.memoryName = n;
        return hasMemory(id);
    }

    bool FileSystemLookupMemory::hasMemory(const MemoryID& id) const
    {
        return id.hasMemoryName() && memoryLookupTable.find(id.memoryName) != memoryLookupTable.end();
    }

    bool FileSystemLookupMemory::hasCoreSegment(const MemoryID& id) const
    {
        if (hasMemory(id))
        {
            const auto& m = getDiskMemory(id);
            return id.hasCoreSegmentName() && m.find(id.coreSegmentName) != m.end();
        }
        return false;
    }

    bool FileSystemLookupMemory::hasProviderSegment(const MemoryID& id) const
    {
        if (hasCoreSegment(id))
        {
            const auto& m = getDiskCoreSegment(id);
            return id.hasProviderSegmentName() && m.find(id.providerSegmentName) != m.end();
        }
        return false;
    }

    bool FileSystemLookupMemory::hasEntity(const MemoryID& id) const
    {
        if (hasProviderSegment(id))
        {
            const auto& m = getDiskProviderSegment(id);
            return id.hasEntityName() && m.find(id.entityName) != m.end();
        }
        return false;
    }

    bool FileSystemLookupMemory::hasEntitySnapshot(const MemoryID& id) const
    {
        if (hasEntity(id))
        {
            const auto& m = getDiskEntity(id);
            return id.hasTimestamp() && m.find(id.timestamp) != m.end();
        }
        return false;
    }

    bool FileSystemLookupMemory::hasEntityInstance(const MemoryID& id) const
    {
        if (hasEntitySnapshot(id))
        {
            const auto& m = getDiskEntitySnapshot(id);
            return id.hasInstanceIndex() && m.size() >= (size_t) id.instanceIndex;
        }
        return false;
    }


    void FileSystemLookupMemory::merge(const FileSystemLookupMemory& m)
    {
        for (const auto& [memoryName, memory] : m.memoryLookupTable)
        {
            if (memoryLookupTable.find(memoryName) == memoryLookupTable.end())
            {
                memoryLookupTable[memoryName] = memory;
                continue;
            }

            DiskMemory& lutMemory = memoryLookupTable[memoryName];
            for (const auto& [coreKey, coreSegment] : memory)
            {
                if (lutMemory.find(coreKey) == lutMemory.end())
                {
                    lutMemory[coreKey] = coreSegment;
                    continue;
                }

                DiskCoreSegment& lutCoreSegment = lutMemory[coreKey];
                for (const auto& [providerKey, providerSegment] : coreSegment)
                {
                    if (lutCoreSegment.find(providerKey) == lutCoreSegment.end())
                    {
                        lutCoreSegment[providerKey] = providerSegment;
                        continue;
                    }

                    DiskProviderSegment& lutProviderSegment = lutCoreSegment[providerKey];
                    for (const auto& [entityKey, entity] : providerSegment)
                    {
                        if (lutProviderSegment.find(entityKey) == lutProviderSegment.end())
                        {
                            lutProviderSegment[entityKey] = entity;
                            continue;
                        }

                        DiskEntity& lutEntity = lutProviderSegment[entityKey];
                        for (const auto& [entityHistoryTimestamp, entitySnapshot] : entity)
                        {
                            if (lutEntity.find(entityHistoryTimestamp) == lutEntity.end())
                            {
                                lutEntity[entityHistoryTimestamp] = entitySnapshot;
                                continue;
                            }

                            DiskEntitySnapshot& lutEntitySnapshot = lutEntity[entityHistoryTimestamp];
                            //for (unsigned int i = 0; i < entitySnapshot.size(); ++i)
                            //{
                            //    // clear internal stored memory of that snapshot
                            //    internalMemoryEntitySnapshot.clear();
                            //    internalMemoryEntitySnapshot.push_back(entitySnapshot[i]);
                            //}
                            // for now, simply overwrite all values since one history timestamp should not have two different snapshots
                            lutEntitySnapshot = entitySnapshot;
                        }
                    }
                }
            }
        }
    }


}
