#include "MemoryRemoteGui.h"

#include <ArmarXCore/core/exceptions/local/ExpressionException.h>

#include <RobotAPI/libraries/aron/aroncore/navigators/datanavigator/AronAllDataNavigators.h>
#include <RobotAPI/libraries/aron/aroncore/navigators/visitors/DataVisitor.h>

#include <SimoxUtility/meta/type_name.h>

#include "RemoteGuiAronDataVisitor.h"


namespace armarx::armem::server
{

    MemoryRemoteGui::GroupBox MemoryRemoteGui::makeGroupBox(const Memory& memory) const
    {
        GroupBox group;
        group.setLabel(makeGroupLabel("Memory", memory.name(), memory.coreSegments.size()));

        if (memory.coreSegments.empty())
        {
            group.addChild(Label(makeNoItemsMessage("core segments")));
        }
        for (const auto& [name, coreSegment] : memory.coreSegments)
        {
            group.addChild(makeGroupBox(coreSegment));
        }

        return group;
    }



    MemoryRemoteGui::GroupBox MemoryRemoteGui::makeGroupBox(const CoreSegment& coreSegment) const
    {
        GroupBox group;
        group.setLabel(makeGroupLabel("Core Segment", coreSegment.name(), coreSegment.providerSegments.size()));

        if (coreSegment.providerSegments.empty())
        {
            group.addChild(Label(makeNoItemsMessage("provider segments")));
        }
        for (const auto& [name, providerSegment] : coreSegment.providerSegments)
        {
            group.addChild(makeGroupBox(providerSegment));
        }

        return group;
    }



    MemoryRemoteGui::GroupBox MemoryRemoteGui::makeGroupBox(const ProviderSegment& providerSegment) const
    {
        GroupBox group;
        group.setLabel(makeGroupLabel("Provider Segment", providerSegment.name(), providerSegment.entities.size()));

        if (providerSegment.entities.empty())
        {
            group.addChild(Label(makeNoItemsMessage("entities")));
        }
        for (const auto& [name, entity] : providerSegment.entities)
        {
            group.addChild(makeGroupBox(entity));
        }

        return group;
    }



    MemoryRemoteGui::GroupBox MemoryRemoteGui::makeGroupBox(const Entity& entity) const
    {
        GroupBox group;
        group.setLabel(makeGroupLabel("Entity", entity.name(), entity.history.size()));

        if (entity.history.empty())
        {
            group.addChild(Label(makeNoItemsMessage("snapshots")));
        }
        if (int(entity.history.size()) <= maxHistorySize)
        {
            for (const auto& [time, snapshot] : entity.history)
            {
                group.addChild(makeGroupBox(snapshot));
            }
        }
        else
        {
            int margin = 2;
            auto it = entity.history.begin();
            auto rit = entity.history.end();
            --rit;
            for (int i = 0; i < margin; ++i, ++it)
            {
                group.addChild(makeGroupBox(it->second));
                --rit;
            }
            group.addChild(Label("..."));
            for (; rit != entity.history.end(); ++rit)
            {
                group.addChild(makeGroupBox(rit->second));
            }
        }
        group.setCollapsed(true);

        return group;
    }



    MemoryRemoteGui::GroupBox MemoryRemoteGui::makeGroupBox(const EntitySnapshot& snapshot) const
    {
        GroupBox group;
        group.setLabel(makeGroupLabel("t", armem::toDateTimeMilliSeconds(snapshot.time()),
                                      snapshot.instances.size(), " = ", ""));

        if (snapshot.instances.empty())
        {
            group.addChild(Label(makeNoItemsMessage("instances")));
        }
        for (const EntityInstance& instance : snapshot.instances)
        {
            group.addChild(makeGroupBox(instance));
        }
        group.setCollapsed(true);

        return group;
    }

    MemoryRemoteGui::GroupBox MemoryRemoteGui::makeGroupBox(const EntityInstance& instance) const
    {
        GroupBox group;

        if (instance.data())
        {
            RemoteGuiAronDataVisitor visitor;
            visitor.applyTo(*instance.data());
            group = visitor.result;
        }
        else
        {
            group.addChild(Label("(No data.)"));
        }

        std::stringstream ss;
        ss << "Instance #" << instance.index();
        group.setLabel(ss.str());
        group.setCollapsed(true);

        return group;
    }



    std::string MemoryRemoteGui::makeGroupLabel(
        const std::string& term, const std::string& name, size_t size,
        const std::string& namePrefix, const std::string& nameSuffix) const
    {
        std::stringstream ss;
        ss << term << namePrefix << name << nameSuffix << " (" << size << ")";
        return ss.str();
    }


    std::string MemoryRemoteGui::makeNoItemsMessage(const std::string& term) const
    {
        std::stringstream ss;
        ss << "(no " << term << ")";
        return ss.str();
    }




}
