#include "MemoryToIceAdapter.h"

#include <ArmarXCore/core/exceptions/local/ExpressionException.h>

#include "../error.h"
#include "../core/ice_conversions.h"
#include "query_proc/MemoryQueryProcessor.h"


namespace armarx::armem::server
{

    MemoryToIceAdapter::MemoryToIceAdapter(Memory* memory) : memory(memory)
    {
    }


    void
    MemoryToIceAdapter::setMemoryListener(client::MemoryListenerInterfacePrx memoryListener)
    {
        this->memoryListener = memoryListener;
    }


    // WRITING

    data::AddSegmentResult
    MemoryToIceAdapter::addSegment(const data::AddSegmentInput& input, bool addCoreSegments)
    {
        ARMARX_DEBUG << "Adding segment '" << input.coreSegmentName << "/" << input.providerSegmentName << "'.";
        ARMARX_CHECK_NOT_NULL(memory);

        data::AddSegmentResult output;

        armem::CoreSegment* coreSegment = nullptr;
        try
        {
            coreSegment = &memory->getCoreSegment(input.coreSegmentName);
        }
        catch (const armem::error::MissingEntry& e)
        {
            if (addCoreSegments)
            {
                coreSegment = &memory->addCoreSegment(input.coreSegmentName);
            }
            else
            {
                output.success = false;
                output.errorMessage = e.what();
                return output;
            }
        }
        ARMARX_CHECK_NOT_NULL(coreSegment);

        if (input.providerSegmentName.size() > 0)
        {
            try
            {
                coreSegment->addProviderSegment(input.providerSegmentName);
            }
            catch (const armem::error::ContainerEntryAlreadyExists&)
            {
                // This is ok.
                if (input.clearWhenExists)
                {
                    ProviderSegment& provider = coreSegment->getProviderSegment(input.providerSegmentName);
                    provider.clear();
                }
            }
        }

        armem::MemoryID segmentID;
        segmentID.memoryName = memory->name();
        segmentID.coreSegmentName = input.coreSegmentName;
        segmentID.providerSegmentName = input.providerSegmentName;

        output.success = true;
        output.segmentID = segmentID.str();
        return output;
    }


    data::AddSegmentsResult
    MemoryToIceAdapter::addSegments(const data::AddSegmentsInput& input, bool addCoreSegments)
    {
        ARMARX_CHECK_NOT_NULL(memory);

        data::AddSegmentsResult output;
        for (const auto& i : input)
        {
            output.push_back(addSegment(i, addCoreSegments));
        }
        return output;
    }

    data::CommitResult
    MemoryToIceAdapter::commit(const data::Commit& commitIce, Time timeArrived)
    {
        ARMARX_CHECK_NOT_NULL(memory);

        armem::Commit commit;
        armem::fromIce(commitIce, commit, timeArrived);

        armem::CommitResult result = this->commit(commit);
        data::CommitResult resultIce;
        armem::toIce(resultIce, result);

        return resultIce;
    }


    data::CommitResult
    MemoryToIceAdapter::commit(const data::Commit& commitIce)
    {
        return commit(commitIce, armem::Time::now());
    }


    armem::CommitResult
    MemoryToIceAdapter::commit(const armem::Commit& commit)
    {
        std::vector<data::MemoryID> updatedIDs;
        const bool publishUpdates = memoryListener;

        CommitResult commitResult;
        for (const EntityUpdate& update : commit.updates)
        {
            EntityUpdateResult& result = commitResult.results.emplace_back();
            try
            {
                MemoryID snapshotID = memory->update(update);

                result.success = true;
                result.snapshotID = snapshotID;
                result.timeArrived = update.timeArrived;

                if (publishUpdates)
                {
                    data::MemoryID& id = updatedIDs.emplace_back();
                    armem::toIce(id, snapshotID);
                }
            }
            catch (const error::ArMemError& e)
            {
                result.success = false;
                result.errorMessage = e.what();
            }
        }

        if (publishUpdates)
        {
            memoryListener->memoryUpdated(updatedIDs);
        }

        return commitResult;
    }


    // READING

    armem::query::data::Result
    MemoryToIceAdapter::query(const armem::query::data::Input& input)
    {
        ARMARX_CHECK_NOT_NULL(memory);

        armem::query_proc::MemoryQueryProcessor processor(
            input.withData ? armem::DataMode::WithData
            : armem::DataMode::NoData);
        return processor.process(input, *memory);
    }

}
