#include "MemoryQueryProcessor.h"

#include <regex>

#include <ArmarXCore/core/logging/Logging.h>
#include <ArmarXCore/core/exceptions/local/ExpressionException.h>

#include <RobotAPI/libraries/armem/core/error.h>
#include <RobotAPI/libraries/armem/core/ice_conversions.h>


namespace armarx::armem::query_proc
{

    armem::query::data::Result MemoryQueryProcessor::process(const armem::query::data::Input& input, const Memory& memory) const
    {
        armem::query::data::Result result;
        result.memory = processToIce(input.memoryQueries, memory);
        result.success = true;
        return result;
    }

    data::MemoryPtr MemoryQueryProcessor::processToIce(
        const armem::query::data::MemoryQuery& query, const Memory& memory) const
    {
        return toIce<data::MemoryPtr>(process(query, memory));
    }

    data::MemoryPtr MemoryQueryProcessor::processToIce(const armem::query::data::MemoryQuerySeq& queries, const Memory& memory) const
    {
        return toIce<data::MemoryPtr>(process(queries, memory));
    }

    void MemoryQueryProcessor::process(Memory& result, const armem::query::data::MemoryQuery& query, const Memory& memory) const
    {
        if (auto q = dynamic_cast<const armem::query::data::memory::All*>(&query))
        {
            process(result, *q, memory);
        }
        else if (auto q = dynamic_cast<const armem::query::data::memory::Single*>(&query))
        {
            process(result, *q, memory);
        }
        else if (auto q = dynamic_cast<const armem::query::data::memory::Regex*>(&query))
        {
            process(result, *q, memory);
        }
        else
        {
            throw armem::error::UnknownQueryType("memory segment", query);
        }
    }


    void MemoryQueryProcessor::process(Memory& result, const armem::query::data::memory::All& query, const Memory& memory) const
    {
        for (const auto& [name, coreSegment] : memory.coreSegments)
        {
            result.addCoreSegment(coreSegmentProcessor.process(query.coreSegmentQueries, coreSegment));
        }
    }

    void MemoryQueryProcessor::process(Memory& result, const armem::query::data::memory::Single& query, const Memory& memory) const
    {
        try
        {
            const CoreSegment& coreSegment = memory.getCoreSegment(query.coreSegmentName);
            result.addCoreSegment(coreSegmentProcessor.process(query.coreSegmentQueries, coreSegment));
        }
        catch (const error::MissingEntry&)
        {
            // Leave empty.
        }
    }

    void MemoryQueryProcessor::process(Memory& result, const armem::query::data::memory::Regex& query, const Memory& memory) const
    {
        std::regex regex(query.coreSegmentNameRegex);
        for (const auto& [name, coreSegment] : memory.coreSegments)
        {
            if (std::regex_search(coreSegment.name(), regex))
            {
                result.addCoreSegment(coreSegmentProcessor.process(query.coreSegmentQueries, coreSegment));
            }
        }
    }

}
