/*
 * This file is part of ArmarX.
 *
 * ArmarX is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * ArmarX is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 *
 * @package    RobotAPI::ArmarXObjects::armem
 * @author     Simon Ottenhaus ( simon dot ottenhaus at kit dot edu )
 * @date       2020
 * @copyright  http://www.gnu.org/licenses/gpl-2.0.txt
 *             GNU General Public License
 */

#define BOOST_TEST_MODULE RobotAPI::ArmarXLibraries::armem

#define ARMARX_BOOST_TEST

#include <RobotAPI/Test.h>
#include "../core/Memory.h"
#include "../core/error.h"

#include <RobotAPI/libraries/aron/aroncore/navigators/datanavigator/AronAllDataNavigators.h>
#include <RobotAPI/libraries/aron/aroncore/navigators/typenavigator/AronAllTypeNavigators.h>
#include <RobotAPI/libraries/armem/core/io/MemoryFileSystemStorage.h>
#include <RobotAPI/libraries/aron/aroncore/AronRandomizer.h>

#include <RobotAPI/libraries/aron/aroncore/AronDebug.h>

#include <filesystem>
#include <iostream>


namespace armem = armarx::armem;
namespace aron = armarx::aron;
namespace fs = std::filesystem;

namespace ArMemLTMTest
{
    struct Fixture
    {
        fs::path storagePath = "./TestMemoryExport";

        Fixture()
        {
            clearStoragePath();
        }
        ~Fixture()
        {
            clearStoragePath();
        }

        void clearStoragePath()
        {
            if (fs::exists(storagePath))
            {
                fs::remove_all(storagePath);
            }
            BOOST_TEST_INFO("Storage path: " << storagePath);
            BOOST_REQUIRE(!fs::exists(storagePath));
        }


        static armem::Memory setupMemoryWithType(
            const std::string& memoryName,
            const aron::typenavigator::AronObjectTypeNavigatorPtr& t1,
            const aron::typenavigator::AronObjectTypeNavigatorPtr& t2,
            unsigned int numSnapshots,
            unsigned int numInstances)
        {
            aron::AronRandomizer r;

            armem::Memory memory(memoryName);
            BOOST_CHECK_EQUAL(memory.name(), memoryName);

            armem::CoreSegment& coreSegment = memory.addCoreSegment("TestCoreSegment");
            BOOST_CHECK_EQUAL(coreSegment.name(), "TestCoreSegment");
            BOOST_CHECK(memory.hasCoreSegment(coreSegment.name()));

            armem::ProviderSegment& providerSegment = coreSegment.addProviderSegment("TestProvider");
            BOOST_CHECK_EQUAL(providerSegment.name(), "TestProvider");
            BOOST_CHECK(coreSegment.hasProviderSegment(providerSegment.name()));

            coreSegment.aronType() = t1;
            providerSegment.aronType() = t2;

            aron::typenavigator::AronObjectTypeNavigatorPtr t = t2 != nullptr ? t2 : t1;

            for (unsigned int i = 0; i < numSnapshots; ++i)
            {
                armem::EntityUpdate update;
                std::vector<aron::datanavigator::AronDictDataNavigatorPtr> q;

                for (unsigned int j = 0; j < numInstances; ++j)
                {
                    aron::datanavigator::AronDictDataNavigatorPtr m = aron::datanavigator::AronDictDataNavigator::DynamicCastAndCheck(r.generateAronDataFromType(t));
                    r.initializeRandomly(m, t);
                    q.push_back(m);
                }
                update.entityID = armem::MemoryID::fromString(memoryName + "/TestCoreSegment/TestProvider/TestEntity");
                update.instancesData = q;
                update.timeCreated = armem::Time::now();
                BOOST_CHECK_NO_THROW(providerSegment.update(update));
                BOOST_CHECK_EQUAL(providerSegment.entities.size(), 1);
            }
            BOOST_CHECK(providerSegment.hasEntity("TestEntity"));

            return memory;
        }

        template <class TypeNavigatorT>
        aron::typenavigator::AronObjectTypeNavigatorPtr makeType(const std::string& memberPrefix, int numMembers = 4)
        {
            aron::typenavigator::AronObjectTypeNavigatorPtr t = std::make_shared<aron::typenavigator::AronObjectTypeNavigator>(aron::AronPath());
            t->setObjectName("TestObjectType1");

            for (int i = 0; i < numMembers; ++i)
            {
                t->addAcceptedType(memberPrefix + std::to_string(i + 1), std::make_shared<TypeNavigatorT>());
            }

            return t;
        }
    };
}


BOOST_FIXTURE_TEST_SUITE(ArMemLTMTest, Fixture)


BOOST_AUTO_TEST_CASE(test_memory_export__easy_int_setup)
{
    std::string memoryName = "TestMemory_IntSetup";

    aron::typenavigator::AronObjectTypeNavigatorPtr t = makeType<aron::typenavigator::AronIntTypeNavigator>("theInt");
    armem::Memory memory = setupMemoryWithType(memoryName, t, nullptr, 15, 1);

    // export memory
    BOOST_REQUIRE(!fs::exists(storagePath));
    armem::io::FileSystemMemoryManager mfs(storagePath, true);
    mfs.writeOnDisk(memory);

    std::string mfs_str = mfs.toString();

    armem::io::FileSystemMemoryManager mfs2(storagePath);
    mfs2.update();

    std::string mfs2_str = mfs2.toString();

    //std::cout << "MFS1: " << std::endl;
    //std::cout << mfs_str << std::endl;
    //std::cout << "MFS2: " << std::endl;
    //std::cout << mfs2_str << std::endl;
    BOOST_CHECK_EQUAL(mfs_str == mfs2_str, true);

    armem::Memory memory2 = mfs.readMemoryFromDisk(memoryName);
    BOOST_CHECK_EQUAL(memory.equalsDeep(memory2), true);
}

BOOST_AUTO_TEST_CASE(test_memory_export__easy_long_setup)
{
    std::string memoryName = "TestMemory_LongSetup";

    aron::typenavigator::AronObjectTypeNavigatorPtr t = makeType<aron::typenavigator::AronLongTypeNavigator>("theLong");
    armem::Memory memory = setupMemoryWithType(memoryName, t, nullptr, 15, 1);

    // export memory
    BOOST_REQUIRE(!fs::exists(storagePath));
    armem::io::FileSystemMemoryManager mfs(storagePath, true);
    mfs.writeOnDisk(memory);

    std::string mfs_str = mfs.toString();

    armem::io::FileSystemMemoryManager mfs2(storagePath);
    mfs2.update();

    std::string mfs2_str = mfs2.toString();

    //std::cout << "MFS1: " << std::endl;
    //std::cout << mfs_str << std::endl;
    //std::cout << "MFS2: " << std::endl;
    //std::cout << mfs2_str << std::endl;
    BOOST_CHECK_EQUAL(mfs_str == mfs2_str, true);

    armem::Memory memory2 = mfs.readMemoryFromDisk(memoryName);
    BOOST_CHECK_EQUAL(memory.equalsDeep(memory2), true);
}

BOOST_AUTO_TEST_CASE(test_memory_export__easy_float_setup)
{
    std::string memoryName = "TestMemory_FloatSetup";

    aron::typenavigator::AronObjectTypeNavigatorPtr t = makeType<aron::typenavigator::AronFloatTypeNavigator>("theFloat");
    armem::Memory memory = setupMemoryWithType(memoryName, t, nullptr, 15, 1);

    // export memory
    BOOST_REQUIRE(!fs::exists(storagePath));
    armem::io::FileSystemMemoryManager mfs(storagePath, true);
    mfs.writeOnDisk(memory);

    std::string mfs_str = mfs.toString();

    armem::io::FileSystemMemoryManager mfs2(storagePath);
    mfs2.update();

    std::string mfs2_str = mfs2.toString();

    //std::cout << "MFS1: " << std::endl;
    //std::cout << mfs_str << std::endl;
    //std::cout << "MFS2: " << std::endl;
    //std::cout << mfs2_str << std::endl;
    BOOST_CHECK_EQUAL(mfs_str == mfs2_str, true);

    armem::Memory memory2 = mfs.readMemoryFromDisk(memoryName);
    BOOST_CHECK_EQUAL(memory.equalsDeep(memory2), true);
}

BOOST_AUTO_TEST_CASE(test_memory_export__easy_double_setup)
{
    std::string memoryName = "TestMemory_DoubleSetup";

    aron::typenavigator::AronObjectTypeNavigatorPtr t = makeType<aron::typenavigator::AronDoubleTypeNavigator>("theDouble");
    armem::Memory memory = setupMemoryWithType(memoryName, t, nullptr, 15, 1);

    // export memory
    BOOST_REQUIRE(!fs::exists(storagePath));
    armem::io::FileSystemMemoryManager mfs(storagePath, true);
    mfs.writeOnDisk(memory);

    std::string mfs_str = mfs.toString();

    armem::io::FileSystemMemoryManager mfs2(storagePath);
    mfs2.update();

    std::string mfs2_str = mfs2.toString();

    //std::cout << "MFS1: " << std::endl;
    //std::cout << mfs_str << std::endl;
    //std::cout << "MFS2: " << std::endl;
    //std::cout << mfs2_str << std::endl;
    BOOST_CHECK_EQUAL(mfs_str == mfs2_str, true);

    armem::Memory memory2 = mfs.readMemoryFromDisk(memoryName);
    BOOST_CHECK_EQUAL(memory.equalsDeep(memory2), true);
}

BOOST_AUTO_TEST_CASE(test_memory_export__easy_string_setup)
{
    std::string memoryName = "TestMemory_StringSetup";

    aron::typenavigator::AronObjectTypeNavigatorPtr t = makeType<aron::typenavigator::AronStringTypeNavigator>("theString");
    armem::Memory memory = setupMemoryWithType(memoryName, t, nullptr, 15, 1);

    // export memory
    BOOST_REQUIRE(!fs::exists(storagePath));
    armem::io::FileSystemMemoryManager mfs(storagePath, true);
    mfs.writeOnDisk(memory);

    std::string mfs_str = mfs.toString();

    armem::io::FileSystemMemoryManager mfs2(storagePath);
    mfs2.update();

    std::string mfs2_str = mfs2.toString();

    //std::cout << "MFS1: " << std::endl;
    //std::cout << mfs_str << std::endl;
    //std::cout << "MFS2: " << std::endl;
    //std::cout << mfs2_str << std::endl;
    BOOST_CHECK_EQUAL(mfs_str == mfs2_str, true);

    armem::Memory memory2 = mfs.readMemoryFromDisk(memoryName);
    BOOST_CHECK_EQUAL(memory.equalsDeep(memory2), true);
}

BOOST_AUTO_TEST_CASE(test_memory_export__easy_bool_setup)
{
    std::string memoryName = "TestMemory_BoolSetup";

    aron::typenavigator::AronObjectTypeNavigatorPtr t = makeType<aron::typenavigator::AronBoolTypeNavigator>("theBool");
    armem::Memory memory = setupMemoryWithType(memoryName, t, nullptr, 15, 1);

    // export memory
    BOOST_REQUIRE(!fs::exists(storagePath));
    armem::io::FileSystemMemoryManager mfs(storagePath, true);
    mfs.writeOnDisk(memory);

    std::string mfs_str = mfs.toString();

    armem::io::FileSystemMemoryManager mfs2(storagePath);
    mfs2.update();

    std::string mfs2_str = mfs2.toString();

    //std::cout << "MFS1: " << std::endl;
    //std::cout << mfs_str << std::endl;
    //std::cout << "MFS2: " << std::endl;
    //std::cout << mfs2_str << std::endl;
    BOOST_CHECK_EQUAL(mfs_str == mfs2_str, true);

    armem::Memory memory2 = mfs.readMemoryFromDisk(memoryName);
    BOOST_CHECK_EQUAL(memory.equalsDeep(memory2), true);
}


BOOST_AUTO_TEST_SUITE_END()

