/*
 * This file is part of ArmarX.
 *
 * Copyright (C) 2012-2016, High Performance Humanoid Technologies (H2T),
 * Karlsruhe Institute of Technology (KIT), all rights reserved.
 *
 * ArmarX is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * ArmarX is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 *
 * @author     Fabian Peller-Konrad (fabian dot peller-konrad at kit dot edu)
 * @copyright  http://www.gnu.org/licenses/gpl-2.0.txt
 *             GNU General Public License
 */

// Header
#include "AronPath.h"

// ArmarX
#include <RobotAPI/libraries/aron/aroncore/AronException.h>

namespace armarx
{
    namespace aron
    {
        AronPath::AronPath() :
            rootIdentifier("\\"),
            delimeter("->")
        {

        }

        AronPath::AronPath(const std::vector<std::string>& p) :
            path(p)
        {

        }

        AronPath::AronPath(const std::string& s, const std::string& d):
            rootIdentifier(s),
            delimeter(d)
        {

        }

        AronPath::AronPath(const std::string& s, const std::string& d, const std::vector<std::string>& p) :
            rootIdentifier(s),
            delimeter(d),
            path(p)
        {

        }

        AronPath::AronPath(const AronPath& p) :
            rootIdentifier(p.getRootIdentifier()),
            delimeter(p.getDelimeter()),
            path(p.getPath())
        {

        }

        AronPath::AronPath(const AronPath& p, const std::string& s) :
            AronPath(p)
        {
            appendPath(s);
        }

        AronPath::AronPath(const AronPath& p, const std::string& s, const std::string& s2) :
            AronPath(p)
        {
            appendPath(s);
            appendPath(s2);
        }

        AronPath::AronPath(const AronPath& pa, const std::vector<std::string>& p) :
            AronPath(pa)
        {
            for (const std::string& s : p)
            {
                appendPath(s);
            }
        }

        void AronPath::setRootIdentifier(const std::string& s)
        {
            rootIdentifier = s;
        }

        std::string AronPath::getRootIdentifier() const
        {
            return rootIdentifier;
        }

        void AronPath::setDelimeter(const std::string& d)
        {
            delimeter = d;
        }

        std::string AronPath::getDelimeter() const
        {
            return delimeter;
        }

        void AronPath::appendPath(const std::string& str)
        {
            path.push_back(str);
        }

        std::vector<std::string> AronPath::getPath() const
        {
            return path;
        }

        std::string AronPath::getLastElement() const
        {
            if (!hasElement())
            {
                throw exception::AronException("AronNavigatorPath", "getLastElement", "Try to access last element of empty vector.");
            }
            return path.back();
        }

        std::string AronPath::getFirstElement() const
        {
            if (!hasElement())
            {
                throw exception::AronException("AronNavigatorPath", "getFirstElement", "Try to access last element of empty vector.");
            }
            return path[0];
        }

        bool AronPath::hasElement() const
        {
            return path.size() > 0;
        }

        size_t AronPath::size() const
        {
            return path.size();
        }

        std::string AronPath::toString() const
        {
            std::stringstream ss;
            ss << rootIdentifier;
            for (const std::string& s : path)
            {
                ss << delimeter << s;
            }
            return ss.str();
        }

        AronPath AronPath::withDetachedLastElement() const
        {
            std::vector<std::string> p = path;
            p.pop_back();
            return AronPath(rootIdentifier, delimeter, p);
        }

        AronPath AronPath::withDetachedFirstElement() const
        {
            std::vector<std::string> p = path;
            p.erase(p.begin());
            return AronPath(rootIdentifier, delimeter, p);
        }

        AronPath AronPath::getWithoutPrefix(const AronPath& pref) const
        {
            unsigned int firstWithoutMatch = 0;
            for (const std::string& el : pref.getPath())
            {
                if (path.size() <= firstWithoutMatch || el != path[firstWithoutMatch])
                {
                    break;
                }
                else
                {
                    firstWithoutMatch++;
                }
            }
            std::vector<std::string> elementsWithoutPrefix(path.begin() + firstWithoutMatch, path.end());
            return AronPath(rootIdentifier, delimeter, elementsWithoutPrefix);

        }
    }
}
