/*
 * This file is part of ArmarX.
 *
 * Copyright (C) 2012-2016, High Performance Humanoid Technologies (H2T),
 * Karlsruhe Institute of Technology (KIT), all rights reserved.
 *
 * ArmarX is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * ArmarX is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 *
 * @author     Fabian Peller-Konrad (fabian dot peller-konrad at kit dot edu)
 * @copyright  http://www.gnu.org/licenses/gpl-2.0.txt
 *             GNU General Public License
 */

#pragma once

// STD/STL
#include <memory>
#include <numeric>

// ArmarX
#include <RobotAPI/libraries/aron/aroncore/AronException.h>
#include <RobotAPI/libraries/aron/aroncore/navigators/datanavigator/AronAllDataNavigators.h>
#include <RobotAPI/libraries/aron/aroncore/navigators/typenavigator/AronAllTypeNavigators.h>

namespace armarx
{
    namespace aron
    {
        class AronRandomizer
        {
        public:
            AronRandomizer()
            {
                initialize_random();
            };

            typenavigator::AronTypeNavigatorPtr generateRandomType(bool mustBeObject = false) const
            {
                AronTypeDescriptor nextType = eAronObjectType;
                if (!mustBeObject)
                {
                    std::vector<AronTypeDescriptor> descriptors =
                    {
#define RUN_ARON_MACRO(upperType, lowerType, capsType) \
    eAron##upperType##Type,

                        HANDLE_ALL_ARON_TYPES
#undef RUN_ARON_MACRO
                    };

                    int randomType = generateRandom(descriptors.size(), 0);
                    nextType = descriptors[randomType];
                }

                switch (nextType)
                {
                    case eAronObjectType:
                    {
                        typenavigator::AronObjectTypeNavigatorPtr t = typenavigator::AronObjectTypeNavigatorPtr(new typenavigator::AronObjectTypeNavigator(AronPath()));
                        std::string objectName = generateRandomWord();
                        t->setObjectName(objectName);

                        std::vector<std::string> usedNames({objectName});
                        int members = generateRandom(4, 2);
                        std::set<std::string> usedKeys;
                        for (int i = 0; i < members; ++i)
                        {
                            std::string key = generateRandomWord(usedKeys);
                            usedKeys.insert(key);

                            typenavigator::AronTypeNavigatorPtr m = generateRandomType(false);
                            t->addAcceptedType(key, m);
                        }
                        return t;
                    }
                    case eAronDictType:
                    {
                        typenavigator::AronDictTypeNavigatorPtr t = typenavigator::AronDictTypeNavigatorPtr(new typenavigator::AronDictTypeNavigator(AronPath()));
                        typenavigator::AronTypeNavigatorPtr a = generateRandomType(false);

                        t->setAcceptedType(a);
                        return t;
                    }
                    case eAronTupleType:
                    {
                        typenavigator::AronTupleTypeNavigatorPtr t = typenavigator::AronTupleTypeNavigatorPtr(new typenavigator::AronTupleTypeNavigator(AronPath()));

                        int members = generateRandom(4, 2);
                        for (int i = 0; i < members; ++i)
                        {
                            typenavigator::AronTypeNavigatorPtr m = generateRandomType(false);
                            t->addAcceptedType(m);
                        }
                        return t;
                    }
                    case eAronListType:
                    {
                        typenavigator::AronListTypeNavigatorPtr t = typenavigator::AronListTypeNavigatorPtr(new typenavigator::AronListTypeNavigator(AronPath()));
                        typenavigator::AronTypeNavigatorPtr a = generateRandomType(false);

                        t->setAcceptedType(a);
                        return t;
                    }
#define RUN_ARON_MACRO(upperType, lowerType, capsType) \
case eAron##upperType##Type:

                        HANDLE_COMPLEX_TYPES
#undef RUN_ARON_MACRO

#define RUN_ARON_MACRO(upperType, lowerType, capsType) \
case eAron##upperType##Type: \
{ \
    typenavigator::Aron##upperType##TypeNavigatorPtr t = typenavigator::Aron##upperType##TypeNavigatorPtr(new typenavigator::Aron##upperType##TypeNavigator(AronPath())); \
    return t; \
}

                    HANDLE_PRIMITIVE_TYPES
#undef RUN_ARON_MACRO
                    default:
                    {
                        throw armarx::aron::exception::AronTypeDescriptorNotValidException("AronTest", "generateRandomType", "No valid type found!", nextType);
                    }
                }
            }

            datanavigator::AronDataNavigatorPtr generateAronDataFromType(const typenavigator::AronTypeNavigatorPtr& type) const
            {
                const AronTypeDescriptor desc = type->getDescriptor();
                switch (desc)
                {
                    // In an object, we do not want to edit the keys.
                    case eAronObjectType:
                    {
                        typenavigator::AronObjectTypeNavigatorPtr t = typenavigator::AronObjectTypeNavigator::DynamicCastAndCheck(type);
                        datanavigator::AronDictDataNavigatorPtr d = datanavigator::AronDictDataNavigatorPtr(new datanavigator::AronDictDataNavigator());
                        for (const auto& [k, tt] : t->getAcceptedTypes())
                        {
                            d->addElement(k, generateAronDataFromType(tt));
                        }
                        return d;
                    }

                    // here all totally random
                    case eAronDictType:
                    {
                        typenavigator::AronDictTypeNavigatorPtr t = typenavigator::AronDictTypeNavigator::DynamicCastAndCheck(type);
                        return datanavigator::AronDataNavigatorPtr(new datanavigator::AronDictDataNavigator());
                    }

#define RUN_ARON_MACRO(upperType, lowerType, capsType) \
case eAron##upperType##Type: \
{ \
    typenavigator::Aron##upperType##TypeNavigatorPtr t = typenavigator::Aron##upperType##TypeNavigator::DynamicCastAndCheck(type); \
    return datanavigator::AronDataNavigatorPtr(new datanavigator::AronListDataNavigator()); \
}

                    HANDLE_LIST_SERIALIZER_TYPES
#undef RUN_ARON_MACRO

#define RUN_ARON_MACRO(upperType, lowerType, capsType) \
case eAron##upperType##Type: \
{ \
    typenavigator::Aron##upperType##TypeNavigatorPtr t = typenavigator::Aron##upperType##TypeNavigator::DynamicCastAndCheck(type); \
    return datanavigator::AronDataNavigatorPtr(new datanavigator::AronNDArrayDataNavigator()); \
}

                    HANDLE_COMPLEX_TYPES
#undef RUN_ARON_MACRO

#define RUN_ARON_MACRO(upperType, lowerType, capsType) \
case eAron##upperType##Type: \
{ \
    typenavigator::Aron##upperType##TypeNavigatorPtr t = typenavigator::Aron##upperType##TypeNavigator::DynamicCastAndCheck(type); \
    return datanavigator::AronDataNavigatorPtr(new datanavigator::Aron##upperType##DataNavigator()); \
}

                    HANDLE_PRIMITIVE_TYPES
#undef RUN_ARON_MACRO
                    default:
                    {
                        throw armarx::aron::exception::AronTypeDescriptorNotValidException("AronTest", "generateAronDataFromType", "No valid type found!", desc);
                    }
                }
            }


            void initializeRandomly(datanavigator::AronDictDataNavigatorPtr& data, const typenavigator::AronObjectTypeNavigatorPtr& type) const
            {
                for (auto& [key, nextData] : data->getElements())
                {
                    initializeRandomly(nextData, type->getAcceptedTypes()[key]);
                }
            }

            void initializeRandomly(datanavigator::AronDictDataNavigatorPtr& data, const typenavigator::AronDictTypeNavigatorPtr& type) const
            {
                data->clear();
                int numElements = generateRandom(5, 1);
                std::set<std::string> usedKeys;
                for (int i = 0; i < numElements; ++i)
                {
                    std::string key = generateRandomWord(usedKeys);
                    usedKeys.insert(key);
                    datanavigator::AronDataNavigatorPtr newData = generateAronDataFromType(type->getAcceptedType());
                    initializeRandomly(newData, type->getAcceptedType());
                    data->addElement(key, newData);
                }
            }

            void initializeRandomly(datanavigator::AronListDataNavigatorPtr& data, const typenavigator::AronTupleTypeNavigatorPtr& type) const
            {
                unsigned int i = 0;
                for (auto& nextData : data->getElements())
                {
                    initializeRandomly(nextData, type->getAcceptedTypes()[i++]);
                }
            }

            void initializeRandomly(datanavigator::AronListDataNavigatorPtr& data, const typenavigator::AronListTypeNavigatorPtr& type) const
            {
                data->clear();
                int numElements = generateRandom(5, 1);
                for (int i = 0; i < numElements; ++i)
                {
                    datanavigator::AronDataNavigatorPtr newData = generateAronDataFromType(type->getAcceptedType());
                    initializeRandomly(newData, type->getAcceptedType());
                    data->addElement(newData);
                }
            }

            void initializeRandomly(datanavigator::AronNDArrayDataNavigatorPtr& data, const typenavigator::AronNDArraySerializerTypeNavigatorPtr& type) const
            {
                std::vector<int> dims = data->getDimensions();
                int bytes = std::accumulate(std::begin(dims), std::end(dims), 1, std::multiplies<int>());
                std::vector<unsigned char> blob = generateRandomBlob(bytes);
                data->setData(bytes, blob.data());
            }

            void initializeRandomly(datanavigator::AronIntDataNavigatorPtr& data, const typenavigator::AronPrimitiveTypeNavigatorPtr& type) const
            {
                data->setValue(generateRandom(1000, -1000));
            }

            void initializeRandomly(datanavigator::AronLongDataNavigatorPtr& data, const typenavigator::AronPrimitiveTypeNavigatorPtr& type) const
            {
                data->setValue(generateRandom(1000, -1000));
            }

            void initializeRandomly(datanavigator::AronFloatDataNavigatorPtr& data, const typenavigator::AronPrimitiveTypeNavigatorPtr& type) const
            {
                data->setValue(generateRandom(1000, -1000));
            }

            void initializeRandomly(datanavigator::AronDoubleDataNavigatorPtr& data, const typenavigator::AronPrimitiveTypeNavigatorPtr& type) const
            {
                data->setValue(generateRandom(1000, -1000));
            }

            void initializeRandomly(datanavigator::AronBoolDataNavigatorPtr& data, const typenavigator::AronPrimitiveTypeNavigatorPtr& type) const
            {
                data->setValue(generateRandom(1, 0));
            }

            void initializeRandomly(datanavigator::AronStringDataNavigatorPtr& data, const typenavigator::AronPrimitiveTypeNavigatorPtr& type) const
            {
                data->setValue(generateRandomWord());
            }

            void initializeRandomly(datanavigator::AronDataNavigatorPtr& data, const typenavigator::AronTypeNavigatorPtr& type) const
            {
                // Containers
                AronTypeDescriptor desc = type->getDescriptor();
                switch (desc)
                {

#define RUN_ARON_MACRO(upperType, lowerType, capsType, upperData, lowerData, capsData) \
case eAron##upperType##Type: \
{ \
    typenavigator::Aron##upperType##TypeNavigatorPtr t = typenavigator::Aron##upperType##TypeNavigator::DynamicCastAndCheck(type); \
    datanavigator::Aron##upperData##DataNavigatorPtr d = datanavigator::Aron##upperData##DataNavigator::DynamicCastAndCheck(data); \
    initializeRandomly(d, t); \
    break; \
}

                    HANDLE_ALL_CORRESPONDING
#undef RUN_ARON_MACRO

                    default:
                    {
                        throw armarx::aron::exception::AronTypeDescriptorNotValidException("AronTest", "initializeRandomly", "No valid type found!", desc);
                    }
                }
            }

            int generateRandom(int max, int min) const
            {
                max += 1;
                int random = (std::rand() % (max - min)) + min;
                return random;
            }

            std::string generateRandomWord(const std::set<std::string>& usedKeys = {}) const
            {
                std::vector<string> words =
                {
                    "Lorem", "ipsum", "dolor", "sit", "amet", "consetetur", "sadipscing", "elitr"
                    "sed", "diam", "nonumy", "eirmod", "tempor", "invidunt", "ut", "labore", "et"
                    "dolore", "magna", "aliquyam", "eratsed"
                };

                int i = generateRandom(words.size() - 1, 0);
                std::string key = words[i];

                while (usedKeys.count(key) > 0)
                {
                    key = generateRandomWord(usedKeys);
                }

                return key;
            }

            std::vector<unsigned char> generateRandomBlob(unsigned int size) const
            {
                std::vector<unsigned char> new_blob(size, 0);
                for (unsigned int i = 0; i < size; ++i)
                {
                    new_blob[i] = (generateRandom(127, 0));
                }
                return new_blob;
            }

        protected:
            void initialize_random() const
            {
                std::srand(std::time(nullptr));
            }
        };
    }
}
