/*
* This file is part of ArmarX.
*
* ArmarX is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License version 2 as
* published by the Free Software Foundation.
*
* ArmarX is distributed in the hope that it will be useful, but
* WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program. If not, see <http://www.gnu.org/licenses/>.
*
* @author     Fabian Peller (fabian dot peller at kit dot edu)
* @copyright  http://www.gnu.org/licenses/gpl-2.0.txt
*             GNU General Public License
*/

// STD/STL
#include <memory>
#include <numeric>

// Header
#include "AronDataConverter.h"

#include <RobotAPI/libraries/aron/aroncore/AronDebug.h>



namespace armarx
{
    namespace aron
    {
        namespace io
        {
            void AronDataConverter::ConvertFromReader(AronDataReader& reader, AronDataWriter& writer, const aron::typenavigator::AronTypeNavigatorPtr& expectedStructure)
            {
                AronTypeDescriptor t_desc = expectedStructure != nullptr ? expectedStructure->getDescriptor() : eAronUnknownType;
                AronDataDescriptor desc = reader.getTypeOfNextAron(t_desc);
                switch (desc)
                {
                    case eAronDict:
                    {
                        int elements = reader.readStartDict();
                        writer.writeStartDict();

                        typenavigator::AronTypeNavigatorPtr childType = nullptr;
                        if (t_desc == eAronDictType)
                        {
                            auto t = typenavigator::AronDictTypeNavigator::DynamicCast(expectedStructure);
                            childType = t->getAcceptedType();
                        }

                        for (int i = 0; i < elements; ++i)
                        {
                            std::string key = reader.readKey();
                            writer.writeKey(key);

                            if (t_desc == eAronObjectType)
                            {
                                auto t = typenavigator::AronObjectTypeNavigator::DynamicCast(expectedStructure);
                                childType = t->getAcceptedType(key);
                            }

                            AronDataConverter::ConvertFromReader(reader, writer, childType);
                        }

                        writer.writeEndDict();
                        reader.readEndDict();

                        break;
                    }
                    case eAronList:
                    {
                        int elements = reader.readStartList();
                        writer.writeStartList();

                        typenavigator::AronTypeNavigatorPtr childType = nullptr;
                        if (t_desc == eAronListType)
                        {
                            auto t = typenavigator::AronListTypeNavigator::DynamicCast(expectedStructure);
                            childType = t->getAcceptedType();
                        }

                        for (int i = 0; i < elements; ++i)
                        {
                            if (t_desc == eAronObjectType)
                            {
                                auto t = typenavigator::AronTupleTypeNavigator::DynamicCast(expectedStructure);
                                childType = t->getAcceptedType(i);
                            }

                            AronDataConverter::ConvertFromReader(reader, writer, childType);
                        }

                        writer.writeEndList();
                        reader.readEndList();
                        break;
                    }
#define RUN_ARON_MACRO(upperType, lowerType, capsType) \
case eAron##upperType: \
{ \
    auto [dims, type] = reader.readStart##upperType(); \
    int elements = std::accumulate(std::begin(dims), std::end(dims), 1, std::multiplies<int>()); \
    std::vector<unsigned char> data(elements); \
    reader.readEnd##upperType(data.data()); \
    \
    writer.write##upperType(dims, type, data.data()); \
    break; \
}

                    HANDLE_COMPLEX_DATA
#undef RUN_ARON_MACRO

#define RUN_ARON_MACRO(upperType, lowerType, capsType) \
case eAron##upperType: \
{ \
    lowerType val = reader.readPrimitive<lowerType>(); \
    writer.writePrimitive(val); \
    break; \
}

                    HANDLE_PRIMITIVE_TYPES
#undef RUN_ARON_MACRO

                    default:
                    {
                        throw exception::AronDataDescriptorNotValidException("LegacyAronDataReader", "SetupReaderAndGetResult", "Data-Type could not be resolved", desc);
                    }
                }
            }
        }
    }
}
