/*
* This file is part of ArmarX.
*
* ArmarX is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License version 2 as
* published by the Free Software Foundation.
*
* ArmarX is distributed in the hope that it will be useful, but
* WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program. If not, see <http://www.gnu.org/licenses/>.
*
* @author     Fabian Peller (fabian dot peller at kit dot edu)
* @copyright  http://www.gnu.org/licenses/gpl-2.0.txt
*             GNU General Public License
*/

// STD/STL
#include <memory>
#include <numeric>

// Header
#include "AronDataNlohmannJSONReader.h"

// ArmarX
#include <RobotAPI/libraries/aron/aroncore/AronException.h>
#include <RobotAPI/libraries/aron/aroncore/navigators/datanavigator/AronListDataNavigator.h>
#include <RobotAPI/libraries/aron/aroncore/navigators/datanavigator/AronPrimitiveDataNavigator.h>



namespace armarx
{
    namespace aron
    {
        namespace io
        {
            AronDataNlohmannJSONReader::AronDataNlohmannJSONReader(const nlohmann::json& n) :
                AronDataClassReader<nlohmann::json, AronDataNlohmannJSONReaderTokenPtr>(n)
            {
            }

            AronDataNlohmannJSONReader::AronDataNlohmannJSONReader(const std::string& n) :
                AronDataClassReader<nlohmann::json, AronDataNlohmannJSONReaderTokenPtr>(nlohmann::json::parse(n))
            {
            }

            nlohmann::json AronDataNlohmannJSONReader::getCurrentAndGoToNext()
            {
                if (!readInitialStart)
                {
                    readInitialStart = true;
                    return data;
                }
                AronDataNlohmannJSONReaderTokenPtr lastToken = stack.top();
                return lastToken->getCurrentElementAndIncreaseCnt();
            }

            nlohmann::json AronDataNlohmannJSONReader::getCurrent()
            {
                if (!readInitialStart)
                {
                    readInitialStart = true;
                    return data;
                }
                AronDataNlohmannJSONReaderTokenPtr lastToken = stack.top();
                return lastToken->getCurrentElement();
            }

#define RUN_ARON_MACRO(upperType, lowerType, capsType) \
    int AronDataNlohmannJSONReader::readStart##upperType() \
    { \
        nlohmann::json current_json = getCurrentAndGoToNext(); \
        int c = current_json.size(); \
        AronDataNlohmannJSONReaderTokenPtr newToken = AronDataNlohmannJSONReaderTokenPtr(new AronDataNlohmannJSONReaderToken(eAron##upperType, current_json)); \
        stack.push(newToken); \
        return c; \
    } \
    \
    bool AronDataNlohmannJSONReader::readEnd##upperType() \
    { \
        AronDataNlohmannJSONReaderTokenPtr token = stack.top(); \
        auto desc = token->getDescriptor(); \
        if (desc != AronDataDescriptor::eAron##upperType) \
        { \
            throw exception::AronDataDescriptorNotValidException("AronDataNlohmannJSONReader", "readEnd" + std::string(#upperType), "A token in the stack has the wrong type. Expected " + std::string(#upperType), desc); \
            return false; \
        } \
        \
        if (token->checkedAllChildren()) \
        { \
            stack.pop(); \
            return true; \
        } \
        return false; \
    }

            HANDLE_CONTAINER_DATA
#undef RUN_ARON_MACRO

            // Complex Data
#define RUN_ARON_MACRO(upperType, lowerType, capsType) \
    std::pair<std::vector<int>, std::string> AronDataNlohmannJSONReader::readStart##upperType() \
    { \
        nlohmann::json j = getCurrent(); \
        std::vector<int> dims = j[ARON_DATA_READER_WRITER_NDARRAY_DIMENSIONS_SLUG]; \
        std::string type = j[ARON_DATA_READER_WRITER_NDARRAY_TYPE_SLUG]; \
        return std::make_pair(dims, type); \
    } \
    \
    void AronDataNlohmannJSONReader::readEnd##upperType(unsigned char* data) \
    { \
        if(data == NULL) \
        { \
            throw exception::AronException("AronDataNlohmannJSONReader", "readEnd" + std::string(#upperType), "The corresponding data pointer of an complex aron ptr is NULL."); \
        } \
        nlohmann::json j = getCurrentAndGoToNext(); \
        std::vector<int> dims = j[ARON_DATA_READER_WRITER_NDARRAY_DIMENSIONS_SLUG]; \
        std::vector<unsigned char> d = j[ARON_DATA_READER_WRITER_NDARRAY_DATA_SLUG]; \
        memcpy(data, d.data(), std::accumulate(std::begin(dims), std::end(dims), 1, std::multiplies<int>())); \
    }

            HANDLE_COMPLEX_DATA
#undef RUN_ARON_MACRO

            // Read primitives
#define RUN_ARON_MACRO(upperType, lowerType, capsType) \
    void AronDataNlohmannJSONReader::readPrimitive(lowerType& t) \
    { \
        nlohmann::json j = getCurrentAndGoToNext(); \
        t = j; \
    }

            HANDLE_PRIMITIVE_TYPES
#undef RUN_ARON_MACRO
        }
    }
}
