/*
* This file is part of ArmarX.
*
* ArmarX is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License version 2 as
* published by the Free Software Foundation.
*
* ArmarX is distributed in the hope that it will be useful, but
* WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program. If not, see <http://www.gnu.org/licenses/>.
*
* @author     Fabian Peller (fabian dot peller at kit dot edu)
* @copyright  http://www.gnu.org/licenses/gpl-2.0.txt
*             GNU General Public License
*/

#pragma once

// STD/STL
#include <memory>
#include <string>

// Simox
#include <SimoxUtility/json.h>

// Base Class
#include <RobotAPI/libraries/aron/aroncore/io/AronDataIO/classReaders/AronDataClassReaderToken.h>

// ArmarX
#include <RobotAPI/libraries/aron/aroncore/AronConfig.h>
#include <RobotAPI/libraries/aron/aroncore/AronResolver.h>

namespace armarx
{
    namespace aron
    {
        namespace io
        {
            // BaseClass
            class AronDataNlohmannJSONReaderToken;
            typedef std::shared_ptr<AronDataNlohmannJSONReaderToken> AronDataNlohmannJSONReaderTokenPtr;

            class AronDataNlohmannJSONReaderToken :
                virtual public AronDataClassReaderToken<nlohmann::json, nlohmann::json>
            {
            public:
                using PointerType = AronDataNlohmannJSONReaderTokenPtr;

            public:
                // constructors
                AronDataNlohmannJSONReaderToken() = delete;
                AronDataNlohmannJSONReaderToken(const AronDataDescriptor desc, const nlohmann::json& data) :
                    AronReaderToken<AronDataDescriptor, nlohmann::json, nlohmann::json>(desc, data),
                    AronDataClassReaderToken<nlohmann::json, nlohmann::json>(desc, data)
                {
                    switch (descriptor)
                    {
                        case eAronDict:
                        {
                            for (auto it = data.begin(); it != data.end(); ++it)
                            {
                                allKeys.push_back(it.key());
                            }
                            maxIndex = allKeys.size();
                            break;
                        }
                        case eAronList:
                        {
                            maxIndex = data.size();
                            break;
                        }
                        default:
                            throw exception::AronDataDescriptorNotValidException("AronDataNlohmannJSONReaderToken", "AronDataNlohmannJSONReaderToken", "Received an invalid instance for a reader token.", desc);
                    }
                }

                AronDataDescriptor getTypeOfNextElement(const AronTypeDescriptor hint) const override
                {
                    const nlohmann::json next = getCurrentElement();
                    if (next.is_object())
                    {
                        // Check if specific NDArray key exists
                        if (next.find(ARON_READER_WRITER_TOKEN_NDARRAY_DATA_SLUG) != next.end())
                        {
                            return eAronNDArray;
                        }
                        return eAronDict;
                    }
                    if (next.is_array())
                    {
                        return eAronList;
                    }
                    if (next.is_number_integer())
                    {

                        AronDataDescriptor d = AronResolver::GetFirstIfDescriptorsAreRelated(
                                                   AronResolver::GetCorrespondingDataDescriptor(hint),
                                                   eAronLong);
                        return d;
                    }
                    if (next.is_number_float())
                    {
                        AronDataDescriptor d = AronResolver::GetFirstIfDescriptorsAreRelated(
                                                   AronResolver::GetCorrespondingDataDescriptor(hint),
                                                   eAronDouble);
                        return d;
                    }
                    if (next.is_boolean())
                    {
                        return eAronBool;
                    }
                    if (next.is_string())
                    {
                        return eAronString;
                    }
                    throw exception::AronException("AronDataNlohmannJSONReaderToken", "getTypeOfNextElement", "Could not determine the type of an nlohmann::json object. Could not convert to AronDataDescriptor enum.");
                }

                nlohmann::json getCurrentElement() const override
                {
                    switch (descriptor)
                    {
                        case eAronDict:
                        {
                            nlohmann::json ret = element[resolveCurrentKey()];
                            return ret;
                        }
                        case eAronList:
                        {
                            nlohmann::json ret = element[currentIndex];
                            return ret;
                        }
                        default:
                            throw exception::AronDataDescriptorNotValidException("AronDataNlohmannJSONReaderToken", "getNextElement", "Could not resove a type of a navigator. Allowed are only container types due to performance", descriptor);
                    }
                }

            private:
            };
        }
    }
}
