/*
* This file is part of ArmarX.
*
* ArmarX is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License version 2 as
* published by the Free Software Foundation.
*
* ArmarX is distributed in the hope that it will be useful, but
* WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program. If not, see <http://www.gnu.org/licenses/>.
*
* @author     Fabian Peller (fabian dot peller at kit dot edu)
* @copyright  http://www.gnu.org/licenses/gpl-2.0.txt
*             GNU General Public License
*/

#pragma once

// STD/STL
#include <memory>
#include <string>

// Base class
#include <RobotAPI/libraries/aron/aroncore/io/AronReaderWriterToken.h>

// ArmarX
#include <RobotAPI/libraries/aron/aroncore/AronConfig.h>
#include <RobotAPI/libraries/aron/aroncore/AronException.h>
#include <RobotAPI/libraries/aron/aroncore/AronResolver.h>


namespace armarx
{
    namespace aron
    {
        namespace io
        {
            template<typename AronDescriptorTypename, typename ElementReturnPtrTypename, typename ElementTypename>
            class AronReaderToken;

            template<typename AronDescriptorTypename, typename ElementReturnPtrTypename, typename ElementTypename>
            using AronReaderTokenPtr = std::shared_ptr<AronReaderToken<AronDescriptorTypename, ElementReturnPtrTypename, ElementTypename>> ;

            template<typename AronDescriptorTypename, typename ElementReturnPtrTypename, typename ElementTypename>
            class AronReaderToken :
                public AronReaderWriterToken
            {
            public:
                AronReaderToken() = delete;
                AronReaderToken(const AronDescriptorTypename& desc, const ElementTypename& e) :
                    descriptor(desc), element(e), currentIndex(0), maxIndex(0), currentKey("")
                {
                };

                bool checkedAllChildren() const
                {
                    return currentIndex >= allKeys.size() && currentIndex >= maxIndex;
                }

                AronDescriptorTypename getDescriptor() const
                {
                    return descriptor;
                }

                unsigned int getChildrenSize() const
                {
                    return maxIndex;
                }

                void setCurrentKey(const std::string& s)
                {
                    if (s.empty())
                    {
                        throw exception::AronException("AronReaderToken", "setCurrentKey", "The passed key is empty.");
                    }
                    currentKey = s;
                }

                virtual bool currentIsEligableForResolvingKey() const = 0;
                std::string resolveCurrentKey() const
                {
                    if (!this->currentIsEligableForResolvingKey())
                    {
                        throw exception::AronException("AronReaderToken", "resolveCurrentKey", "Cant get key of token.");
                    }

                    // check outOfBounds
                    if (this->currentIndex >= this->maxIndex)
                    {
                        throw exception::AronSizeNotValidException("AronReaderToken", "resolveCurrentKey", "Index out of bounds.", this->currentIndex, this->maxIndex);
                    }

                    // ugly workaround
                    if (this->currentKey == "") // probably in a dict where the caller does not know which keys are there
                    {
                        if (this->allKeys.size() == 0)
                        {
                            throw exception::AronException("AronReaderToken", "resolveCurrentKey", "The currentKey is not set and the list of all keys is empty!");
                        }
                        return this->allKeys[this->currentIndex];
                    }
                    return this->currentKey;
                }


                virtual ElementReturnPtrTypename getCurrentElement() const = 0;
                ElementReturnPtrTypename getCurrentElementAndIncreaseCnt()
                {
                    ElementReturnPtrTypename next = getCurrentElement();
                    this->currentIndex++;
                    return next;
                }

            protected:
                // members
                AronDescriptorTypename descriptor;
                ElementTypename element;

                unsigned int currentIndex;
                unsigned int maxIndex;

                // dict and object members
                std::string currentKey;
                std::vector<std::string> allKeys;
            };
        }
    }
}
