/*
* This file is part of ArmarX.
*
* ArmarX is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License version 2 as
* published by the Free Software Foundation.
*
* ArmarX is distributed in the hope that it will be useful, but
* WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program. If not, see <http://www.gnu.org/licenses/>.
*
* @author     Fabian Peller (fabian dot peller at kit dot edu)
* @copyright  http://www.gnu.org/licenses/gpl-2.0.txt
*             GNU General Public License
*/

// STD/STL
#include <memory>
#include <numeric>

// Header
#include "AronTypeNavigatorReader.h"

// ArmarX
#include <RobotAPI/libraries/aron/aroncore/AronException.h>
#include <RobotAPI/libraries/aron/aroncore/navigators/typenavigator/AronListTypeNavigator.h>
#include <RobotAPI/libraries/aron/aroncore/navigators/typenavigator/AronPrimitiveTypeNavigator.h>



namespace armarx
{
    namespace aron
    {
        namespace io
        {
            AronTypeNavigatorReader::AronTypeNavigatorReader(const typenavigator::AronContainerTypeNavigatorPtr& n) :
                AronTypeClassReader<typenavigator::AronContainerTypeNavigatorPtr, AronTypeNavigatorReaderTokenPtr>(n)
            {
            }

            AronTypeNavigatorReader::AronTypeNavigatorReader(const type::AronContainerTypePtr& n) :
                AronTypeClassReader<typenavigator::AronContainerTypeNavigatorPtr, AronTypeNavigatorReaderTokenPtr>(
                    typenavigator::AronContainerTypeNavigator::DynamicCast(typenavigator::AronTypeNavigator::FromAronType(n))
                )
            {
            }

            typenavigator::AronTypeNavigatorPtr AronTypeNavigatorReader::getCurrentAndGoToNext()
            {
                if (!readInitialStart)
                {
                    readInitialStart = true;
                    return input;
                }
                AronTypeNavigatorReaderTokenPtr lastToken = stack.top();
                return lastToken->getCurrentElementAndIncreaseCnt();
            }

            typenavigator::AronTypeNavigatorPtr AronTypeNavigatorReader::getCurrent()
            {
                if (!readInitialStart)
                {
                    readInitialStart = true;
                    return input;
                }
                AronTypeNavigatorReaderTokenPtr lastToken = stack.top();
                return lastToken->getCurrentElement();
            }

#define RUN_ARON_MACRO(upperType, lowerType, capsType) \
    int AronTypeNavigatorReader::readStart##upperType##Type() \
    { \
        typenavigator::AronTypeNavigatorPtr current_nav = getCurrentAndGoToNext(); \
        auto desc = current_nav->getDescriptor(); \
        if (desc != AronTypeDescriptor::eAron##upperType##Type) \
        { \
            throw exception::AronTypeDescriptorNotValidException("AronTypeNavigatorReader", "readStart" + std::string(#upperType), "A token in the stack has the wrong type. Expected " + std::string(#upperType), desc); \
            return false; \
        } \
        typenavigator::Aron##upperType##TypeNavigatorPtr current_nav_casted = typenavigator::Aron##upperType##TypeNavigator::DynamicCast(current_nav); \
        AronTypeNavigatorReaderTokenPtr newToken = AronTypeNavigatorReaderTokenPtr(new AronTypeNavigatorReaderToken(current_nav->getDescriptor(), current_nav_casted)); \
        stack.push(newToken); \
        return newToken->getChildrenSize(); \
    } \
    \
    bool AronTypeNavigatorReader::readEnd##upperType##Type() \
    { \
        AronTypeNavigatorReaderTokenPtr token = stack.top(); \
        auto desc = token->getDescriptor(); \
        if (desc != AronTypeDescriptor::eAron##upperType##Type) \
        { \
            throw exception::AronTypeDescriptorNotValidException("AronTypeNavigatorReader", "readEnd" + std::string(#upperType), "A token in the stack has the wrong type. Expected " + std::string(#upperType), desc); \
            return false; \
        } \
        \
        if (token->checkedAllChildren()) \
        { \
            stack.pop(); \
            return true; \
        } \
        return false; \
    }

            HANDLE_CONTAINER_TYPES
#undef RUN_ARON_MACRO

            // Complex Type
#define RUN_ARON_MACRO(upperType, lowerType, capsType) \
    std::pair<std::vector<int>, std::string> AronTypeNavigatorReader::read##upperType##Type() \
    { \
        typenavigator::AronTypeNavigatorPtr nav = getCurrentAndGoToNext(); \
        typenavigator::Aron##upperType##TypeNavigatorPtr casted = typenavigator::Aron##upperType##TypeNavigator::DynamicCastAndCheck(nav); \
        return std::make_pair(casted->getDimensions(), casted->getUsedType()); \
    }

            HANDLE_COMPLEX_TYPES
#undef RUN_ARON_MACRO

            // Read primitives
#define RUN_ARON_MACRO(upperType, lowerType, capsType) \
    void AronTypeNavigatorReader::read##upperType##Type() \
    { \
        typenavigator::AronTypeNavigatorPtr nav = getCurrentAndGoToNext(); \
        typenavigator::Aron##upperType##TypeNavigatorPtr casted = typenavigator::Aron##upperType##TypeNavigator::DynamicCastAndCheck(nav); \
    }

            HANDLE_PRIMITIVE_TYPES
#undef RUN_ARON_MACRO
        }
    }
}
