/*
* This file is part of ArmarX.
*
* ArmarX is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License version 2 as
* published by the Free Software Foundation.
*
* ArmarX is distributed in the hope that it will be useful, but
* WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program. If not, see <http://www.gnu.org/licenses/>.
*
* @author     Fabian Peller (fabian dot peller at kit dot edu)
* @copyright  http://www.gnu.org/licenses/gpl-2.0.txt
*             GNU General Public License
*/

#pragma once

// STD/STL
#include <memory>
#include <string>

// Base Class
#include <RobotAPI/libraries/aron/aroncore/io/AronTypeIO/classReaders/AronTypeClassReaderToken.h>

// ArmarX
#include <RobotAPI/libraries/aron/aroncore/AronConfig.h>
#include <RobotAPI/libraries/aron/aroncore/navigators/typenavigator/AronAllTypeNavigators.h>

namespace armarx
{
    namespace aron
    {
        namespace io
        {
            // BaseClass
            class AronTypeNavigatorReaderToken;
            typedef std::shared_ptr<AronTypeNavigatorReaderToken> AronTypeNavigatorReaderTokenPtr;

            class AronTypeNavigatorReaderToken :
                virtual public AronTypeClassReaderToken<typenavigator::AronTypeNavigatorPtr, typenavigator::AronContainerTypeNavigatorPtr>
            {
            public:
                using PointerType = AronTypeNavigatorReaderTokenPtr;

            public:
                // constructors
                AronTypeNavigatorReaderToken() = delete;
                AronTypeNavigatorReaderToken(const AronTypeDescriptor desc, const typenavigator::AronContainerTypeNavigatorPtr& type) :
                    AronReaderToken<AronTypeDescriptor, typenavigator::AronTypeNavigatorPtr, typenavigator::AronContainerTypeNavigatorPtr>(desc, type),
                    AronTypeClassReaderToken<typenavigator::AronTypeNavigatorPtr, typenavigator::AronContainerTypeNavigatorPtr>(desc, type)
                {
                    switch (descriptor)
                    {
                        case eAronObjectType:
                        {
                            allKeys = typenavigator::AronObjectTypeNavigator::DynamicCast(element)->getAllKeys();
                            maxIndex = allKeys.size();
                            break;
                        }
                        case eAronDictType:
                        {
                            maxIndex = 1;
                            break;
                        }
                        case eAronTupleType:
                        {
                            maxIndex = typenavigator::AronTupleTypeNavigator::DynamicCast(element)->childrenSize();
                            break;
                        }
                        case eAronListType:
                        {
                            maxIndex = 1;
                            break;
                        }
                        default:
                            throw exception::AronTypeDescriptorNotValidException("AronTypeNavigatorReaderToken", "AronTypeNavigatorReaderToken", "Received an invalid instance for a reader token.", desc);
                    }
                }

                std::string getObjectName() const override
                {
                    if(descriptor == eAronObjectType)
                    {
                        const typenavigator::AronObjectTypeNavigatorPtr next = typenavigator::AronObjectTypeNavigator::DynamicCastAndCheck(getCurrentElement());
                        return next->getObjectName();
                    }
                    throw exception::AronTypeDescriptorNotValidException("AronTypeNavigatorReaderToken", "AronTypeNavigatorReaderToken", "Could not get object nme of a non-object token", descriptor);
                }

                AronTypeDescriptor getTypeOfNextElement() const override
                {
                    const typenavigator::AronTypeNavigatorPtr next = getCurrentElement();
                    return next->getDescriptor();
                }

                typenavigator::AronTypeNavigatorPtr getCurrentElement() const override
                {
                    switch (descriptor)
                    {
                        case eAronObjectType:
                        {
                            typenavigator::AronObjectTypeNavigatorPtr casted = typenavigator::AronObjectTypeNavigator::DynamicCastAndCheck(element);
                            typenavigator::AronTypeNavigatorPtr ret = casted->getAcceptedType(resolveCurrentKey());
                            return ret;
                        }
                        case eAronDictType:
                        {
                            typenavigator::AronDictTypeNavigatorPtr casted = typenavigator::AronDictTypeNavigator::DynamicCastAndCheck(element);
                            typenavigator::AronTypeNavigatorPtr ret = casted->getAcceptedType();
                            return ret;
                        }
                        case eAronTupleType:
                        {
                            typenavigator::AronTupleTypeNavigatorPtr casted = typenavigator::AronTupleTypeNavigator::DynamicCastAndCheck(element);
                            typenavigator::AronTypeNavigatorPtr ret = casted->getAcceptedType(currentIndex);
                            return ret;
                        }
                        case eAronListType:
                        {
                            typenavigator::AronListTypeNavigatorPtr casted = typenavigator::AronListTypeNavigator::DynamicCastAndCheck(element);
                            typenavigator::AronTypeNavigatorPtr ret = casted->getAcceptedType();
                            return ret;
                        }
                        default:
                            throw exception::AronTypeDescriptorNotValidException("AronTypeNavigatorReaderToken", "getNextElement", "Could not resove a type of a navigator. Allowed are only container types due to performance", descriptor);
                    }
                }

            private:
            };
        }
    }
}
