/*
* This file is part of ArmarX.
*
* ArmarX is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License version 2 as
* published by the Free Software Foundation.
*
* ArmarX is distributed in the hope that it will be useful, but
* WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program. If not, see <http://www.gnu.org/licenses/>.
*
* @author     Fabian Peller (fabian dot peller at kit dot edu)
* @copyright  http://www.gnu.org/licenses/gpl-2.0.txt
*             GNU General Public License
*/

// STD/STL
#include <memory>
#include <numeric>

// Header
#include "AronTypeNlohmannJSONReader.h"

// ArmarX
#include <RobotAPI/libraries/aron/aroncore/AronException.h>
#include <RobotAPI/libraries/aron/aroncore/navigators/typenavigator/AronListTypeNavigator.h>
#include <RobotAPI/libraries/aron/aroncore/navigators/typenavigator/AronPrimitiveTypeNavigator.h>



namespace armarx
{
    namespace aron
    {
        namespace io
        {
            AronTypeNlohmannJSONReader::AronTypeNlohmannJSONReader(const nlohmann::json& n) :
                AronTypeClassReader<nlohmann::json, AronTypeNlohmannJSONReaderTokenPtr>(n)
            {
            }

            AronTypeNlohmannJSONReader::AronTypeNlohmannJSONReader(const std::string& n) :
                AronTypeClassReader<nlohmann::json, AronTypeNlohmannJSONReaderTokenPtr>(nlohmann::json::parse(n))
            {
            }

            nlohmann::json AronTypeNlohmannJSONReader::getCurrentAndGoToNext()
            {
                if (!readInitialStart)
                {
                    readInitialStart = true;
                    return input;
                }
                AronTypeNlohmannJSONReaderTokenPtr lastToken = stack.top();
                return lastToken->getCurrentElementAndIncreaseCnt();
            }

            nlohmann::json AronTypeNlohmannJSONReader::getCurrent()
            {
                if (!readInitialStart)
                {
                    readInitialStart = true;
                    return input;
                }
                AronTypeNlohmannJSONReaderTokenPtr lastToken = stack.top();
                return lastToken->getCurrentElement();
            }

#define RUN_ARON_MACRO(upperType, lowerType, capsType) \
    int AronTypeNlohmannJSONReader::readStart##upperType##Type() \
    { \
        nlohmann::json current_json = getCurrentAndGoToNext(); \
        AronTypeNlohmannJSONReaderTokenPtr newToken = AronTypeNlohmannJSONReaderTokenPtr(new AronTypeNlohmannJSONReaderToken(eAron##upperType##Type, current_json)); \
        stack.push(newToken); \
        return newToken->getChildrenSize(); \
    } \
    \
    bool AronTypeNlohmannJSONReader::readEnd##upperType##Type() \
    { \
        AronTypeNlohmannJSONReaderTokenPtr token = stack.top(); \
        auto desc = token->getDescriptor(); \
        if (desc != AronTypeDescriptor::eAron##upperType##Type) \
        { \
            throw exception::AronTypeDescriptorNotValidException("AronTypeNlohmannJSONReader", "readEnd" + std::string(#upperType), "A token in the stack has the wrong type. Expected " + std::string(#upperType) + "Type", desc); \
            return false; \
        } \
        \
        if (token->checkedAllChildren()) \
        { \
            stack.pop(); \
            return true; \
        } \
        return false; \
    }

            HANDLE_CONTAINER_TYPES
#undef RUN_ARON_MACRO

            // Complex Type
#define RUN_ARON_MACRO(upperType, lowerType, capsType) \
    std::pair<std::vector<int>, std::string> AronTypeNlohmannJSONReader::read##upperType##Type() \
    { \
        nlohmann::json j = getCurrent(); \
        std::vector<int> dims = j[ARON_TYPE_READER_WRITER_NDARRAY_DIMENSIONS_SLUG]; \
        std::string type = j[ARON_TYPE_READER_WRITER_NDARRAY_TYPE_SLUG]; \
        return std::make_pair(dims, type); \
    }

            HANDLE_COMPLEX_TYPES
#undef RUN_ARON_MACRO

            // Read primitives
#define RUN_ARON_MACRO(upperType, lowerType, capsType) \
    void AronTypeNlohmannJSONReader::read##upperType##Type() \
    { \
        nlohmann::json j = getCurrentAndGoToNext(); \
    }

            HANDLE_PRIMITIVE_TYPES
#undef RUN_ARON_MACRO
        }
    }
}
