/*
* This file is part of ArmarX.
*
* ArmarX is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License version 2 as
* published by the Free Software Foundation.
*
* ArmarX is distributed in the hope that it will be useful, but
* WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program. If not, see <http://www.gnu.org/licenses/>.
*
* @author     Fabian Peller (fabian dot peller at kit dot edu)
* @copyright  http://www.gnu.org/licenses/gpl-2.0.txt
*             GNU General Public License
*/

#pragma once

// STD/STL
#include <memory>
#include <string>

// Simox
#include <SimoxUtility/json.h>

// Base Class
#include <RobotAPI/libraries/aron/aroncore/io/AronTypeIO/classReaders/AronTypeClassReaderToken.h>

// ArmarX
#include <RobotAPI/libraries/aron/aroncore/AronConfig.h>
#include <RobotAPI/libraries/aron/aroncore/AronResolver.h>

namespace armarx
{
    namespace aron
    {
        namespace io
        {
            // BaseClass
            class AronTypeNlohmannJSONReaderToken;
            typedef std::shared_ptr<AronTypeNlohmannJSONReaderToken> AronTypeNlohmannJSONReaderTokenPtr;

            class AronTypeNlohmannJSONReaderToken :
                virtual public AronTypeClassReaderToken<nlohmann::json, nlohmann::json>
            {
            public:
                using PointerType = AronTypeNlohmannJSONReaderTokenPtr;

            public:
                // constructors
                AronTypeNlohmannJSONReaderToken() = delete;
                AronTypeNlohmannJSONReaderToken(const AronTypeDescriptor desc, const nlohmann::json& type) :
                    AronReaderToken<AronTypeDescriptor, nlohmann::json, nlohmann::json>(desc, type),
                    AronTypeClassReaderToken<nlohmann::json, nlohmann::json>(desc, type)
                {
                    switch (descriptor)
                    {
                        case eAronObjectType:
                        {
                            for (auto it = type.begin(); it != type.end(); ++it)
                            {
                                if (
                                    it.key() == ARON_READER_WRITER_TOKEN_OBJECT_NAME_SLUG ||
#define RUN_ARON_MACRO(upperType, lowerType, capsType) \
it.key() == ARON_READER_WRITER_TOKEN_##capsType##_TYPENAME_SLUG ||

                                    HANDLE_PRIMITIVE_TYPES
#undef RUN_ARON_MACRO
                                    it.key() == ARON_READER_WRITER_TOKEN_DICT_ACCEPTED_TYPE_SLUG ||
                                    it.key() == ARON_READER_WRITER_TOKEN_LIST_ACCEPTED_TYPE_SLUG ||
                                    it.key() == ARON_READER_WRITER_TOKEN_NDARRAY_NAME_SLUG ||
                                    it.key() == ARON_READER_WRITER_TOKEN_NDARRAY_DIMENSIONS_SLUG ||
                                    it.key() == ARON_READER_WRITER_TOKEN_NDARRAY_TYPE_SLUG
                                )
                                {
                                    continue;
                                }

                                allKeys.push_back(it.key());
                            }
                            maxIndex = allKeys.size();
                            break;
                        }
                        case eAronDictType:
                        {
                            maxIndex = 1;
                            break;
                        }
                        case eAronTupleType:
                        {
                            maxIndex = type.size();
                            break;
                        }
                        case eAronListType:
                        {
                            maxIndex = 1;
                            break;
                        }
                        default:
                            throw exception::AronTypeDescriptorNotValidException("AronTypeNlohmannJSONReaderToken", "AronTypeNlohmannJSONReaderToken", "Received an invalid instance for a reader token.", desc);
                    }
                }

                std::string getObjectName() const override
                {
                    if (descriptor == eAronObjectType)
                    {
                        return element[ARON_READER_WRITER_TOKEN_OBJECT_NAME_SLUG];
                    }
                    throw exception::AronTypeDescriptorNotValidException("AronTypeNavigatorReaderToken", "AronTypeNavigatorReaderToken", "Could not get object nme of a non-object token", descriptor);
                }

                AronTypeDescriptor getTypeOfNextElement() const override
                {
                    const nlohmann::json next = getCurrentElement();

                    if (next.is_object())
                    {
                        if (next.find(ARON_READER_WRITER_TOKEN_DICT_ACCEPTED_TYPE_SLUG) != next.end())
                        {
                            return eAronDictType;
                        }
                        if (next.find(ARON_READER_WRITER_TOKEN_LIST_ACCEPTED_TYPE_SLUG) != next.end())
                        {
                            return eAronListType;
                        }
                        if (next.find(ARON_READER_WRITER_TOKEN_NDARRAY_NAME_SLUG) != next.end())
                        {
#define RUN_ARON_MACRO(upperType, lowerType, capsType) \
    if (next[ARON_READER_WRITER_TOKEN_NDARRAY_NAME_SLUG] == std::string(#upperType)) \
    { \
        return eAron##upperType##Type; \
    }

                            HANDLE_COMPLEX_TYPES
#undef RUN_ARON_MACRO
                            throw exception::AronException("AronTypeNlohmannJSONReaderToken", "getTypeOfNextElement", "Could not determine the type of an nlohmann::json object. Found indicator of NDArray but could not resolve real type. Found JSON: " + next[ARON_READER_WRITER_TOKEN_NDARRAY_NAME_SLUG].dump(2));
                        }
                        return eAronObjectType;
                    }
                    if (next.is_array())
                    {
                        return eAronTupleType;
                    }
                    if (next.is_string())
                    {
#define RUN_ARON_MACRO(upperType, lowerType, capsType) \
    if (next == ARON_READER_WRITER_TOKEN_##capsType##_TYPENAME_SLUG) \
    { \
        return eAron##upperType##Type; \
    }

                        HANDLE_PRIMITIVE_TYPES
#undef RUN_ARON_MACRO
                    }
                    throw exception::AronException("AronTypeNlohmannJSONReaderToken", "getTypeOfNextElement", "Could not determine the type of an nlohmann::json object. Could not convert to AronTypeDescriptor enum. Found JSON: " + next.dump(2));
                }

                nlohmann::json getCurrentElement() const override
                {
                    switch (descriptor)
                    {
                        case eAronObjectType:
                        {
                            nlohmann::json ret = element[resolveCurrentKey()];
                            return ret;
                        }
                        case eAronDictType:
                        {
                            nlohmann::json ret = element[ARON_READER_WRITER_TOKEN_DICT_ACCEPTED_TYPE_SLUG];
                            return ret;
                        }
                        case eAronTupleType:
                        {
                            nlohmann::json ret = element[currentIndex];
                            return ret;
                        }
                        case eAronListType:
                        {
                            nlohmann::json ret = element[ARON_READER_WRITER_TOKEN_LIST_ACCEPTED_TYPE_SLUG];
                            return ret;
                        }
                        default:
                            throw exception::AronTypeDescriptorNotValidException("AronTypeNlohmannJSONReaderToken", "getNextElement", "Could not resove a type of a navigator. Allowed are only container types due to performance", descriptor);
                    }
                }

            private:
            };
        }
    }
}
