/*
 * This file is part of ArmarX.
 *
 * Copyright (C) 2012-2016, High Performance Humanoid Technologies (H2T),
 * Karlsruhe Institute of Technology (KIT), all rights reserved.
 *
 * ArmarX is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * ArmarX is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 *
 * @author     Fabian Peller-Konrad (fabian dot peller-konrad at kit dot edu)
 * @copyright  http://www.gnu.org/licenses/gpl-2.0.txt
 *             GNU General Public License
 */

#pragma once

// STD/STL
#include <memory>
#include <string>
#include <unordered_map>

// ArmarX
#include <RobotAPI/interface/aron.h>
#include <RobotAPI/libraries/aron/aroncore/navigators/AronNavigator.h>
#include <RobotAPI/libraries/aron/aroncore/AronPath.h>
#include <RobotAPI/libraries/aron/aroncore/navigators/typenavigator/AronTypeNavigator.h>

namespace armarx
{
    namespace aron
    {
        namespace datanavigator
        {
            class AronDataNavigatorFactory;
            typedef std::shared_ptr<AronDataNavigatorFactory> AronDataNavigatorFactoryPtr;

            class AronDataNavigator;
            typedef std::shared_ptr<AronDataNavigator> AronDataNavigatorPtr;

            class AronDataNavigator :
                virtual public navigator::AronNavigator<AronDataDescriptor, data::AronData>
            {
            public:
                using PointerType = AronDataNavigatorPtr;

            public:
                // constructors
                AronDataNavigator() =  delete;
                AronDataNavigator(const AronDataDescriptor&, const AronPath& p = AronPath());

                // operators
                virtual bool equalsDataNavigator(const AronDataNavigatorPtr&) const = 0;

                // static methods
                static AronDataNavigatorPtr FromAronData(const data::AronDataPtr&, const AronPath& = AronPath());
                static std::vector<AronDataNavigatorPtr> FromAronData(const std::vector<data::AronDataPtr>&, const AronPath& = AronPath());
                static std::vector<data::AronDataPtr> ToAronData(const std::vector<AronDataNavigatorPtr>&);

                template<typename AronNavigatorType>
                static typename AronNavigatorType::PointerType DynamicCast(const AronDataNavigatorPtr& n)
                {
                    return AronNavigatorType::DynamicCast(n);
                }

                template<typename AronNavigatorType>
                static typename AronNavigatorType::PointerType DynamicCastAndCheck(const AronDataNavigatorPtr& n)
                {
                    return AronNavigatorType::DynamicCastAndCheck(n);
                }

                // virtual definitions of base class
                virtual data::AronDataPtr getResult() const override = 0;
                virtual std::string getName() const override = 0;

                // virtual definitions
                virtual std::vector<AronDataNavigatorPtr> getChildren() const = 0;
                virtual size_t childrenSize() const = 0;

                virtual typenavigator::AronTypeNavigatorPtr recalculateType() const = 0;
                virtual bool fullfillsType(const typenavigator::AronTypeNavigatorPtr&) const = 0;

            protected:
                static void CheckDataNavigatorPtrForNull(const std::string&, const std::string&, const AronDataNavigatorPtr&);
                static void CheckDataNavigatorPtrForNull(const std::string&, const std::string&, const AronPath&, const AronDataNavigatorPtr&);

            private:
                static const AronDataNavigatorFactoryPtr FACTORY;
            };

            class AronContainerDataNavigator;
            typedef std::shared_ptr<AronContainerDataNavigator> AronContainerDataNavigatorPtr;

            class AronContainerDataNavigator :
                virtual public AronDataNavigator
            {
            public:
                using PointerType = AronContainerDataNavigatorPtr;

                // virtual definitions
                virtual AronDataNavigatorPtr navigateAbsolute(const AronPath& path) const = 0;

                // member functions
                AronDataNavigatorPtr navigateRelative(const AronPath& path) const;

                // static definitions
                static AronContainerDataNavigatorPtr DynamicCast(const AronDataNavigatorPtr& n);
                static AronContainerDataNavigatorPtr DynamicCastAndCheck(const AronDataNavigatorPtr& n);
            };

            class AronComplexDataNavigator;
            typedef std::shared_ptr<AronComplexDataNavigator> AronComplexDataNavigatorPtr;

            class AronComplexDataNavigator :
                virtual public AronDataNavigator
            {
            public:
                using PointerType = AronComplexDataNavigatorPtr;

                static AronComplexDataNavigatorPtr DynamicCast(const AronDataNavigatorPtr& n)
                {
                    AronComplexDataNavigatorPtr casted = std::dynamic_pointer_cast<AronComplexDataNavigator>(n);
                    return casted;
                }
            };

            class AronPrimitiveDataNavigator;
            typedef std::shared_ptr<AronPrimitiveDataNavigator> AronPrimitiveDataNavigatorPtr;

            class AronPrimitiveDataNavigator :
                virtual public AronDataNavigator
            {
            public:
                using PointerType = AronPrimitiveDataNavigatorPtr;

                static AronPrimitiveDataNavigatorPtr DynamicCast(const AronDataNavigatorPtr& n)
                {
                    AronPrimitiveDataNavigatorPtr casted = std::dynamic_pointer_cast<AronPrimitiveDataNavigator>(n);
                    return casted;
                }
            };

        }
    }
}
