/*
 * This file is part of ArmarX.
 *
 * Copyright (C) 2012-2016, High Performance Humanoid Technologies (H2T),
 * Karlsruhe Institute of Technology (KIT), all rights reserved.
 *
 * ArmarX is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * ArmarX is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 *
 * @author     Fabian Peller-Konrad (fabian dot peller-konrad at kit dot edu)
 * @copyright  http://www.gnu.org/licenses/gpl-2.0.txt
 *             GNU General Public License
 */

// Header
#include "AronDictDataNavigator.h"

// ArmarX
#include <ArmarXCore/core/exceptions/Exception.h>
#include <RobotAPI/libraries/aron/aroncore/navigators/datanavigator/AronDataNavigatorFactory.h>
#include <RobotAPI/libraries/aron/aroncore/navigators/typenavigator/AronDictTypeNavigator.h>

namespace armarx
{
    namespace aron
    {
        namespace datanavigator
        {

            // constructors
            AronDictDataNavigator::AronDictDataNavigator(const AronPath& path) :
                AronNavigator<AronDataDescriptor, data::AronData>::AronNavigator(AronDataDescriptor::eAronDict, path),
                AronDataNavigator(AronDataDescriptor::eAronDict, path),
                aron(new data::AronDict())
            {
            }

            AronDictDataNavigator::AronDictDataNavigator(const data::AronDictPtr& o, const AronPath& path) :
                AronNavigator<AronDataDescriptor, data::AronData>::AronNavigator(AronDataDescriptor::eAronDict, path),
                AronDataNavigator(AronDataDescriptor::eAronDict, path),
                aron(o)
            {
                CheckAronPtrForNull("AronDictDataNavigator", "AronDictDataNavigator", getPath(), aron);

                for (const auto& [key, dataPtr] : aron->elements)
                {
                    childrenNavigators[key] = AronDataNavigator::FromAronData(dataPtr, AronPath(path, key));
                }
            }

            AronDictDataNavigator::AronDictDataNavigator(const data::AronDataDict& d, const AronPath& path) :
                AronDictDataNavigator(data::AronDictPtr(new data::AronDict(d)), path)
            {
            }

            AronDictDataNavigator::AronDictDataNavigator(const std::map<std::string, AronDataNavigatorPtr>& m, const AronPath& path) :
                AronDictDataNavigator(path)
            {
                for (const auto& [key, dataPtr] : m)
                {
                    addElement(key, dataPtr);
                }
            }

            // operators
            bool AronDictDataNavigator::operator==(const AronDictDataNavigator& other) const
            {
                for (const auto& [key, nav] : childrenNavigators)
                {
                    if (not(other.hasElement(key)))
                    {
                        return false;
                    }
                    if (not(nav->equalsDataNavigator(other.getElement(key))))
                    {
                        return false;
                    }
                }
                return true;
            }

            bool AronDictDataNavigator::equalsDataNavigator(const AronDataNavigatorPtr& other) const
            {
                if (other == nullptr)
                {
                    return false;
                }
                AronDictDataNavigatorPtr casted = DynamicCast(other);
                return equalsDeep(casted);
            }

            bool AronDictDataNavigator::equalsDeep(const AronDictDataNavigatorPtr& other) const
            {
                return *this == *other;
            }

            // static methods
            AronDictDataNavigatorPtr AronDictDataNavigator::DynamicCast(const AronDataNavigatorPtr& n)
            {
                return std::dynamic_pointer_cast<AronDictDataNavigator>(n);
            }

            AronDictDataNavigator AronDictDataNavigator::DynamicCast(AronDataNavigator& n)
            {
                return dynamic_cast<AronDictDataNavigator&>(n);
            }

            AronDictDataNavigatorPtr AronDictDataNavigator::DynamicCastAndCheck(const AronDataNavigatorPtr& n)
            {
                CheckDataNavigatorPtrForNull("AronDictDataNavigator", "DynamicCastAndCheck[Before]", n);
                AronDictDataNavigatorPtr casted = AronDictDataNavigator::DynamicCast(n);
                CheckDataNavigatorPtrForNull("AronDictDataNavigator", "DynamicCastAndCheck[After]", n->getPath(), casted);
                return casted;
            }

            AronDictDataNavigatorPtr AronDictDataNavigator::FromAronDictPtr(const data::AronDictPtr& aron)
            {
                return std::make_shared<AronDictDataNavigator>(aron);
            }

            data::AronDictPtr AronDictDataNavigator::ToAronDictPtr(const AronDictDataNavigatorPtr& navigator)
            {
                return navigator ? navigator->toAronDictPtr() : nullptr;
            }

            data::AronDictPtr AronDictDataNavigator::toAronDictPtr() const
            {
                return aron;
            }

            // public member functions
            std::vector<std::string> AronDictDataNavigator::getAllKeys() const
            {
                std::vector<std::string> ret;
                for (const auto& [key, _] : childrenNavigators)
                {
                    ret.push_back(key);
                }
                return ret;
            }

            void AronDictDataNavigator::addElement(const std::string& key, const AronDataNavigatorPtr& data)
            {
                this->childrenNavigators[key] = data;
                this->aron->elements[key] = data->getResult();
            }

            bool AronDictDataNavigator::hasElement(const std::string& key) const
            {
                return childrenNavigators.count(key) > 0;
            }

            AronDataNavigatorPtr AronDictDataNavigator::getElement(const std::string& key) const
            {
                auto it = childrenNavigators.find(key);
                if (it == childrenNavigators.end())
                {
                    std::string all_keys = "";
                    for (const auto& child : this->getAllKeys())
                    {
                        all_keys += child + ", ";
                    }
                    throw exception::AronExceptionWithPathInfo("AronDictDataNavigator", "getElement", "Could not find key '" + key + "'. But I found the following keys: [" + all_keys + "]", getPath());
                }
                return it->second;
            }

            std::map<std::string, AronDataNavigatorPtr> AronDictDataNavigator::getElements() const
            {
                return childrenNavigators;
            }

            void AronDictDataNavigator::clear()
            {
                childrenNavigators.clear();
                aron->elements.clear();
            }

            // virtual implementations
            data::AronDataPtr AronDictDataNavigator::getResult() const
            {
                return toAronDictPtr();
            }

            std::string AronDictDataNavigator::getName() const
            {
                return "AronDict";
            }

            typenavigator::AronTypeNavigatorPtr AronDictDataNavigator::recalculateType() const
            {
                typenavigator::AronDictTypeNavigatorPtr typenav = typenavigator::AronDictTypeNavigatorPtr(new typenavigator::AronDictTypeNavigator(getPath()));
                for (const auto& [key, nav] : childrenNavigators)
                {
                    if (typenav->getAcceptedType() == nullptr)
                    {
                        typenav->setAcceptedType(nav->recalculateType());
                        continue;
                    }

                    if (!nav->fullfillsType(typenav->getAcceptedType()))
                    {
                        throw exception::AronExceptionWithPathInfo("AronDictDataNavigator", "recalculateType", "Cannot recalculate the aronType. Inconsistency found for key: " + key, getPath());
                    }
                }
                return typenav->getAcceptedType();
            }

            bool AronDictDataNavigator::fullfillsType(const typenavigator::AronTypeNavigatorPtr& type) const
            {
                if (!AronResolver::TypesAreCorresponding(type->getDescriptor(), getDescriptor()))
                {
                    return false;
                }

                {
                    typenavigator::AronSingleAcceptedTypeHavingTypeNavigatorPtr dictTypenav = typenavigator::AronSingleAcceptedTypeHavingTypeNavigator::DynamicCast(type);
                    if (dictTypenav)
                    {
                        for (const auto& [_, nav] : childrenNavigators)
                        {
                            if (!nav->fullfillsType(dictTypenav->getAcceptedType()))
                            {
                                return false;
                            }
                        }
                        return true;
                    }
                }

                {
                    typenavigator::AronMultipleDictAcceptedTypeHavingTypeNavigatorPtr dictTypenav = typenavigator::AronMultipleDictAcceptedTypeHavingTypeNavigator::DynamicCast(type);
                    if (dictTypenav)
                    {
                        for (const auto& [key, nav] : childrenNavigators)
                        {
                            if (!nav->fullfillsType(dictTypenav->getAcceptedType(key)))
                            {
                                return false;
                            }
                        }
                        return true;
                    }
                }
                return false;
            }

            std::vector<AronDataNavigatorPtr> AronDictDataNavigator::getChildren() const
            {
                std::vector<AronDataNavigatorPtr> ret(childrenNavigators.size());
                for (const auto& [key, nav] : childrenNavigators)
                {
                    ret.push_back(nav);
                }
                return ret;
            }

            size_t AronDictDataNavigator::childrenSize() const
            {
                return childrenNavigators.size();
            }

            AronDataNavigatorPtr AronDictDataNavigator::navigateAbsolute(const AronPath& path) const
            {
                if (!path.hasElement())
                {
                    throw exception::AronException("AronDictDataNavigator", "navigate", "Could not navigate without a valid path. The path was empty.");
                }
                std::string el = path.getFirstElement();
                if (!hasElement(el))
                {
                    throw exception::AronStringNotValidException("AronDictDataNavigator", "navigate", "Could not find an element of a path.", el);
                }

                if (path.size() == 1)
                {
                    return childrenNavigators.at(el);
                }
                else
                {
                    AronPath next = path.withDetachedFirstElement();
                    AronContainerDataNavigatorPtr c = AronContainerDataNavigator::DynamicCastAndCheck(childrenNavigators.at(el));
                    return c->navigateAbsolute(next);
                }
            }
        }
    }
}
