/*
 * This file is part of ArmarX.
 *
 * Copyright (C) 2012-2016, High Performance Humanoid Technologies (H2T),
 * Karlsruhe Institute of Technology (KIT), all rights reserved.
 *
 * ArmarX is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * ArmarX is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 *
 * @author     Fabian Peller-Konrad (fabian dot peller-konrad at kit dot edu)
 * @copyright  http://www.gnu.org/licenses/gpl-2.0.txt
 *             GNU General Public License
 */

// Header
#include "AronListDataNavigator.h"

// ArmarX
#include <ArmarXCore/core/exceptions/Exception.h>
#include <RobotAPI/libraries/aron/aroncore/navigators/datanavigator/AronDataNavigatorFactory.h>
#include <RobotAPI/libraries/aron/aroncore/navigators/typenavigator/AronListTypeNavigator.h>

namespace armarx
{
    namespace aron
    {
        namespace datanavigator
        {
            // constructors
            AronListDataNavigator::AronListDataNavigator(const AronPath& path) :
                AronNavigator<AronDataDescriptor, data::AronData>::AronNavigator(AronDataDescriptor::eAronList, path),
                AronDataNavigator(AronDataDescriptor::eAronList, path),
                aron(new data::AronList())
            {
            }

            AronListDataNavigator::AronListDataNavigator(const data::AronListPtr& l, const AronPath& path) :
                AronNavigator<AronDataDescriptor, data::AronData>::AronNavigator(AronDataDescriptor::eAronList, path),
                AronDataNavigator(AronDataDescriptor::eAronList, path),
                aron(l)
            {
                CheckAronPtrForNull("AronListDataNavigator", "AronListDataNavigator", getPath(), aron);

                unsigned int i = 0;
                for (const auto& dataPtr : l->elements)
                {
                    childrenNavigators.push_back(FromAronData(dataPtr, AronPath(path, std::to_string(i++))));
                }
            }

            AronListDataNavigator::AronListDataNavigator(const data::AronDataList& d, const AronPath& path) :
                AronListDataNavigator(data::AronListPtr(new data::AronList(d)), path)
            {
            }

            AronListDataNavigator::AronListDataNavigator(const std::vector<AronDataNavigatorPtr>& n, const AronPath& path) :
                AronListDataNavigator(path)
            {
                for (const auto& dataPtr : n)
                {
                    addElement(dataPtr);
                }
            }

            // operators
            bool AronListDataNavigator::operator==(const AronListDataNavigator& other) const
            {
                unsigned int i = 0;
                for (const auto& nav : childrenNavigators)
                {
                    if (not(other.hasElement(i)))
                    {
                        return false;
                    }
                    if (not(nav->equalsDataNavigator(other.getElement(i))))
                    {
                        return false;
                    }
                    i++;
                }
                return true;
            }

            bool AronListDataNavigator::equalsDataNavigator(const AronDataNavigatorPtr& other) const
            {
                if (other == nullptr)
                {
                    return false;
                }
                AronListDataNavigatorPtr casted = DynamicCast(other);
                return equalsDeep(casted);
            }

            bool AronListDataNavigator::equalsDeep(const AronListDataNavigatorPtr& other) const
            {
                return *this == *other;
            }

            // static methods
            AronListDataNavigatorPtr AronListDataNavigator::DynamicCast(const AronDataNavigatorPtr& n)
            {
                AronListDataNavigatorPtr casted = std::dynamic_pointer_cast<AronListDataNavigator>(n);
                return casted;
            }

            AronListDataNavigator AronListDataNavigator::DynamicCast(AronDataNavigator& n)
            {
                return dynamic_cast<AronListDataNavigator&>(n);
            }

            AronListDataNavigatorPtr AronListDataNavigator::DynamicCastAndCheck(const AronDataNavigatorPtr& n)
            {
                CheckDataNavigatorPtrForNull("AronListDataNavigator", "DynamicCastAndCheck[Before]", n);
                AronListDataNavigatorPtr casted = AronListDataNavigator::DynamicCast(n);
                CheckDataNavigatorPtrForNull("AronListDataNavigator", "DynamicCastAndCheck[After]", n->getPath(), casted);
                return casted;
            }

            AronListDataNavigatorPtr AronListDataNavigator::FromAronListPtr(const data::AronListPtr& aron)
            {
                return std::make_shared<AronListDataNavigator>(aron);
            }

            data::AronListPtr AronListDataNavigator::ToAronListPtr(const AronListDataNavigatorPtr& navigator)
            {
                return navigator ? navigator->toAronListPtr() : nullptr;
            }

            // public member functions
            void AronListDataNavigator::addElement(const AronDataNavigatorPtr& n)
            {
                childrenNavigators.push_back(n);
                aron->elements.push_back(n->getResult());
            }

            bool AronListDataNavigator::hasElement(unsigned int i) const
            {
                return i < childrenNavigators.size();
            }

            AronDataNavigatorPtr AronListDataNavigator::getElement(unsigned int i) const
            {
                if (i >= childrenNavigators.size())
                {
                    throw exception::AronExceptionWithPathInfo("AronListDataNavigator", "getElement", "The index i = " + std::to_string(i) + " is out of bounds (size = " + std::to_string(childrenNavigators.size()) + ")", getPath());
                }
                return childrenNavigators[i];
            }

            std::vector<AronDataNavigatorPtr> AronListDataNavigator::getElements() const
            {
                return childrenNavigators;
            }

            void AronListDataNavigator::clear()
            {
                childrenNavigators.clear();
                aron->elements.clear();
            }

            data::AronListPtr AronListDataNavigator::toAronListPtr() const
            {
                return aron;
            }

            // virtual implementations
            data::AronDataPtr AronListDataNavigator::getResult() const
            {
                return toAronListPtr();
            }

            std::string AronListDataNavigator::getName() const
            {
                return "AronList";
            }

            typenavigator::AronTypeNavigatorPtr AronListDataNavigator::recalculateType() const
            {
                typenavigator::AronListTypeNavigatorPtr typenav = typenavigator::AronListTypeNavigatorPtr(new typenavigator::AronListTypeNavigator(getPath()));
                unsigned int i = 0;
                for (const auto& nav : childrenNavigators)
                {
                    if (typenav->getAcceptedType() == nullptr)
                    {
                        typenav->setAcceptedType(nav->recalculateType());
                        continue;
                    }

                    if (!nav->fullfillsType(typenav->getAcceptedType()))
                    {
                        throw exception::AronExceptionWithPathInfo("AronListDataNavigator", "recalculateType", "Cannot recalculate the aronType. Inconsistency found for index: " + std::to_string(i), getPath());
                    }
                    ++i;
                }
                return typenav->getAcceptedType();
            }

            bool AronListDataNavigator::fullfillsType(const typenavigator::AronTypeNavigatorPtr& type) const
            {
                if (!AronResolver::TypesAreCorresponding(type->getDescriptor(), getDescriptor()))
                {
                    return false;
                }

                {
                    typenavigator::AronSingleAcceptedTypeHavingTypeNavigatorPtr listTypenav = typenavigator::AronSingleAcceptedTypeHavingTypeNavigator::DynamicCast(type);
                    if (listTypenav)
                    {
                        for (const auto& nav : childrenNavigators)
                        {
                            if (!nav->fullfillsType(listTypenav->getAcceptedType()))
                            {
                                return false;
                            }
                        }
                        return true;
                    }
                }

                {
                    typenavigator::AronMultipleListAcceptedTypeHavingTypeNavigatorPtr listTypenav = typenavigator::AronMultipleListAcceptedTypeHavingTypeNavigator::DynamicCast(type);
                    if (listTypenav)
                    {
                        unsigned int i = 0;
                        for (const auto& nav : childrenNavigators)
                        {
                            if (!nav->fullfillsType(listTypenav->getAcceptedType(i)))
                            {
                                return false;
                            }
                        }
                        return true;
                    }
                }

                return false;
            }

            std::vector<AronDataNavigatorPtr> AronListDataNavigator::getChildren() const
            {
                return childrenNavigators;
            }

            size_t AronListDataNavigator::childrenSize() const
            {
                return childrenNavigators.size();
            }

            AronDataNavigatorPtr AronListDataNavigator::navigateAbsolute(const AronPath& path) const
            {
                if (!path.hasElement())
                {
                    throw exception::AronException("AronListDataNavigator", "navigate", "Could not navigate without a valid path. The path was empty.");
                }
                unsigned int i = std::stoi(path.getFirstElement());
                if (!hasElement(i))
                {
                    throw exception::AronIndexNotValidException("AronListDataNavigator", "navigate", "Could not find an element of a path.", i, childrenSize());
                }

                if (path.size() == 1)
                {
                    return childrenNavigators.at(i);
                }
                else
                {
                    AronPath next = path.withDetachedFirstElement();
                    AronContainerDataNavigatorPtr c = AronContainerDataNavigator::DynamicCastAndCheck(childrenNavigators.at(i));
                    return c->navigateAbsolute(next);
                }
            }
        }
    }
}
