/*
 * This file is part of ArmarX.
 *
 * Copyright (C) 2012-2016, High Performance Humanoid Technologies (H2T),
 * Karlsruhe Institute of Technology (KIT), all rights reserved.
 *
 * ArmarX is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * ArmarX is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 *
 * @author     Fabian Peller-Konrad (fabian dot peller-konrad at kit dot edu)
 * @copyright  http://www.gnu.org/licenses/gpl-2.0.txt
 *             GNU General Public License
 */

// Header
#include "AronEigenMatrixTypeNavigator.h"

namespace armarx
{
    namespace aron
    {
        namespace typenavigator
        {
            const std::map<std::string, std::vector<std::string>> AronEigenMatrixTypeNavigator::ACCEPTED_TYPES =
            {
                {"short", {}},
                {"int", {}},
                {"long", {}},
                {"float", {}},
                {"double", {}},
                //{"complex float", {"std::complex<float>"}},
                //{"complex double", {"std::complex<double>"}},
            };

            // constructors
            AronEigenMatrixTypeNavigator::AronEigenMatrixTypeNavigator(const AronPath& path) :
                navigator::AronNavigator<AronTypeDescriptor, type::AronType>::AronNavigator(AronTypeDescriptor::eAronEigenMatrixType, path),
                AronTypeNavigator(AronTypeDescriptor::eAronEigenMatrixType, path),
                type(new type::AronEigenMatrixType({1, 1}, ""))
            {
            }

            AronEigenMatrixTypeNavigator::AronEigenMatrixTypeNavigator(const type::AronEigenMatrixTypePtr& o, const AronPath& path) :
                AronNavigator<AronTypeDescriptor, type::AronType>::AronNavigator(AronTypeDescriptor::eAronDictType, path),
                AronTypeNavigator(AronTypeDescriptor::eAronEigenMatrixType, path),
                type(o)
            {
                CheckAronPtrForNull("AronEigenMatrixTypeNavigator", "AronEigenMatrixTypeNavigator", getPath(), o);
                if (type->dimensions.size() != 2)
                {
                    throw exception::AronSizeNotValidException("AronEigenMatrixTypeNavigator", "AronEigenMatrixTypeNavigator", "The size of the AronPtr is too small", type->dimensions.size(), 2);
                }
            }

            unsigned int AronEigenMatrixTypeNavigator::getRows() const
            {
                return type->dimensions[0];
            }

            unsigned int AronEigenMatrixTypeNavigator::getCols() const
            {
                return type->dimensions[1];
            }

            void AronEigenMatrixTypeNavigator::setRows(const unsigned int& w)
            {
                if (w == 0)
                {
                    throw exception::AronExceptionWithPathInfo("AronEigenMatrixTypeNavigator", "setRows", "The rows of cannot be 0", getPath());
                }
                type->dimensions[0] = w;
            }

            void AronEigenMatrixTypeNavigator::setCols(const unsigned int& h)
            {
                if (h == 0)
                {
                    throw exception::AronExceptionWithPathInfo("AronEigenMatrixTypeNavigator", "setCols", "The cols cannot be 0", getPath());
                }
                type->dimensions[1] = h;
            }

            type::AronEigenMatrixTypePtr AronEigenMatrixTypeNavigator::toAronEigenMatrixTypePtr() const
            {
                if (type->dimensions.empty())
                {
                    throw exception::AronExceptionWithPathInfo("AronEigenMatrixTypeNavigator", "getCastedResult", "The dimension eigen matrix is empty", getPath());
                }
                if (std::any_of(type->dimensions.begin(), type->dimensions.end(), [](int i)
            {
                return i < -1;
            }))
                {
                    throw exception::AronExceptionWithPathInfo("AronEigenMatrixTypeNavigator", "getCastedResult", "The dimension size is wrong. At least one empty is <-1", getPath());
                }
                if (type->typeName.empty())
                {
                    throw exception::AronExceptionWithPathInfo("AronEigenMatrixTypeNavigator", "getCastedResult", "The useType of eigen matrix is empty.", getPath());
                }
                return type;
            }

            std::string AronEigenMatrixTypeNavigator::getUsedType() const
            {
                return type->typeName;
            }

            std::vector<int> AronEigenMatrixTypeNavigator::getDimensions() const
            {
                return type->dimensions;
            }

            void AronEigenMatrixTypeNavigator::setUsedType(const std::string& u)
            {
                if (u.empty())
                {
                    throw exception::AronExceptionWithPathInfo("AronEigenMatrixTypeNavigator", "setUsedType", "The useType cannot be empty", getPath());
                }

                for (const auto& [key, vec] : ACCEPTED_TYPES)
                {
                    if (u == key || std::find(vec.begin(), vec.end(), key) != vec.end())
                    {
                        type->typeName = key;
                        return;
                    }
                }
                throw exception::AronStringNotValidException("AronEigenMatrixTypeNavigator", "setUsedType", "The type was not found in the ACCEPTED_TYPES", u);
            }

            void AronEigenMatrixTypeNavigator::setDimensions(const std::vector<int>& d)
            {
                if (d.empty())
                {
                    throw exception::AronExceptionWithPathInfo("AronEigenMatrixTypeNavigator", "setDimensions", "The dimensions of an eigen matrix cannot be empty", getPath());
                }
                if (std::any_of(d.begin(), d.end(), [](int i)
            {
                return i < -1;
            }))
                {
                    throw exception::AronExceptionWithPathInfo("AronEigenMatrixTypeNavigator", "setDimensions", "The dimension size is wrong. At least one empty is <-1", getPath());
                }
                type->dimensions = d;
            }

            // static methods
            AronEigenMatrixTypeNavigatorPtr AronEigenMatrixTypeNavigator::DynamicCast(const AronTypeNavigatorPtr& n)
            {
                return std::dynamic_pointer_cast<AronEigenMatrixTypeNavigator>(n);
            }

            AronEigenMatrixTypeNavigatorPtr AronEigenMatrixTypeNavigator::DynamicCastAndCheck(const AronTypeNavigatorPtr& n)
            {
                CheckTypeNavigatorPtrForNull("AronEigenMatrixTypeNavigator", "DynamicCast[Before]", n);
                AronEigenMatrixTypeNavigatorPtr casted = AronEigenMatrixTypeNavigator::DynamicCast(n);
                CheckTypeNavigatorPtrForNull("AronEigenMatrixTypeNavigator", "DynamicCast[After]", n->getPath(), casted);
                return casted;
            }

            // virtual implementations
            type::AronTypePtr AronEigenMatrixTypeNavigator::getResult() const
            {
                return toAronEigenMatrixTypePtr();
            }

            std::string AronEigenMatrixTypeNavigator::getName() const
            {
                return "AronEigenMatrixType<" + simox::alg::to_string(type->dimensions, ", ") + ", " + type->typeName + ">";
            }
        }
    }
}

