/*
 * This file is part of ArmarX.
 *
 * Copyright (C) 2012-2016, High Performance Humanoid Technologies (H2T),
 * Karlsruhe Institute of Technology (KIT), all rights reserved.
 *
 * ArmarX is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * ArmarX is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 *
 * @author     Fabian Peller-Konrad (fabian dot peller-konrad at kit dot edu)
 * @copyright  http://www.gnu.org/licenses/gpl-2.0.txt
 *             GNU General Public License
 */

// Header
#include "AronPCLPointCloudTypeNavigator.h"

namespace armarx
{
    namespace aron
    {
        namespace typenavigator
        {
            const std::map<std::string, std::vector<std::string>> AronPCLPointCloudTypeNavigator::ACCEPTED_TYPES =
            {
                {"PointXYZ", {"pcl::PointXYZ"}},
                {"PointXYZI", {"pcl::PointXYZI"}},
                {"PointXYZL", {"pcl::PointXYZL"}},
                {"PointXYZRGB", {"pcl::PointXYZRGB"}},
                {"PointXYZRGBL", {"pcl::PointXYZRGBL"}},
                {"PointXYZRGBA", {"pcl::PointXYZRGBA"}},
                {"PointXYZHSV", {"pcl::PointXYZHSV"}},
            };

            // constructors
            AronPCLPointCloudTypeNavigator::AronPCLPointCloudTypeNavigator(const AronPath& path) :
                navigator::AronNavigator<AronTypeDescriptor, type::AronType>::AronNavigator(AronTypeDescriptor::eAronPCLPointCloudType, path),
                AronTypeNavigator(AronTypeDescriptor::eAronPCLPointCloudType, path),
                type(new type::AronPCLPointCloudType({0, 0}, ""))
            {
            }

            AronPCLPointCloudTypeNavigator::AronPCLPointCloudTypeNavigator(const type::AronPCLPointCloudTypePtr& o, const AronPath& path) :
                navigator::AronNavigator<AronTypeDescriptor, type::AronType>::AronNavigator(AronTypeDescriptor::eAronPCLPointCloudType, path),
                AronTypeNavigator(AronTypeDescriptor::eAronPCLPointCloudType, path),
                type(o)
            {
                CheckAronPtrForNull("AronPCLPointCloudTypeNavigator", "AronPCLPointCloudTypeNavigator", getPath(), o);
            }

            type::AronPCLPointCloudTypePtr AronPCLPointCloudTypeNavigator::toAronPCLPointCloudTypePtr() const
            {
                if (type->dimensions.size() != 2)
                {
                    throw exception::AronExceptionWithPathInfo("AronPCLPointCloudTypeNavigator", "getCastedResult", "The dimension size is wrong. Got size: " + std::to_string(type->dimensions.size()), getPath());
                }
                if (std::any_of(type->dimensions.begin(), type->dimensions.end(), [](int i)
            {
                return i <= 0;
            }))
                {
                    throw exception::AronExceptionWithPathInfo("AronPCLPointCloudTypeNavigator", "getCastedResult", "The dimension size is wrong. At least one empty is <=0", getPath());
                }
                if (type->typeName.empty())
                {
                    throw exception::AronExceptionWithPathInfo("AronPCLPointCloudTypeNavigator", "getCastedResult", "The useType of image is empty.", getPath());
                }
                return type;
            }

            std::vector<int> AronPCLPointCloudTypeNavigator::getDimensions() const
            {
                return type->dimensions;
            }

            void AronPCLPointCloudTypeNavigator::setDimensions(const std::vector<int>& d)
            {
                if (d.size() != 2)
                {
                    throw exception::AronExceptionWithPathInfo("AronPCLPointCloudTypeNavigator", "setDimensions", "Wrong number of elements in dimension array. The size must be equal to 2. The size was: " + std::to_string(d.size()), getPath());
                }
                type->dimensions = d;
            }

            unsigned int AronPCLPointCloudTypeNavigator::getWidth() const
            {
                return type->dimensions[0];
            }

            unsigned int AronPCLPointCloudTypeNavigator::getHeight() const
            {
                return type->dimensions[1];
            }

            std::string AronPCLPointCloudTypeNavigator::getUsedType() const
            {
                return type->typeName;
            }

            void AronPCLPointCloudTypeNavigator::setWidth(const unsigned int& w)
            {
                if (w == 0)
                {
                    throw exception::AronExceptionWithPathInfo("AronPCLPointCloudTypeNavigator", "setWidth", "The width of aa pointcloud cannot be 0", getPath());
                }
                type->dimensions[0] = w;
            }

            void AronPCLPointCloudTypeNavigator::setHeight(const unsigned int& h)
            {
                if (h == 0)
                {
                    throw exception::AronExceptionWithPathInfo("AronPCLPointCloudTypeNavigator", "setHeight", "The height of aa pointcloud cannot be 0", getPath());
                }
                type->dimensions[1] = h;
            }

            void AronPCLPointCloudTypeNavigator::setUsedType(const std::string& u)
            {
                if (u.empty())
                {
                    throw exception::AronExceptionWithPathInfo("AronPCLPointCloudTypeNavigator", "setUsedType", "The useType cannot be empty", getPath());
                }

                for (const auto& [key, vec] : ACCEPTED_TYPES)
                {
                    if (u == key || std::find(vec.begin(), vec.end(), key) != vec.end())
                    {
                        type->typeName = key;
                        return;
                    }
                }
                throw exception::AronStringNotValidException("AronPCLPointCloudTypeNavigator", "setUsedType", "The type was not found in the ACCEPTED_TYPES", u);
            }


            // static methods
            AronPCLPointCloudTypeNavigatorPtr AronPCLPointCloudTypeNavigator::DynamicCast(const AronTypeNavigatorPtr& n)
            {
                return std::dynamic_pointer_cast<AronPCLPointCloudTypeNavigator>(n);
            }

            AronPCLPointCloudTypeNavigatorPtr AronPCLPointCloudTypeNavigator::DynamicCastAndCheck(const AronTypeNavigatorPtr& n)
            {
                CheckTypeNavigatorPtrForNull("AronPCLPointCloudTypeNavigator", "DynamicCast[Before]", n);
                AronPCLPointCloudTypeNavigatorPtr casted = AronPCLPointCloudTypeNavigator::DynamicCast(n);
                CheckTypeNavigatorPtrForNull("AronPCLPointCloudTypeNavigator", "DynamicCast[After]", n->getPath(), casted);
                return casted;
            }

            // virtual implementations
            type::AronTypePtr AronPCLPointCloudTypeNavigator::getResult() const
            {
                return toAronPCLPointCloudTypePtr();
            }

            std::string AronPCLPointCloudTypeNavigator::getName() const
            {
                return "AronPCLPointCloudType<" + simox::alg::to_string(type->dimensions, ", ") + ", " + type->typeName + ">";
            }
        }
    }
}

