/*
* This file is part of ArmarX.
*
* Copyright (C) 2012-2016, High Performance Humanoid Technologies (H2T),
* Karlsruhe Institute of Technology (KIT), all rights reserved.
*
* ArmarX is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License version 2 as
* published by the Free Software Foundation.
*
* ArmarX is distributed in the hope that it will be useful, but
* WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program. If not, see <http://www.gnu.org/licenses/>.
*
* @author     Fabian Peller-Konrad (fabian dot peller-konrad at kit dot edu)
* @copyright  http://www.gnu.org/licenses/gpl-2.0.txt
*             GNU General Public License
*/

#pragma once

// STD/STL
#include <memory>
#include <vector>
#include <map>
#include <string>

// Simox
#include <SimoxUtility/algorithm/string.h>

// ArmarX
#include <RobotAPI/libraries/aron/aroncore/AronException.h>
#include <RobotAPI/interface/aron.h>
#include <RobotAPI/libraries/aron/aroncore/navigators/AronNavigator.h>

namespace armarx
{
    namespace aron
    {
        namespace typenavigator
        {
            class AronTypeNavigatorFactory;
            typedef std::shared_ptr<AronTypeNavigatorFactory> AronTypeNavigatorFactoryPtr;

            class AronTypeNavigator;
            typedef std::shared_ptr<AronTypeNavigator> AronTypeNavigatorPtr;

            class AronTypeNavigator :
                virtual public navigator::AronNavigator<AronTypeDescriptor, type::AronType>
            {
            public:
                using PointerType = AronTypeNavigatorPtr;

            public:
                // constructors
                AronTypeNavigator() = delete;
                AronTypeNavigator(const AronTypeDescriptor&, const AronPath&);

                // virtual methods
                virtual type::AronTypePtr getResult() const override = 0;
                virtual std::string getName() const override = 0;
                virtual size_t childrenSize() const = 0;

                // static methods
                static AronTypeNavigatorPtr FromAronType(const type::AronTypePtr&, const AronPath& = AronPath());

                template<typename AronNavigatorType>
                static typename AronNavigatorType::PointerType DynamicCast(const AronTypeNavigatorPtr& n)
                {
                    return AronNavigatorType::DynamicCast(n);
                }

                template<typename AronNavigatorType>
                static typename AronNavigatorType::PointerType DynamicCastAndCheck(const AronTypeNavigatorPtr& n)
                {
                    return AronNavigatorType::DynamicCastAndCheck(n);
                }

            protected:
                static void CheckTypeNavigatorPtrForNull(const std::string&, const std::string&, const AronTypeNavigatorPtr& data);
                static void CheckTypeNavigatorPtrForNull(const std::string&, const std::string&, const AronPath&, const AronTypeNavigatorPtr& data);

            private:
                static const AronTypeNavigatorFactoryPtr FACTORY;
            };


            /* More hierarchy */
            class AronContainerTypeNavigator;
            typedef std::shared_ptr<AronContainerTypeNavigator> AronContainerTypeNavigatorPtr;

            class AronContainerTypeNavigator :
                virtual public AronTypeNavigator
            {
            public:
                using PointerType = AronContainerTypeNavigatorPtr;

            public:
                // static methods
                static AronContainerTypeNavigatorPtr DynamicCast(const AronTypeNavigatorPtr& n)
                {
                    AronContainerTypeNavigatorPtr casted = std::dynamic_pointer_cast<AronContainerTypeNavigator>(n);
                    return casted;
                }
            };



            class AronSingleAcceptedTypeHavingTypeNavigator;
            typedef std::shared_ptr<AronSingleAcceptedTypeHavingTypeNavigator> AronSingleAcceptedTypeHavingTypeNavigatorPtr;

            class AronSingleAcceptedTypeHavingTypeNavigator :
                virtual public AronContainerTypeNavigator
            {
            public:
                using PointerType = AronSingleAcceptedTypeHavingTypeNavigatorPtr;

            public:
                // virtual methods
                virtual void setAcceptedType(const AronTypeNavigatorPtr&) = 0;
                virtual AronTypeNavigatorPtr getAcceptedType() const;

                virtual size_t childrenSize() const override;


                // static methods
                static AronSingleAcceptedTypeHavingTypeNavigatorPtr DynamicCast(const AronTypeNavigatorPtr& n)
                {
                    AronSingleAcceptedTypeHavingTypeNavigatorPtr casted = std::dynamic_pointer_cast<AronSingleAcceptedTypeHavingTypeNavigator>(n);
                    return casted;
                }
            protected:
                AronTypeNavigatorPtr acceptedTypeNavigator;
            };



            class AronMultipleListAcceptedTypeHavingTypeNavigator;
            typedef std::shared_ptr<AronMultipleListAcceptedTypeHavingTypeNavigator> AronMultipleListAcceptedTypeHavingTypeNavigatorPtr;

            class AronMultipleListAcceptedTypeHavingTypeNavigator :
                virtual public AronContainerTypeNavigator
            {
            public:
                using PointerType = AronMultipleListAcceptedTypeHavingTypeNavigatorPtr;

            public:
                virtual void addAcceptedType(const AronTypeNavigatorPtr&) = 0;
                virtual std::vector<AronTypeNavigatorPtr> getAcceptedTypes() const;
                virtual AronTypeNavigatorPtr getAcceptedType(unsigned int) const;

                virtual size_t childrenSize() const override;

                // static methods
                static AronMultipleListAcceptedTypeHavingTypeNavigatorPtr DynamicCast(const AronTypeNavigatorPtr& n)
                {
                    AronMultipleListAcceptedTypeHavingTypeNavigatorPtr casted = std::dynamic_pointer_cast<AronMultipleListAcceptedTypeHavingTypeNavigator>(n);
                    return casted;
                }
            protected:
                // members
                std::vector<AronTypeNavigatorPtr> acceptedTypeNavigators;
            };



            class AronMultipleDictAcceptedTypeHavingTypeNavigator;
            typedef std::shared_ptr<AronMultipleDictAcceptedTypeHavingTypeNavigator> AronMultipleDictAcceptedTypeHavingTypeNavigatorPtr;

            class AronMultipleDictAcceptedTypeHavingTypeNavigator :
                virtual public AronContainerTypeNavigator
            {
            public:
                using PointerType = AronMultipleDictAcceptedTypeHavingTypeNavigatorPtr;

            public:
                virtual void addAcceptedType(const std::string&, const AronTypeNavigatorPtr&) = 0;
                virtual std::map<std::string, AronTypeNavigatorPtr> getAcceptedTypes() const;
                virtual AronTypeNavigatorPtr getAcceptedType(const std::string&) const;

                virtual size_t childrenSize() const override;

                std::vector<std::string> getAllKeys() const;

                // static methods
                static AronMultipleDictAcceptedTypeHavingTypeNavigatorPtr DynamicCast(const AronTypeNavigatorPtr& n)
                {
                    AronMultipleDictAcceptedTypeHavingTypeNavigatorPtr casted = std::dynamic_pointer_cast<AronMultipleDictAcceptedTypeHavingTypeNavigator>(n);
                    return casted;
                }
            protected:
                // members
                std::map<std::string, AronTypeNavigatorPtr> acceptedTypeNavigators;
            };



            class AronDictSerializerTypeNavigator;
            typedef std::shared_ptr<AronDictSerializerTypeNavigator> AronDictSerializerTypeNavigatorPtr;

            class AronDictSerializerTypeNavigator :
                virtual public AronContainerTypeNavigator
            {
            public:
                using PointerType = AronDictSerializerTypeNavigatorPtr;

            public:
                // static methods
                static AronDictSerializerTypeNavigatorPtr DynamicCast(const AronTypeNavigatorPtr& n)
                {
                    AronDictSerializerTypeNavigatorPtr casted = std::dynamic_pointer_cast<AronDictSerializerTypeNavigator>(n);
                    return casted;
                }
            };



            class AronListSerializerTypeNavigator;
            typedef std::shared_ptr<AronListSerializerTypeNavigator> AronListSerializerTypeNavigatorPtr;

            class AronListSerializerTypeNavigator :
                virtual public AronContainerTypeNavigator
            {
            public:
                using PointerType = AronListSerializerTypeNavigatorPtr;

            public:
                // static methods
                static AronListSerializerTypeNavigatorPtr DynamicCast(const AronTypeNavigatorPtr& n)
                {
                    AronListSerializerTypeNavigatorPtr casted = std::dynamic_pointer_cast<AronListSerializerTypeNavigator>(n);
                    return casted;
                }
            };



            class AronComplexTypeNavigator;
            typedef std::shared_ptr<AronComplexTypeNavigator> AronComplexTypeNavigatorPtr;

            class AronComplexTypeNavigator :
                virtual public AronTypeNavigator
            {
            public:
                using PointerType = AronComplexTypeNavigatorPtr;

                virtual size_t childrenSize() const override;

                // static methods
                static AronComplexTypeNavigatorPtr DynamicCast(const AronTypeNavigatorPtr& n)
                {
                    AronComplexTypeNavigatorPtr casted = std::dynamic_pointer_cast<AronComplexTypeNavigator>(n);
                    return casted;
                }
            };



            class AronNDArraySerializerTypeNavigator;
            typedef std::shared_ptr<AronNDArraySerializerTypeNavigator> AronNDArraySerializerTypeNavigatorPtr;

            class AronNDArraySerializerTypeNavigator :
                virtual public AronComplexTypeNavigator
            {
            public:
                using PointerType = AronNDArraySerializerTypeNavigatorPtr;

            public:
                // virtual methods
                virtual void setDimensions(const std::vector<int>&) = 0;
                virtual void setUsedType(const std::string&) = 0;

                virtual std::string getUsedType() const = 0;
                virtual std::vector<int> getDimensions() const = 0;

                // static methods
                static AronNDArraySerializerTypeNavigatorPtr DynamicCast(const AronTypeNavigatorPtr& n)
                {
                    AronNDArraySerializerTypeNavigatorPtr casted = std::dynamic_pointer_cast<AronNDArraySerializerTypeNavigator>(n);
                    return casted;
                }
            };



            class AronPrimitiveTypeNavigator;
            typedef std::shared_ptr<AronPrimitiveTypeNavigator> AronPrimitiveTypeNavigatorPtr;

            class AronPrimitiveTypeNavigator :
                virtual public AronTypeNavigator
            {
            public:
                using PointerType = AronPrimitiveTypeNavigatorPtr;

                virtual size_t childrenSize() const override;

                // static methods
                static AronPrimitiveTypeNavigatorPtr DynamicCast(const AronTypeNavigatorPtr& n)
                {
                    AronPrimitiveTypeNavigatorPtr casted = std::dynamic_pointer_cast<AronPrimitiveTypeNavigator>(n);
                    return casted;
                }
            };
        }
    }
}
