#pragma once

#include <RobotAPI/libraries/aron/aroncore/navigators/datanavigator/AronAllDataNavigators.h>
#include <RobotAPI/libraries/aron/aroncore/navigators/typenavigator/AronAllTypeNavigators.h>


namespace armarx::aron::visitor
{

    /**
     * @brief A visitor traversing an `aron::Type::AronTypePtr` using
     * `aron::Typenavigator::AronTypeNavigator`s, taking care of type checking,
     * casting and hierarchical traversal.
     *
     * To use it, create a class deriving from `AronTypeVisitor` and
     * override the overloads of `visitEnter()`, `visitExit()` and `visit()`
     * for the types you would like to handle (see blow for a copy-and-paste
     * example).
     * Then, instantiate your derived visitor and call `applyTo()`, passing
     * the `AronTypePtr` or the root `AronTypeNavigator`.
     *
     * For dicts and lists, you will get a `visitEnter()`/`visitExit()` pair.
     * For types without children (e.g. primitive types), you get a single
     * `visit()` call.
     *
     * Each visit function should return true to continue traversal or
     * false to stop it. All default implementations return true.
     *
     * To the the location of the passed navigator in the original `AronType`,
     * use `navigator.getPath()`.
     *
     * @see  `aron::Type::AronTypePtr`
     * @see  `aron::Typenavigator::AronTypeNavigator`
     * @see `std::stack` To manage Type with stack semantics.
     */
    class TypedDataVisitor
    {
    public:

        // Define types to ease writing code in this and derived classes.

        // Data - possible data representations
        using DataNavigator = datanavigator::AronDataNavigator;

        using DictDataNavigator = datanavigator::AronDictDataNavigator;
        using ListDataNavigator = datanavigator::AronListDataNavigator;

        using BoolDataNavigator = datanavigator::AronBoolDataNavigator;
        using DoubleDataNavigator = datanavigator::AronDoubleDataNavigator;
        using FloatDataNavigator = datanavigator::AronFloatDataNavigator;
        using IntDataNavigator = datanavigator::AronIntDataNavigator;
        using LongDataNavigator = datanavigator::AronLongDataNavigator;
        using StringDataNavigator = datanavigator::AronStringDataNavigator;

        using NDArrayDataNavigator = datanavigator::AronNDArrayDataNavigator;


        // Type
        using TypeNavigator = typenavigator::AronTypeNavigator;

        // Dict-valued
        using DictTypeNavigator = typenavigator::AronDictTypeNavigator;
        using ObjectTypeNavigator = typenavigator::AronObjectTypeNavigator;

        // List-valued
        using ListTypeNavigator = typenavigator::AronListTypeNavigator;
        using TupleTypeNavigator = typenavigator::AronTupleTypeNavigator;

        // Primitive-valued
        using BoolTypeNavigator = typenavigator::AronBoolTypeNavigator;
        using DoubleTypeNavigator = typenavigator::AronDoubleTypeNavigator;
        using FloatTypeNavigator = typenavigator::AronFloatTypeNavigator;
        using IntTypeNavigator = typenavigator::AronIntTypeNavigator;
        using LongTypeNavigator = typenavigator::AronLongTypeNavigator;
        using StringTypeNavigator = typenavigator::AronStringTypeNavigator;

        // Array-valued
        using EigenMatrixTypeNavigator = typenavigator::AronEigenMatrixTypeNavigator;
        using IVTCByteImageTypeNavigator = typenavigator::AronIVTCByteImageTypeNavigator;
        using OpenCVMatTypeNavigator = typenavigator::AronOpenCVMatTypeNavigator;
        using PCLPointCloudTypeNavigator = typenavigator::AronPCLPointCloudTypeNavigator;


    public:

        TypedDataVisitor();
        virtual ~TypedDataVisitor();


        bool applyTo(TypeNavigator& type, DataNavigator& data);
        bool applyTo(const std::string& key, TypeNavigator& type, DataNavigator& data);


        virtual bool visitEnter(DictTypeNavigator& type, DictDataNavigator& data)
        {
            (void) type, (void) data;
            return true;
        }
        virtual bool visitExit(DictTypeNavigator& type, DictDataNavigator& data)
        {
            (void) type, (void) data;
            return true;
        }

        virtual bool visitEnter(ObjectTypeNavigator& type, DictDataNavigator& data)
        {
            (void) type, (void) data;
            return true;
        }
        virtual bool visitExit(ObjectTypeNavigator& type, DictDataNavigator& data)
        {
            (void) type, (void) data;
            return true;
        }

        virtual bool visitEnter(ListTypeNavigator& type, ListDataNavigator& data)
        {
            (void) type, (void) data;
            return true;
        }
        virtual bool visitExit(ListTypeNavigator& type, ListDataNavigator& data)
        {
            (void) type, (void) data;
            return true;
        }

        virtual bool visitEnter(TupleTypeNavigator& type, ListDataNavigator& data)
        {
            (void) type, (void) data;
            return true;
        }
        virtual bool visitExit(TupleTypeNavigator& type, ListDataNavigator& data)
        {
            (void) type, (void) data;
            return true;
        }


        virtual bool visit(BoolTypeNavigator& type, BoolDataNavigator& data)
        {
            (void) type, (void) data;
            return true;
        }
        virtual bool visit(DoubleTypeNavigator& type, DoubleDataNavigator& data)
        {
            (void) type, (void) data;
            return true;
        }
        virtual bool visit(FloatTypeNavigator& type, FloatDataNavigator& data)
        {
            (void) type, (void) data;
            return true;
        }
        virtual bool visit(IntTypeNavigator& type, IntDataNavigator& data)
        {
            (void) type, (void) data;
            return true;
        }
        virtual bool visit(LongTypeNavigator& type, LongDataNavigator& data)
        {
            (void) type, (void) data;
            return true;
        }
        virtual bool visit(StringTypeNavigator& type, StringDataNavigator& data)
        {
            (void) type, (void) data;
            return true;
        }


        virtual bool visit(EigenMatrixTypeNavigator& type, NDArrayDataNavigator& data)
        {
            (void) type, (void) type, (void) data;
            return true;
        }
        virtual bool visit(IVTCByteImageTypeNavigator& type, NDArrayDataNavigator& data)
        {
            (void) type, (void) type, (void) data;
            return true;
        }
        virtual bool visit(OpenCVMatTypeNavigator& type, NDArrayDataNavigator& data)
        {
            (void) type, (void) type, (void) data;
            return true;
        }
        virtual bool visit(PCLPointCloudTypeNavigator& type, NDArrayDataNavigator& data)
        {
            (void) type, (void) type, (void) data;
            return true;
        }



        virtual bool visitEnter(DictTypeNavigator& type, const std::string& key, DictDataNavigator& data)
        {
            (void) key;
            visitEnter(type, data);
            return true;
        }
        virtual bool visitExit(DictTypeNavigator& type, const std::string& key, DictDataNavigator& data)
        {
            (void) key;
            visitExit(type, data);
            return true;
        }

        virtual bool visitEnter(ObjectTypeNavigator& type, const std::string& key, DictDataNavigator& data)
        {
            (void) key;
            visitEnter(type, data);
            return true;
        }
        virtual bool visitExit(ObjectTypeNavigator& type, const std::string& key, DictDataNavigator& data)
        {
            (void) key;
            visitExit(type, data);
            return true;
        }

        virtual bool visitEnter(ListTypeNavigator& type, const std::string& key, ListDataNavigator& data)
        {
            (void) key;
            visitEnter(type, data);
            return true;
        }
        virtual bool visitExit(ListTypeNavigator& type, const std::string& key, ListDataNavigator& data)
        {
            (void) key;
            visitExit(type, data);
            return true;
        }

        virtual bool visitEnter(TupleTypeNavigator& type, const std::string& key, ListDataNavigator& data)
        {
            (void) key;
            visitEnter(type, data);
            return true;
        }
        virtual bool visitExit(TupleTypeNavigator& type, const std::string& key, ListDataNavigator& data)
        {
            (void) key;
            visitExit(type, data);
            return true;
        }


        virtual bool visit(BoolTypeNavigator& type, const std::string& key, BoolDataNavigator& data)
        {
            (void) key;
            visit(type, data);
            return true;
        }
        virtual bool visit(DoubleTypeNavigator& type, const std::string& key, DoubleDataNavigator& data)
        {
            (void) key;
            visit(type, data);
            return true;
        }
        virtual bool visit(FloatTypeNavigator& type, const std::string& key, FloatDataNavigator& data)
        {
            (void) key;
            visit(type, data);
            return true;
        }
        virtual bool visit(IntTypeNavigator& type, const std::string& key, IntDataNavigator& data)
        {
            (void) key;
            visit(type, data);
            return true;
        }
        virtual bool visit(LongTypeNavigator& type, const std::string& key, LongDataNavigator& data)
        {
            (void) key;
            visit(type, data);
            return true;
        }
        virtual bool visit(StringTypeNavigator& type, const std::string& key, StringDataNavigator& data)
        {
            (void) key;
            visit(type, data);
            return true;
        }


        virtual bool visit(EigenMatrixTypeNavigator& type, const std::string& key, NDArrayDataNavigator& data)
        {
            (void) type, (void) key;
            visit(type, data);
            return true;
        }
        virtual bool visit(IVTCByteImageTypeNavigator& type, const std::string& key, NDArrayDataNavigator& data)
        {
            (void) type, (void) key;
            visit(type, data);
            return true;
        }
        virtual bool visit(OpenCVMatTypeNavigator& type, const std::string& key, NDArrayDataNavigator& data)
        {
            (void) type, (void) key;
            visit(type, data);
            return true;
        }
        virtual bool visit(PCLPointCloudTypeNavigator& type, const std::string& key, NDArrayDataNavigator& data)
        {
            (void) type, (void) key;
            visit(type, data);
            return true;
        }



    private:

        template <class DataT, class TypeT>
        bool applyToCollection(TypeT& type, const std::string& key, DataNavigator& data);

        bool applyToChildren(ObjectTypeNavigator& type, DictDataNavigator& data);
        bool applyToChildren(DictTypeNavigator& type, DictDataNavigator& data);
        bool applyToChildren(ListTypeNavigator& type, ListDataNavigator& data);
        bool applyToChildren(TupleTypeNavigator& type, ListDataNavigator& data);

    };

}


/* Copy-and-paste example below.
 * Remove functions you don't need.
 * Add `const std::string& key, ` to parameter list if you need an items's key
 * in the parent container.
 */

#if 0

struct MyDerivedAronTypeVisitor : public armarx::aron::visitor::AronTypeVisitor
{

    bool visitEnter(DictTypeNavigator& dict) override
    {
        (void) dict;
        return true;
    }
    bool visitExit(DictTypeNavigator& dict) override
    {
        (void) dict;
        return true;
    }
    bool visitEnter(ListTypeNavigator& list) override
    {
        (void) list;
        return true;
    }
    bool visitExit(ListTypeNavigator& list) override
    {
        (void) list;
        return true;
    }

    bool visit(BoolTypeNavigator& b) override
    {
        (void) b;
        return true;
    }
    bool visit(DoubleTypeNavigator& d) override
    {
        (void) d;
        return true;
    }
    bool visit(FloatTypeNavigator& f) override
    {
        (void) f;
        return true;
    }
    bool visit(IntTypeNavigator& i) override
    {
        (void) i;
        return true;
    }
    bool visit(LongTypeNavigator& l) override
    {
        (void) l;
        return true;
    }
    bool visit(NDArrayTypeNavigator& array) override
    {
        (void) array;
        return true;
    }
    bool visit(StringTypeNavigator& string) override
    {
        (void) string;
        return true;
    }

};

#endif
