/*
 * This file is part of ArmarX.
 *
 * ArmarX is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * ArmarX is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 *
 * @package    RobotAPI::ArmarXObjects::aron
 * @author     Simon Ottenhaus ( simon dot ottenhaus at kit dot edu )
 * @date       2019
 * @copyright  http://www.gnu.org/licenses/gpl-2.0.txt
 *             GNU General Public License
 */

#define BOOST_TEST_MODULE RobotAPI::ArmarXLibraries::aron

#define ARMARX_BOOST_TEST

// STD/STL
#include <iostream>
#include <cstdlib>
#include <ctime>
#include <numeric>

// IVT
#include <Image/ByteImage.h>

// Eigen
#include <Eigen/Core>

// OpenCV
#include <opencv2/core/core.hpp>

// PCL
#include <pcl/point_cloud.h>
#include <pcl/point_types.h>

// Boost
#include <boost/algorithm/string.hpp>

// Test
#include <RobotAPI/Test.h>

// ArmarX
#include <ArmarXCore/libraries/cppgen/CppMethod.h>
#include <ArmarXCore/libraries/cppgen/CppClass.h>
#include <RobotAPI/libraries/aron/aroncore/AronException.h>

// Aron
#include <RobotAPI/libraries/aron/aroncore/AronDebug.h>
#include <RobotAPI/libraries/aron/aroncore/AronRandomizer.h>
#include <RobotAPI/libraries/aron/aroncore/io/AronDataIO/classReaders/NavigatorReader/AronDataNavigatorReader.h>

#include <RobotAPI/libraries/aron/aroncore/io/AronDataIO/classWriters/NavigatorWriter/AronDataNavigatorWriter.h>
#include <RobotAPI/libraries/aron/aroncore/io/AronDataIO/classWriters/NlohmannJSONWriter/AronDataNlohmannJSONWriter.h>
#include <RobotAPI/libraries/aron/aroncore/io/AronDataIO/classWriters/RapidXMLWriter/AronDataRapidXMLWriter.h>
#include <RobotAPI/libraries/aron/aroncore/io/AronDataIO/classReaders/NlohmannJSONReader/AronDataNlohmannJSONReader.h>

// Files without code generation
#include "aronDataWithoutCodeGeneration.h"

// Generated File
#include <RobotAPI/libraries/aron/aroncore/test/aron/NaturalIK.aron.generated.h>
#include <RobotAPI/libraries/aron/aroncore/test/aron/ListTest.aron.generated.h>
#include <RobotAPI/libraries/aron/aroncore/test/aron/SimpleDictTest.aron.generated.h>
#include <RobotAPI/libraries/aron/aroncore/test/aron/PrimitiveTest.aron.generated.h>
#include <RobotAPI/libraries/aron/aroncore/test/aron/SimpleObjectTest.aron.generated.h>
#include <RobotAPI/libraries/aron/aroncore/test/aron/ImageTest.aron.generated.h>
#include <RobotAPI/libraries/aron/aroncore/test/aron/EigenMatrixTest.aron.generated.h>
#include <RobotAPI/libraries/aron/aroncore/test/aron/OpenCVMatTest.aron.generated.h>
#include <RobotAPI/libraries/aron/aroncore/test/aron/PCLPointCloudTest.aron.generated.h>


using namespace armarx;
using namespace aron;

template <typename T>
void runTestWithInstances(T& k1, T& k2)
{
    AronRandomizer r;

    std::cout << "\t getting type 1" << std::endl;
    typenavigator::AronObjectTypeNavigatorPtr k1_type_nav = k1.toInitialAronType();
    type::AronObjectTypePtr k1_type = k1_type_nav->toAronObjectTypePtr();
    BOOST_CHECK_NE(k1_type.get(), nullptr);

    std::cout << "\t getting type 2" << std::endl;
    typenavigator::AronObjectTypeNavigatorPtr k2_type_nav = k2.toInitialAronType();
    type::AronObjectTypePtr k2_type = k2_type_nav->toAronObjectTypePtr();
    BOOST_CHECK_NE(k2_type.get(), nullptr);

    std::cout << "\t getting aron 1" << std::endl;
    datanavigator::AronDictDataNavigatorPtr k1_aron_nav = k1.toAron();
    data::AronDictPtr k1_aron = k1_aron_nav->toAronDictPtr();
    BOOST_CHECK_NE(k1_aron.get(), nullptr);

    std::cout << "\t initialize aron 1 randomly" << std::endl;
    r.initializeRandomly(k1_aron_nav, k1_type_nav);

    //std::cout << "K Aron:" << std::endl;
    //std::cout << AronDebug::AronDataPtrToString(k_aron) << std::endl;

    std::cout << "\t setting aron 2 from aron 1" << std::endl;
    k2.fromAron(k1_aron);

    std::cout << "\t getting aron 2" << std::endl;
    datanavigator::AronDictDataNavigatorPtr k2_aron_nav = k2.toAron();
    data::AronDictPtr k2_aron = k2_aron_nav->toAronDictPtr();
    BOOST_CHECK_NE(k2_aron.get(), nullptr);

    //std::cout << "K2 Aron:" << std::endl;
    //std::cout << AronDebug::AronDataPtrToString(k2_aron) << std::endl;

    std::cout << "\t setting aron 1 from aron 2 and check for equality" << std::endl;
    k1.fromAron(k2_aron_nav);

    BOOST_CHECK_EQUAL((k1 == k2), true);

    //std::cout << "\t getting current type 1" << std::endl;
    //typenavigator::AronTypeNavigatorPtr k1_current_type_nav = k1.toCurrentAronType();
    //type::AronTypePtr k1_current_type = k1_current_type_nav->getResult();
    //BOOST_CHECK_NE(k1_current_type.get(), nullptr);

    //std::cout << "\t getting type 2" << std::endl;
    //typenavigator::AronTypeNavigatorPtr k2_current_type_nav = k2.toCurrentAronType();
    //type::AronTypePtr k2_current_type = k2_current_type_nav->getResult();
    //BOOST_CHECK_NE(k2_current_type.get(), nullptr);

    r.initializeRandomly(k1_aron_nav, k1_type_nav);
    k1.fromAron(k1_aron);

    std::cout << "\t check JSON export of k and k2 for equality" << std::endl;
    armarx::aron::io::AronDataNlohmannJSONWriter json_writer_for_k1;
    armarx::aron::io::AronDataNlohmannJSONWriter json_writer_for_k2;

    k1.write(json_writer_for_k1);
    nlohmann::json k1_aron_json = json_writer_for_k1.getResult();
    std::string k1_aron_json_str = k1_aron_json.dump(4);

    armarx::aron::io::AronDataNlohmannJSONWriter direct_json_writer_for_k1;
    armarx::aron::io::AronDataWriterVisitor::SetupWriterFromAronDataPtr(direct_json_writer_for_k1, k1_aron);
    nlohmann::json direct_k1_aron_json = direct_json_writer_for_k1.getResult();
    std::string direct_k1_aron_json_str = direct_k1_aron_json.dump(4);
    BOOST_CHECK_EQUAL((k1_aron_json_str == direct_k1_aron_json_str), true);

    armarx::aron::io::AronDataNlohmannJSONReader json_reader_for_k2(k1_aron_json);
    k2.read(json_reader_for_k2);

    k2.write(json_writer_for_k2);
    nlohmann::json k2_aron_json = json_writer_for_k2.getResult();
    std::string k2_aron_json_str = k2_aron_json.dump(4);

    //std::cout << "\t K1 as json: " << std::endl << k1_aron_json_str << std::endl;
    //std::cout << "\t K2 as json: " << std::endl << k2_aron_json_str << std::endl;
    BOOST_CHECK_EQUAL((k1_aron_json_str == k2_aron_json_str), true);

    //std::cout << "\t check XML export of k and k2 for equality" << std::endl;
    //armarx::aron::io::AronDataRapidXMLWriter xml_writer1;
    //armarx::aron::io::AronDataRapidXMLWriter xml_writer2;

    //k.write(xml_writer1);
    //k_aron_rapid_xml = xml_writer1.getResult();

    //k2.write(xml_writer2);
    //k2_aron_xml = xml_writer2.getResult();

    //std::cout << "\t K as XML: " << std::endl << k_aron_xml << std::endl;
    //std::cout << "\t K2: " << std::endl << k2_aron_json << std::endl;
    //BOOST_CHECK_EQUAL((k_aron_xml == k2_aron_xml), true);*/
}


BOOST_AUTO_TEST_CASE(AronNaturalIKTest)
{
    std::cout << "Running NaturalIK test" << std::endl;
    NaturalIKResult k;
    NaturalIKResult k2;
    runTestWithInstances<NaturalIKResult>(k, k2);

    // also test AronPath
    datanavigator::AronDictDataNavigatorPtr aron = k.toAron();
    AronPath path = aron->getPath(); // should be empty since aron is top level object
    AronPath memberReached(path, "reached");
    AronPath memberJointValues(path, "jointValues");
    AronPath indexJointValues0(memberJointValues, "0");
    AronPath indexJointValues1(memberJointValues, "1");

    BOOST_CHECK_EQUAL(path.toString(), "\\");
    BOOST_CHECK_EQUAL(path.size(), 0);

    BOOST_CHECK_EQUAL(memberReached.toString(), "\\->reached");
    BOOST_CHECK_EQUAL(memberReached.size(), 1);

    BOOST_CHECK_EQUAL(memberJointValues.toString(), "\\->jointValues");
    BOOST_CHECK_EQUAL(memberJointValues.size(), 1);

    BOOST_CHECK_EQUAL(indexJointValues0.toString(), "\\->jointValues->0");
    BOOST_CHECK_EQUAL(indexJointValues0.size(), 2);

    BOOST_CHECK_EQUAL(indexJointValues1.toString(), "\\->jointValues->1");
    BOOST_CHECK_EQUAL(indexJointValues1.size(), 2);

    datanavigator::AronBoolDataNavigatorPtr reached = datanavigator::AronBoolDataNavigator::DynamicCastAndCheck(aron->navigateAbsolute(memberReached));
    datanavigator::AronListDataNavigatorPtr jointValues = datanavigator::AronListDataNavigator::DynamicCastAndCheck(aron->navigateAbsolute(memberJointValues));

    if (jointValues->childrenSize() > 0)
    {
        datanavigator::AronFloatDataNavigatorPtr el = datanavigator::AronFloatDataNavigator::DynamicCastAndCheck(aron->navigateAbsolute(indexJointValues0));
    }
    if (jointValues->childrenSize() > 1)
    {
        datanavigator::AronFloatDataNavigatorPtr el = datanavigator::AronFloatDataNavigator::DynamicCastAndCheck(aron->navigateAbsolute(indexJointValues1));
    }

    AronPath diff = indexJointValues1.getWithoutPrefix(indexJointValues0);
    BOOST_CHECK_EQUAL(diff.toString(), "\\->1");
    BOOST_CHECK_EQUAL(diff.size(), 1);
}


BOOST_AUTO_TEST_CASE(AronListTest)
{
    std::cout << "Running List test" << std::endl;
    ListTest l;
    ListTest l2;
    runTestWithInstances<ListTest>(l, l2);
}

BOOST_AUTO_TEST_CASE(AronDictTest)
{
    std::cout << "Running Dict test" << std::endl;
    DictTest d;
    DictTest d2;
    runTestWithInstances<DictTest>(d, d2);
}

BOOST_AUTO_TEST_CASE(AronPrimitiveTest)
{
    std::cout << "Running Primitive test" << std::endl;
    PrimitiveTest p;
    PrimitiveTest p2;
    runTestWithInstances<PrimitiveTest>(p, p2);
}

BOOST_AUTO_TEST_CASE(AronSimpleClassTest)
{
    std::cout << "Running SimpleClass test" << std::endl;
    SimpleClass o1;
    SimpleClass o12;
    runTestWithInstances<SimpleClass>(o1, o12);
}

BOOST_AUTO_TEST_CASE(AronObjectTest)
{
    std::cout << "Running Object test" << std::endl;
    ObjectTest o2;
    ObjectTest o22;
    runTestWithInstances<ObjectTest>(o2, o22);
}

BOOST_AUTO_TEST_CASE(AronImageTest)
{
    std::cout << "Running Image test" << std::endl;
    ImageTest i;
    ImageTest i2;
    runTestWithInstances<ImageTest>(i, i2);
}

BOOST_AUTO_TEST_CASE(AronImageChangeSizeTest)
{
    std::cout << "Running Image test with changed sizes" << std::endl;
    ImageTest ii;
    ImageTest ii2;
    ii.the_ivt_image->Set(100, 100, ii.the_ivt_image->type);
    ii2.the_ivt_image->Set(10, 10, ii.the_ivt_image->type);
    runTestWithInstances<ImageTest>(ii, ii2);
}

BOOST_AUTO_TEST_CASE(AronEigenMatrixTest)
{
    std::cout << "Running EigenMatrix test" << std::endl;
    EigenMatrixTest em;
    EigenMatrixTest em2;
    runTestWithInstances<EigenMatrixTest>(em, em2);
}

BOOST_AUTO_TEST_CASE(AronOpenCVTest)
{
    std::cout << "Running OpenCVMat test" << std::endl;
    OpenCVMatTest ocv;
    OpenCVMatTest ocv2;
    runTestWithInstances<OpenCVMatTest>(ocv, ocv2);
}

BOOST_AUTO_TEST_CASE(AronPCLPointCloudTest)
{
    std::cout << "Running PCLPointCloud test" << std::endl;
    PointCloudTest pc;
    PointCloudTest pc2;
    runTestWithInstances<PointCloudTest>(pc, pc2);
}

/*
BOOST_AUTO_TEST_CASE(AronTests)
{
    initialize_random();

    std::cout << "Running (Raphael) Trivial test" << std::endl;
    AronTrivialDataWithoutCodeGeneration rt1;
    AronTrivialDataWithoutCodeGeneration rt2;
    runTestWithInstances<AronTrivialDataWithoutCodeGeneration>(rt1, rt2);

    std::cout << "Running (Raphael) Simple test" << std::endl;
    AronSimpleDataWithoutCodeGeneration rs1;
    AronSimpleDataWithoutCodeGeneration rs2;
    runTestWithInstances<AronSimpleDataWithoutCodeGeneration>(rs1, rs2);

    std::cout << "Running (Raphael) Nested test" << std::endl;
    AronNestedDataWithoutCodeGeneration rn1;
    AronNestedDataWithoutCodeGeneration rn2;
    runTestWithInstances<AronNestedDataWithoutCodeGeneration>(rn1, rn2);
}*/
